#include "SortLast.h"

#include <stdio.h>
#include <stdlib.h>

#include <iostream>

#include <GL/glut.h>

#include "glutredirect.h"
#include "saveppm.h"

// XC FIXME: hardware compositing has to be optimized (we glRead too
// much pixels)
#undef HARDWARE_COMPOSITING

#undef DEBUG_TRACE_SORT_LAST

#undef DEBUG_SAVE_PPM

#undef DEBUG_TIME_READ_PIXELS
#undef DEBUG_TIME_DRAW_PIXELS
#if defined(DEBUG_TIME_READ_PIXELS) || defined(DEBUG_TIME_DRAW_PIXELS)
#include "Timer.h"
#endif

// XC FIXME:
#define MAXNODES 10000

// XC FIXME: order of read pixels? (OK for Linux)
#define R 3
#define G 2
#define B 1
#define A 0

using namespace std;

//____________________________________________________________________________
//
SortLast::SortLast(unsigned int stages, unsigned int *config, 
		   int width, int height,
		   MPI_Comm &worldComm, MPI_Comm &slavesComm, 
		   int slavesSize, int masterIsSlave) :
  Stages(0), Config(NULL),
  Width(width), Height(height),
  WorldComm(worldComm), SlavesComm(slavesComm), 
  SlavesSize(slavesSize), MasterIsSlave(masterIsSlave),
  FrameBuffer(NULL), TmpFrameBuffer(NULL), WBehindFrameBuffer(NULL)
  //____________________________________________________________________________
{
  MPI_Comm_size(this->WorldComm, &this->WorldSize);
  MPI_Comm_rank(this->WorldComm, &this->MyMpiRank);
  this->MyAlgRank = -1;
  
#ifdef DEBUG_TRACE_SORT_LAST
  char traceName[255];
  sprintf(traceName, "trace%d.info", this->MyMpiRank);
  this->MyTrace = new ofstream(traceName);	  
  if ( ! this->MyTrace->is_open() )
    {
      cerr << "Can not open file " << traceName << endl;
      this->MyTrace = NULL;
    }
#endif
  
  unsigned int check = 1;
  for ( int i = 0; i < stages; ++i ) check *= config[i];
  if ( check != this->SlavesSize || stages < 2 )
    {
      cerr << "Invalid config, switching to parallel pipeline." << endl;
      this->Stages = 2;
      this->Config = new unsigned int [this->Stages];
      this->Config[0] = this->SlavesSize;
      this->Config[1] = 1;
    }
  else
    {
      this->Stages = stages;
      this->Config = new unsigned int [this->Stages];
      for ( int i = 0; i < this->Stages; ++i )
	{
	  this->Config[i] = config[i];
	}
    }
  
  this->Resize();
}

//____________________________________________________________________________
//
SortLast::~SortLast()
  //____________________________________________________________________________
{
  free(this->FrameBuffer);
  free(this->TmpFrameBuffer);
  free(this->WBehindFrameBuffer);

#ifdef DEBUG_TRACE_SORT_LAST
  delete this->MyTrace;
#endif
}

//____________________________________________________________________________
//
void SortLast::Resize(int width, int height)
  //____________________________________________________________________________
{
  //alog(<< "SL: started Resize(w,h) to " << width << "," << height << endl);

  if ( this->Width != width || this->Height != height )
    {
      this->Width = width;
      this->Height = height;
      this->Resize();
    }

  //alog(<< "SL: finished Resize(w,h) to " << width << "," << height << endl);
}

//____________________________________________________________________________
//
void SortLast::Resize(void)
  //____________________________________________________________________________ 
{
  //alog(<< "SL: started Resize() " << this->Width << "x" << this->Height << endl);

  this->Size = this->Width * this->Height; 

  int tsize = (this->Width + (8-(this->Width % 8))) * this->Height; //account for 8 byte alignment
  this->AdjustedSize = tsize + tsize % this->SlavesSize;
  
  free(this->FrameBuffer);
  this->FrameBuffer = 
    (uint32_t *) malloc(this->AdjustedSize * sizeof(uint32_t));
  
  free(this->TmpFrameBuffer);
  this->TmpFrameBuffer = 
    (uint32_t *) malloc(this->AdjustedSize * sizeof(uint32_t));

  free(this->WBehindFrameBuffer);
  this->WBehindFrameBuffer = 
    (uint32_t *) malloc(this->AdjustedSize * sizeof(uint32_t));

  //alog(<< "SL: finished Resize()" << endl);
}

//____________________________________________________________________________
//
void SortLast::Compose(MpiIdType *algRankToProc, int *algProcToRank)
  //____________________________________________________________________________
{ 
  // Find out total area.
  unsigned int totalArea = this->AdjustedSize;   

  //alog(<< "compose size = " << totalArea << endl);

  // Each processor calculates beginning of its active area.
  unsigned int beginActiveArea, beginSendArea;
  unsigned int beginRecvArea = 0;
  
  // The gap.
  unsigned int gap = 1;

  // The left and right neighbours.
  unsigned int leftNeighbour, rightNeighbour;
  
  // Misc.
  MPI_Status status;   
  unsigned tmpRank;
  
  if ( this->MyMpiRank || this->MasterIsSlave )
    {
      this->MyAlgRank = algProcToRank[this->MyMpiRank];

#ifdef DEBUG_TRACE_SORT_LAST
      if ( this->MyTrace )
	(*this->MyTrace) << "MPI #" << this->MyMpiRank 
			 << " composes a new frame in rank " 
			 << this->MyAlgRank << "..." << endl; 
#endif
      
      // Get the current frame buffer (subvolume drawn by itself).
      this->GetFrameBuffer();

      alog(<<"s barr 1" << endl);
      MPI_Barrier(this->SlavesComm);
      alog(<<"e barr 1" << endl);

#ifdef DEBUG_SAVE_PPM
      {
	char tmp[256];
	sprintf(tmp, "s0-p%d.ppm", this->MyMpiRank);
	save_ppm(tmp, this->Width, this->Height, 
		 4, (const unsigned char *)(this->FrameBuffer));
      }
#endif
      
      // Do the compositing stuff.
      // _________________________
      //
      
      // Iterate over the stages.
      // ________________________
      //
      for ( int s = 0; s < this->Stages; ++s )
	{
#ifdef DEBUG_TRACE_SORT_LAST
	  if ( this->MyTrace )
	    (*this->MyTrace) << "  * entering stage " << s + 1 
			     << "/" << this->Stages << endl;
#endif
	  
	  // Compute the total area.
	  totalArea = totalArea / this->Config[s];
	  
	  // Calculate the beginning of active area.
	  beginActiveArea = beginRecvArea;
	  
	  // Calculate gap and wrap around.
	  if ( s != 0 ) gap = gap * this->Config[s - 1];   
	  
	  // Find the left and right neighbour for each processor.
	  //
	  
	  // Find right neighbour.
	  if ( ((this->MyAlgRank / gap) % this->Config[s]) == (this->Config[s] - 1) )
	    {
	      // Wrap around.
	      rightNeighbour = this->MyAlgRank - gap * (this->Config[s] - 1);
	    }
	  else
	    {
	      rightNeighbour = this->MyAlgRank + gap;
	    }
	  
	  // Find left neighbour.
	  if ( ((this->MyAlgRank / gap) % this->Config[s]) == 0 )
	    {
	      // Wrap around.
	      leftNeighbour = this->MyAlgRank + gap * (this->Config[s] - 1);
	    }
	  else
	    {
	      leftNeighbour = this->MyAlgRank - gap;
	    }
	  
	  
	  // Calculate begin of send and receive area.
	  beginSendArea = beginActiveArea + 
	    totalArea * ((this->MyAlgRank / gap) % this->Config[s]);
	  
	  if ( ((this->MyAlgRank / gap) % this->Config[s]) == 0 )
	    {
	      // Wrap around.
	      beginRecvArea = beginActiveArea  + 
		totalArea * (this->Config[s] - 1);
	    } 
	  else 
	    {
	      beginRecvArea = beginActiveArea + 
		totalArea * (((this->MyAlgRank / gap) % this->Config[s]) - 1);
	    }
	  
	  // Now the iterations.
	  // ___________________
	  //
	  for ( int it = 0; it < this->Config[s] - 1; ++it )
	    {
#ifdef DEBUG_TRACE_SORT_LAST
	      if ( this->MyTrace )
		(*this->MyTrace) << "	 - entering iteration " 
				 << it + 1 << "/" << this->Config[s] - 1 << endl;
#endif
	      
	      // Find the send and receive boundary.
	      if ( it != 0 )
		{
		  beginSendArea = beginRecvArea;
		  if ( this->MyAlgRank < gap * it )
		    {
		      // XC FIXME: check this "1 *"
		      tmpRank = this->SlavesSize + this->MyAlgRank - 1 * gap * it;
		    }
		  else
		    {
		      tmpRank = this->MyAlgRank - gap * it;
		    }
		  
		  if ( ((tmpRank / gap) % this->Config[s]) == 0 )
		    {
		      // Wrap around.
		      beginRecvArea = beginActiveArea  + 
			totalArea * (this->Config[s] - 1);
		    } 
		  else 
		    {
		      beginRecvArea = beginActiveArea + 
			totalArea * (((tmpRank / gap) % this->Config[s]) - 1);
		    }
		}
	      
#ifdef DEBUG_TRACE_SORT_LAST
	      if ( this->MyTrace )
		(*this->MyTrace) << "	   sends " << totalArea << " pixels"
				 << " over " << this->Size
				 << " starting " << beginSendArea
				 << " to #" << algRankToProc[rightNeighbour]
				 << " and receive the same from #" 
				 << algRankToProc[leftNeighbour]
				 << endl;
#endif
	      
#ifdef MPE_LOGGING
	      MPE_Log_event(MPE_EVENT_SEND_RECV_PIXELS_B, 0, "sendrecv");
#endif
	      
	      // Do actual send and receive.
	      alog(<<"s srecv" << endl);
	      MPI_Sendrecv(this->FrameBuffer + beginSendArea, totalArea, 
			   MPI_UNSIGNED, algRankToProc[rightNeighbour], 
			   TAG_PBUF,
			   this->TmpFrameBuffer, totalArea, 
			   MPI_UNSIGNED, algRankToProc[leftNeighbour], 
			   TAG_PBUF, 
			   this->WorldComm, &status);
	      alog(<<"e srecv" << endl);
	      
#ifdef MPE_LOGGING
	      MPE_Log_event(MPE_EVENT_SEND_RECV_PIXELS_E, 0, "sentrecvd");
#endif
	      
	      // Do composition.
	      if ( this->MyAlgRank < leftNeighbour )
		{
#ifdef DEBUG_TRACE_SORT_LAST
		  if ( this->MyTrace )
		    (*this->MyTrace) << "      I am back compared to #"
				     << algRankToProc[leftNeighbour]
				     << endl;
#endif
		  this->DoComposition(beginRecvArea, totalArea, 0);
		}
	      else
		{
#ifdef DEBUG_TRACE_SORT_LAST
		  if ( this->MyTrace )
		    (*this->MyTrace) << "      I am front compared to #"
				     << algRankToProc[leftNeighbour]
				     << endl;
#endif
		  this->DoComposition(beginRecvArea, totalArea, 1);
		}
	      
#ifdef DEBUG_SAVE_PPM
	      {
		char tmp[256];
		sprintf(tmp, "s%d-i%d-p%d.ppm", s+1, it+1, this->MyMpiRank);
		save_ppm(tmp, this->Width, this->Height, 
			 4, (const unsigned char *)(this->FrameBuffer));
	      }
#endif
	      
	      alog(<<"s barr 2" << endl);
	      MPI_Barrier(this->SlavesComm);
	      alog(<<"e barr 2" << endl);
	    } // End of iterations.
	} // End of stages.
      
      // comp_time = MPI_Wtime() - start;
    }
  
#ifdef DEBUG_TRACE_SORT_LAST
  if ( this->MyTrace )
    (*this->MyTrace) << "  * final gathering" << endl;
#endif
  
  // Determine which segment each processor is sending.
  // __________________________________________________
  //
  unsigned int tmpRecv, tmpTotal;
  int where;
  unsigned int which[MAXNODES];
  if ( this->MyMpiRank == 0 )
    {
      for( int p = 0; p < this->SlavesSize; ++p )
	{
	  tmpRecv =0;
	  which[p] = 0;
	  gap = 1;
	  tmpTotal = this->AdjustedSize;
	  
	  for ( int i = 0; i < this->Stages; ++i )
	    {
	      tmpTotal = tmpTotal / this->Config[i];
	      if ( i != 0 )
		{
		  gap = gap * this->Config[i-1];
		}
	      if ( ((p / gap) % this->Config[i]) == 0 )
		{
		  tmpRecv = which[p] + 
		    tmpTotal * (this->Config[i] - 1);
		}
	      else
		{
		  tmpRecv = which[p] + 
		    tmpTotal * (((p / gap) % this->Config[i]) - 1);
		}
	      
	      for (int j = 0; j < this->Config[i] - 1; ++j )
		{
		  if ( j !=0 )
		    {
		      if ( p < gap * j )
			{
			  // XC FIXME: check this "1 *"
			  tmpRank = this->SlavesSize + p - 1 * gap * j;
			}
		      else
			{
			  tmpRank = p - gap * j;
			}
		      if ( ((tmpRank / gap) % this->Config[i]) == 0 )
			{
			  tmpRecv = which[p] + 
			    tmpTotal * (this->Config[i] - 1);
			}
		      else
			{
			  tmpRecv = which[p] + 
			    tmpTotal * (((tmpRank / gap) % this->Config[i]) - 1);
			}
		    }
		}
	      which[p] = tmpRecv;
	    }
	}
    }
  
  // Send results back to a single node, 0 in this case.
  // ___________________________________________________
  //
  
#ifdef MPE_LOGGING
  if ( this->MyMpiRank == 0 )
    {
      MPE_Log_event(MPE_EVENT_COLLECT_PIXELS_B, 0, "recv");
    }
#endif

  CompositeBackWidgets(beginRecvArea, totalArea);
  
  
  for ( int p = 1; p < this->WorldSize; ++p )
    {
      if ( this->MyMpiRank == p )
	{
#ifdef DEBUG_TRACE_SORT_LAST
	  if ( this->MyTrace )
	    (*this->MyTrace) << "      sends final from " << beginRecvArea 
			     << " for " << totalArea << " pixels" << endl;
#endif
	  
#ifdef MPE_LOGGING
	  MPE_Log_event(MPE_EVENT_COLLECT_PIXELS_B, 0, "send");
#endif
	  //alog(<< "Sending " << beginRecvArea << " + " << totalArea << endl);
	  alog(<<"s send" << endl);
	  MPI_Send(this->FrameBuffer + beginRecvArea, totalArea,
		   MPI_UNSIGNED, 0, TAG_PBUF, this->WorldComm);
	  alog(<<"e send" << endl);
	  
#ifdef MPE_LOGGING
	  MPE_Log_event(MPE_EVENT_COLLECT_PIXELS_E, 0, "sent");
#endif
	}
      else if ( this->MyMpiRank == 0 )
	{
	  where = which[algProcToRank[p]];
	  
	  //alog(<< "Receiving " << where << " + " << totalArea << " from " << p << endl);
	  alog(<<"s recv" << endl);
	  MPI_Recv(this->FrameBuffer + where, totalArea, 
	     MPI_UNSIGNED, p, TAG_PBUF, this->WorldComm, &status);
	  alog(<<"e recv" << endl);
	  
#ifdef DEBUG_TRACE_SORT_LAST
	  if ( this->MyTrace )
	    (*this->MyTrace) << "      receives final from #" << p
			     << " at " << where 
			     << " for " << totalArea << " pixels" << endl;
#endif
	}
      
      // XC FIXME: see if there is a way to remove the barrier! The
      // barrier is better to know where to put the received buffer.
      alog(<<"s barr 3" << endl);
      MPI_Barrier(this->WorldComm);
      alog(<<"e barr 3" << endl);
      // XC FIXME: actually, this barrier does not need to be. We
      // shall do here non blocking send and receive to overlap with
      // some computations.
    }
  
#ifdef MPE_LOGGING
  if ( this->MyMpiRank == 0 )
    {
      MPE_Log_event(MPE_EVENT_COLLECT_PIXELS_E, 0, "recvd");
    }
#endif
  
  
  //gather_time = MPI_Wtime() - (comp_time+start);
  
#ifdef DEBUG_TRACE_SORT_LAST
  if ( this->MyTrace )
    (*this->MyTrace) << "MPI #" << this->MyMpiRank 
		     << " ends the frame." << endl << endl; 
#endif

#ifdef DEBUG_SAVE_PPM
  once = 1;
#endif

}

//____________________________________________________________________________
//
void SortLast::Display(void* font, char* string)
  //____________________________________________________________________________
{
#ifdef DEBUG_TIME_DRAW_PIXELS
  Timer_t start = Timer::GetRealNsec();
#endif

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  gluOrtho2D(0.0, (GLfloat) this->Width, 0.0, (GLfloat) this->Height);
  
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_LIGHTING);
  glDisable(GL_BLEND);
  
  glPixelStorei(GL_PACK_ALIGNMENT, 8);
  
  glRasterPos2i(0, 0);
  glDrawPixels(this->Width, this->Height, GL_RGBA, GL_UNSIGNED_INT_8_8_8_8, 
	       this->FrameBuffer);

  if (worldId == 0) {
    if ( font && string )
      {
	float tcolor = 0.5f;//0.5;//(float)(1-gluvv.env.bgColor);
	glColor3f(tcolor, tcolor, tcolor);
	glRasterPos2i(15, 15);
	for (char *c = string; *c != '\0'; ++c) glutBitmapCharacter(font, *c);
      }
  }

  glEnable(GL_DEPTH_TEST);

  glMatrixMode(GL_PROJECTION);
  glPopMatrix();

  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();

#ifdef DEBUG_TIME_DRAW_PIXELS  
  Timer_t stop = Timer::GetRealNsec();
  
  cerr << "glDrawPixels (" << this->Width << "x" << this->Height << "): "
       << Timer::DeltaMsec(start, stop) / 1000.0 << " s, that's "
       << (stop - start) / (this->Width * this->Height)
       << " ns per pixel, or "
       << 1000 * (this->Width * this->Height) / Timer::DeltaMsec(start, stop)
       << " pixels per s" << endl;
#endif
}

//____________________________________________________________________________
//
void SortLast::GetFrameBuffer(void)
  //____________________________________________________________________________
{
#ifdef MPE_LOGGING
  MPE_Log_event(MPE_EVENT_READ_PIXELS_B, 0, "read");
#endif
  
#ifdef DEBUG_TIME_READ_PIXELS
  Timer_t start = Timer::GetRealNsec();
#endif

  // Read back the back frame buffer.
  // ________________________________
  //
  
  // XC FIXME: check receive format and pack (See OpenGL1.2 page 303).
  
  glReadBuffer(GL_BACK);
  
  glPixelStorei(GL_UNPACK_ALIGNMENT, 8);
  
  glReadPixels(0, 0, this->Width, this->Height, 
  	       GL_RGBA, GL_UNSIGNED_INT_8_8_8_8, this->FrameBuffer);

#ifdef DEBUG_TIME_READ_PIXELS  
  Timer_t stop = Timer::GetRealNsec();
  
  cerr << "glReadPixels (" << this->Width << "x" << this->Height << "): "
       << Timer::DeltaMsec(start, stop) / 1000.0 << " s, that's "
       << (stop - start) / (this->Width * this->Height)
       << " ns per pixel, or "
       << 1000 * (this->Width * this->Height) / Timer::DeltaMsec(start, stop)
       << " pixels per s" << endl;
#endif

#ifdef MPE_LOGGING
  MPE_Log_event(MPE_EVENT_READ_PIXELS_E, 0, "readed");
#endif
}

//____________________________________________________________________________
//
void SortLast::DoComposition(unsigned int startArea, unsigned int area,
			     unsigned int reverse)
  //____________________________________________________________________________
{
#ifdef MPE_LOGGING
  MPE_Log_event(MPE_EVENT_COMPOSE_B, 0, "compose");
#endif

#ifdef DEBUG_TRACE_SORT_LAST
  if ( this->MyTrace )
    (*this->MyTrace) << "      does composition starting " << startArea 
		     << " for " << area << " pixels" << endl;
#endif

#ifdef HARDWARE_COMPOSITING
  
  // Hardware compositing
  // ____________________
  //
  glPushMatrix();
  {
    glDrawBuffer(GL_BACK);

    glDisable(GL_DEPTH_TEST);

    glEnable(GL_BLEND);

    if ( reverse )
      {
	glBlendFunc(GL_ONE_MINUS_DST_ALPHA, GL_ONE);
      }
    else
      {
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
      }
    
#ifdef DEBUG_TRACE_SORT_LAST
    if ( this->MyTrace ) {    
      (*this->MyTrace) << "	   glRaster at 0, " << startArea / this->Width << endl;
      (*this->MyTrace) << "	   glDraw " << this->Width << "x" << area/this->Width << endl;
    }
#endif
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0.0, (GLdouble) this->Width, 0.0, (GLdouble) this->Height);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glRasterPos2i(0, startArea / this->Width);	
    glDrawPixels(this->Width, area / this->Width, GL_RGBA, GL_UNSIGNED_INT_8_8_8_8, 
		 this->TmpFrameBuffer);

    glDisable(GL_BLEND);

#ifdef DEBUG_TRACE_SORT_LAST
    if ( this->MyTrace ) {    
      (*this->MyTrace) << "	   ";
      (*this->MyTrace) << 0 << "," << startArea / this->Width << " - ";
      (*this->MyTrace) << this->Width << "," << startArea / this->Width << " - ";
      (*this->MyTrace) << this->Width << "," << (startArea + area) / this->Width << " - ";
      (*this->MyTrace) << 0 << "," << (startArea + area) / this->Width << endl;
    }
#endif

#ifdef DEBUG_SAVE_PPM
    glColor3f(1.0, 1.0, 1.0);
    glBegin(GL_LINE_LOOP);
    glVertex2i(0, startArea / this->Width);
    glVertex2i(this->Width, startArea / this->Width);
    glVertex2i(this->Width, (startArea + area) / this->Width);
    glVertex2i(0, (startArea + area) / this->Width);
    glEnd();
#endif

    glEnable(GL_DEPTH_TEST);
  }
  glPopMatrix();

  this->GetFrameBuffer();

#else

  // Software compositing
  // ____________________
  //  
  unsigned char *out = (unsigned char *)(this->FrameBuffer + startArea);
  unsigned char *back;
  unsigned char *front;
  
  out = (unsigned char *)(this->FrameBuffer + startArea);
  
  if ( ! reverse )
    {
      back  = (unsigned char *)(this->FrameBuffer + startArea);
      front = (unsigned char *)(this->TmpFrameBuffer);	    
    }
  else
    {
      back  = (unsigned char *)(this->TmpFrameBuffer);
      front = (unsigned char *)(this->FrameBuffer + startArea);
    }
  
  long i;
  unsigned int tmp1, tmp2;
  unsigned char t;
#ifdef _OPENMP
#pragma omp parallel for \
   private(i, t, tmp1, tmp2) \
   shared(area, out, front, back)
#endif
  for ( i = 0; i < 4 * area; i += 4 )
    { 
      // XC FIXME: find how to compose the right way!!!
      //
      // out[R] = (front[A] * front[R] + (255 - front[A]) * back[R]) / 255;
      // out[G] = (front[A] * front[G] + (255 - front[A]) * back[G]) / 255;
      // out[B] = (front[A] * front[B] + (255 - front[A]) * back[B]) / 255;
      // out[A] = (255	    * front[A] + (255 - front[A]) * back[A]) / 255;
     
      // XC NOTE: there is a special case when front[i+A] is 1, but it
      // does not happen with us.
      
#if 1 
      if ( front[i+A] == 0 )
	{
	  out[i+R] = front[i+R] + back[i+R];
	  out[i+G] = front[i+G] + back[i+G];
	  out[i+B] = front[i+B] + back[i+B];
	  out[i+A] = back[i+A];
	}
      else
	{
	  t = 255 - front[i+A];
	  out[i+R] = (255 * front[i+R] + t * back[i+R]) / 255;
	  out[i+G] = (255 * front[i+G] + t * back[i+G]) / 255;
	  out[i+B] = (255 * front[i+B] + t * back[i+B]) / 255;
	  out[i+A] = (255 * front[i+A] + t * back[i+A]) / 255;	  
	}
#else
      // XC FIXME: this code with bit shifting is faster, but it is
      // wrong in order to avoid dividing by 65280.

      if ( front[i+A] == 0 )
	{
	  out[i+R] = front[i+R] + back[i+R];
	  out[i+G] = front[i+G] + back[i+G];
	  out[i+B] = front[i+B] + back[i+B];
	  out[i+A] = back[i+A];
	}
      else
	{
	  t = 255 - front[i+A];
	  
	  tmp1 = front[i+R];
	  tmp1 <<= 8;
	  tmp1 = (tmp1 - front[i+R]) + t * back[i+R];	   
	  // tmp2 = tmp1 / 65280;
	  // tmp1 >>= 8;
	  // out[i+R] = tmp1 + tmp2;
	  tmp1 >>= 8;
	  out[i+R] = tmp1;
	  
	  tmp1 = front[i+G];
	  tmp1 <<= 8;
	  tmp1 = (tmp1 - front[i+G]) + t * back[i+G];
	  // tmp2 = tmp1 / 65280;
	  // tmp1 >>= 8;
	  // out[i+G] = tmp1 + tmp2;
	  tmp1 >>= 8;
	  out[i+G] = tmp1;
	  
	  tmp1 = front[i+B];
	  tmp1 <<= 8;
	  tmp1 = (tmp1 - front[i+B]) + t * back[i+B];
	  // tmp2 = tmp1 / 65280;
	  // tmp1 >>= 8;
	  // out[i+B] = tmp1 + tmp2;
	  tmp1 >>= 8;
	  out[i+B] = tmp1;
	  
	  tmp1 = front[i+A];
	  tmp1 <<= 8;
	  tmp1 = (tmp1 - front[i+A]) + t * back[i+A];
	  // tmp2 = tmp1 / 65280;
	  // tmp1 >>= 8;
	  // out[i+A] = tmp1 + tmp2;	  
	  tmp1 >>= 8;
	  out[i+A] = tmp1;
	}
#endif
    }
  
#ifdef DEBUG_SAVE_PPM
  {    
    unsigned int ii;
    for ( ii = 0; ii < startArea; ++ii )
      {
	this->FrameBuffer[ii] = 0xffffffff;
      }
    for ( ii = startArea + area; ii < this->Width * this->Height; ++ii )
      {
	this->FrameBuffer[ii] = 0xffffffff;
      }
  }
#endif
  
#endif

#ifdef MPE_LOGGING
  MPE_Log_event(MPE_EVENT_COMPOSE_E, 0, "composed");
#endif
}


extern void drawBackWidgets();

void SortLast::CompositeBackWidgets(unsigned int start, unsigned int length)
{
  //get result of compositing workers
  uint32_t *frontBuffer;
  frontBuffer = this->FrameBuffer;
  
  //render only widgets behind everything else
  drawBackWidgets();
  
  //read the pixels back
  glReadBuffer(GL_BACK);
  
  glPixelStorei(GL_UNPACK_ALIGNMENT, 8);
  
  glReadPixels(0, 0, this->Width, this->Height, 
	       GL_RGBA, GL_UNSIGNED_INT_8_8_8_8, WBehindFrameBuffer);


  glClear(GL_COLOR_BUFFER_BIT);

  // composite the behind widget pixels in my region in
  unsigned char *out = (unsigned char *)(this->FrameBuffer + start);
  unsigned char *back  = (unsigned char *)(WBehindFrameBuffer + start);
  unsigned char *front = (unsigned char *)(frontBuffer + start);
  
  long i;
  unsigned int tmp1, tmp2;
  unsigned char t;

  for ( i = 0; i < 4 * length; i += 4 )
    { 
      if ( front[i+A] == 0 )
	{
	  out[i+R] = front[i+R] + back[i+R];
	  out[i+G] = front[i+G] + back[i+G];
	  out[i+B] = front[i+B] + back[i+B];
	  out[i+A] = back[i+A];
	}
      else
	{
	  t = 255 - front[i+A];
	  out[i+R] = (255 * front[i+R] + t * back[i+R]) / 255;
	  out[i+G] = (255 * front[i+G] + t * back[i+G]) / 255;
	  out[i+B] = (255 * front[i+B] + t * back[i+B]) / 255;
	  out[i+A] = (255 * front[i+A] + t * back[i+A]) / 255;	  
	}
    }  

}


//____________________________________________________________________________
//
// This is the end, my friend, the end.
//____________________________________________________________________________
