/*
 * THC: Tools and Headers for C++
 * Copyright (C) 2002
 */

#ifndef __THC_MONITOR_TIMER_H__
#define __THC_MONITOR_TIMER_H__

#if defined (sgi) || defined(__sgi)
#include <invent.h>
#include <time.h>
#elif defined (linux)
#else
#include <time.h>
#endif

//____________________________________________________________________________

/**
 * A simple timer class.
 */

/** Monitor_t is nanoseconds (1E-9). */
typedef unsigned long long Timer_t;

class Timer
{
 public:
  static Timer_t GetRealNsec(void);

  static Timer_t DeltaMsec(Timer_t b, Timer_t e);
  static Timer_t DeltaUsec(Timer_t b, Timer_t e);
  static Timer_t DeltaNsec(Timer_t b, Timer_t e);
  
 private:  
  static bool InitMe(void);
#if defined (sgi) || defined(__sgi)
  static int scan_cpu_info(inventory_t *item, void */*foo*/);
#endif
  static float MHz;
  static bool Init;
};

//____________________________________________________________________________

inline Timer_t Timer::GetRealNsec(void)
{
  Timer_t retval;

#if defined (sgi) || defined(__sgi)
  // SGI specific code, using free running hardware counter
  timespec_t t;
  clock_gettime(CLOCK_SGI_CYCLE, &t);
  retval = ((Timer_t)t.tv_sec * (Timer_t)1000000000) + (Timer_t)(t.tv_nsec);
#elif defined (linux)
  // Linux specific code
  unsigned long long t;
  __asm__ __volatile__("rdtsc" : "=A" (t) : /* no inputs */);
  t *= (unsigned long long)1000;
  retval = t / (unsigned long long)Timer::MHz;
#else
  // General case, takes time and not precise
  timespec_t t;
  clock_gettime(CLOCK_REALTIME, &t);
  retval = ((Timer_t)t.tv_sec * (Timer_t)1000000000) +	(Timer_t)(t.tv_nsec);
#endif
  
  return(retval);  
}

inline Timer_t Timer::DeltaMsec(Timer_t b, Timer_t e) 
{
  return (e - b) / 1000000;
}  

inline Timer_t Timer::DeltaUsec(Timer_t b, Timer_t e) 
{
  return (e - b) / 1000;
}  

inline Timer_t Timer::DeltaNsec(Timer_t b, Timer_t e) 
{
  return (e - b);
}  

#endif
