#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include "VectorMath.h"

#define MAX(x,y) (((x)>(y)) ? (x) : (y))
enum
  {
    _DOUBLE,//8byte
    _FLOAT, //4byte
    _INT,   //4byte
    _UINT,     //4byte
    _SHORT, //2byte
    _USHORT,//2byte
    _UCHAR, //1byte
  };

int dims;
int xiSize, yiSize, ziSize;
int nelts;
float       xSpc, ySpc, zSpc;
int           dataType;
float         xfSize, yfSize, zfSize;
int           numSubVols; 
unsigned char *currentData;
unsigned char *currentGrad;


//============================================================ read Nrrd
//======================================================================
int parseNrrd(char *line);
unsigned int readNrrd(int debug, const char* nrrdFile);
void normalsVGH(int blurnorms, int only1st);
void padOut();
int isPow2(int sz);
int getPow2(int sz);
void pow2Celing(unsigned char *&data0,   //there must be a better way!!!!
		unsigned char *&grad,
		const unsigned char *d0,
		const unsigned char *g,
		int &sx, int &sy, int &sz);
void brick(int maxsz);
unsigned int writeAll(char *filename, int nelts);

class Volume{
 public:
  Volume();
  Volume(Volume &v); //copy constructor does not copy data!
  ~Volume();

  unsigned int writeVol(FILE *trex, char *filename, int nelts);

  int		subVolNum;
  
  int		xiSize, yiSize, ziSize;
  int	      xiVSize, yiVSize, ziVSize; // size of the original volume (before padding etc)
  float		xfSize, yfSize, zfSize;
  float		xfVSize, yfVSize, zfVSize; // size of the original volume (before padding etc)

  int		xiPos, yiPos, ziPos;
  float		xfPos, yfPos, zfPos;
  
  unsigned char *dbData[2];   //points to two blocks of data
  //unsigned char *dbData1[2];	 //two blocks for gradient magnitude (secondary data)
  //unsigned char *dbData2[2];	 //two blocks for gradient magnitude (secondary data)
  //unsigned char *dbData3[2];	 //two blocks for gradient magnitude (secondary data)
  unsigned char *dbGrad[2];   // two blocks for gradients (Vectors)
  unsigned char *currentData; //points to the right block of data
  //unsigned char *currentData1; //points to the right gradient
  //unsigned char *currentData2; //points to the right gradient
  //unsigned char *currentData3; //points to the right gradient
  unsigned char *currentGrad;

  unsigned char **oldData;       //ring buffer for cached data
  unsigned char **oldGrad;       //ring buffer for cached gradients
  int	       *oldTimeStep;
  int		tstepCache;
  int		oldTSpos;

  int		extradivs;

  void *nativeData;
 
  int timestep;

 private:
	
};

Volume::Volume()
{
  xiSize = INT_MAX;
  yiSize = INT_MAX;
  ziSize = INT_MAX;
  xiVSize = INT_MAX;
  yiVSize = INT_MAX;
  ziVSize = INT_MAX;
  xfSize = 0.0;
  yfSize = 0.0;
  zfSize = 0.0;
  xfVSize = 0.0;
  yfVSize = 0.0;
  zfVSize = 0.0;
  xiPos = 0;
  yiPos = 0;
  ziPos = 0;
  xfPos = 0.0;
  yfPos = 0.0;
  zfPos = 0.0;
  subVolNum = 0;
  dbData[0] = 0;
  dbData[1] = 0;
  dbGrad[0] = 0;
  dbGrad[1] = 0;
  currentData = 0;
  currentGrad = 0;
  extradivs = 0;
  oldData = 0;
  oldTimeStep = 0;
  tstepCache = 0;
}

Volume::Volume(Volume &v)
{
  xiSize = v.xiSize;
  yiSize = v.yiSize;
  ziSize = v.ziSize;
  xiVSize = v.xiVSize;
  yiVSize = v.yiVSize;
  ziVSize = v.ziVSize;
  xfSize = v.xfSize;
  yfSize = v.yfSize;
  zfSize = v.zfSize;
  xfVSize = v.xfVSize;
  yfVSize = v.yfVSize;
  zfVSize = v.zfVSize;
  xiPos = v.xiPos;
  yiPos = v.yiPos;
  ziPos = v.ziPos;
  xfPos = v.xfPos;
  yfPos = v.yfPos;
  zfPos = v.zfPos;
  subVolNum = v.subVolNum;
  dbData[0] = 0;
  dbData[1] = 0;
  dbGrad[0] = 0;
  dbGrad[1] = 0;
  currentData = 0;
  currentGrad = 0;
  extradivs = 0;
  oldData = 0;
  oldTimeStep = 0;
  tstepCache = v.tstepCache;
}

Volume::~Volume()
{
  if(dbData[0]) delete dbData[0];
  if(dbData[1]) delete dbData[1];
}

unsigned int Volume::writeVol(FILE *trex, char *filename,int nelts)
{
  fprintf(trex, "SubVolume {\n");
  fprintf(trex, "\tSize int:	  %d, %d, %d\n", xiSize, yiSize, ziSize);
  fprintf(trex, "\tSize float:	  %f, %f, %f\n", xfSize, yfSize, zfSize);
  fprintf(trex, "\tPos int:	  %d, %d, %d\n", xiPos, yiPos, ziPos);
  fprintf(trex, "\tPos float:	  %f, %f, %f\n", xfPos, yfPos, zfPos);
  fprintf(trex, "}\n");

  FILE *f;
  if(!(f = fopen(filename, "wb"))){
    cerr << "Error: MetaVolume::writeAll, failed to open " 
	 << filename << " for writing" << endl;
    return 0;
  }

  unsigned int n;

  n= fwrite((void *) currentData, sizeof(unsigned char), xiSize*yiSize*ziSize*nelts, f);
  if( n != (unsigned int)(xiSize*yiSize*ziSize*nelts)){
    cerr << "Error: MetaVolume:writeVol, write failed " << n << " of " << xiSize*yiSize*ziSize*nelts << endl;
    return n;
  }

  //hartner
  //write out scaled and biased normals
  //new unsigned char[sx*sy*sz*3]
  n= fwrite((void *) currentGrad, sizeof(unsigned char), xiSize*yiSize*ziSize*3, f);
  if( n != (unsigned int)(xiSize*yiSize*ziSize*3)){
    cerr << "Error: MetaVolume:writeVol, write scaled and biased grad failed " << n << " of " << xiSize*yiSize*ziSize*3 << endl;
    return n;
  }
	
	
  //n= fwrite((void *) currentGrad, sizeof(unsigned char), xiSize*yiSize*ziSize, f);
  //if( n != (unsigned int)(xiSize*yiSize*ziSize)){
  //	cerr << "Error: MetaVolume:writeVol-Grad, write failed " << n << " of " << xiSize*yiSize*ziSize << endl;
  //	return n;
  //}

  cerr << "   " << filename << " successfully written" << endl;

  fclose(f);

  return n;
}


Volume       *volumes;                //all sub volumes are kept here


int main(){
  //call genVGH to make V, G, and H and put it in an array
  readNrrd(1,"foo-vgh.nrrd");

  //call blur3D on the float GxGyGz data and then stuff it back into an array of chars for the NxNyNz volume
  int blurnorms = 0;
  normalsVGH(blurnorms,1);


  //copy VGH and NxNyNz volume into a larger volume that is a power of 2 size for the GPU
  padOut();
  int bricksz = 128 * 128 * 256; //unused

  int worldSize = 32;
  int dataSize = xiSize * yiSize * ziSize;

  bricksz = dataSize/worldSize;
  
  brick(bricksz); //bricks into subvolumes
  return 1;
}

unsigned int readNrrd(int debug, const char* nrrdFile)
{
  FILE *f;
  if(!(f = fopen(nrrdFile, "rb"))){
    cerr << "Error: Reader::readNrrd, failed to open " 
	 << nrrdFile << " for reading" << endl;
    return 0;
  }
  char line[256];
  if(debug)
    cerr << "Nrrd info:" << endl;

  int ret = 0;
  while(strlen(fgets(line,256,f)) > 1){
    if(debug)
      cerr << line;
    ret += parseNrrd(line);
  }

  if(ret != 0) return -1;

  size_t n;

  //check to see if we have unsigned chars
  if(dataType == _UCHAR){
    //read it
    unsigned char *data = new unsigned char[xiSize*yiSize*ziSize*nelts];
    n = fread((void *)data, 
	      sizeof(unsigned char), 
	      xiSize * yiSize * ziSize * nelts, f);
    //sanity check
    if(n != (unsigned int)(xiSize * yiSize * ziSize * nelts)){
      cerr << "Error: Reader::readNrrd, UC read failed\n";
      cerr << "	 n = " << n << endl;
      return 0; //error
    }
    //close it to force a flush
    fclose(f);
    cerr << "read : " << n << " bytes " << endl;

    xfSize = xSpc * xiSize;
    yfSize = ySpc * yiSize;
    zfSize = zSpc * ziSize;
    float max = MAX(xfSize, yfSize);
    max = MAX(max,zfSize);
    xfSize /=max;
    yfSize /=max;
    zfSize /=max;
	   
    numSubVols = 1;
    //volumes = new Volume[1];
    currentData = data;
    //volumes[0].xiSize = xiSize;
    //volumes[0].yiSize = yiSize;
    //volumes[0].ziSize = ziSize;
    //volumes[0].xfSize = xfSize;
    //volumes[0].yfSize = yfSize;
    //volumes[0].zfSize = zfSize;
	   
    return n;
  }
	
  return 0;
}


//=========================================================== parse NRRD
//======================================================================

int parseNrrd(char *line)
{
  char *tmp;
  tmp = strtok(line, ":");
  if(!strcmp(tmp, "dimension")){
    tmp = strtok(NULL, "\n");
    dims = atoi(tmp);
  }
  else if(!strcmp(tmp, "sizes")){
    tmp = strtok(NULL, "\n");
    if(dims == 3){
      sscanf(tmp,"%d %d %d", &xiSize, &yiSize, &ziSize);
    }
    else if(dims = 4){
      sscanf(tmp, "%d %d %d %d", &nelts, &xiSize, &yiSize, &ziSize);
    } else {
      cerr << "Error parsing nrrd file: incorrect dimension: " << dims << endl;
      return 1;
    }
  }
  else if(!strcmp(tmp, "spacings")){
    tmp = strtok(NULL, "\n");
    if(dims == 3){
      sscanf(tmp,"%f %f %f", &xSpc, &ySpc, &zSpc);
    }
    else if(dims == 4){
      tmp = strtok(tmp, " ");
      tmp = strtok(NULL, "\n");
      sscanf(tmp, "%f %f %f", &xSpc, &ySpc, &zSpc);
    } else {
      cerr << "Error parsing nrrd file: incorrect dimension: " << dims << endl;
      return 1;
    }
  }

  else if(!strcmp(tmp, "type")){
    tmp = strtok(NULL, "\n");
    if(!strcmp(tmp, "unsigned char") || !strcmp(tmp, " unsigned char")){
      dataType = _UCHAR;
    }
    else if(!strcmp(tmp, "unsigned short") || !strcmp(tmp, " unsigned short")){
      dataType = _USHORT;
    }
    else{
      cerr << "Error parsing nrrd file: only unsigned char/short data type supported" << endl;
      return 1;
    }
  }
  return 0;
}


void normalsVGH(int blurnorms, int only1st)
{
  if(only1st){ //VGH volume???
    int sx = xiSize;
    int sy = yiSize;
    int sz = ziSize;
    currentGrad = new unsigned char[sx*sy*sz*3];
    float *fgrad = new float[sx*sy*sz*3];	   //temp float grad

    //created GxGyGz volume from scratch
    derivative3DVGH(fgrad,sx,sy,sz,currentData);

    //runs some blurring operation on them
    if(blurnorms){
      cerr << "  bluring normals";
      blurV3D(fgrad,(float)1,(float).3,(float).2,(float).1,sx,sy,sz);
      cerr << " -done" << endl;
    }

    //stuffs them into chars
    scalebiasN(currentGrad, fgrad,sx,sy,sz);
    delete[] fgrad;
  }
}


void padOut()
{
  //for(int i =0; i<numSubVols; ++i){
  int sx = xiSize; 
  int sy = yiSize; 
  int sz = ziSize;
  if(!isPow2(sx) || !isPow2(sy) || !isPow2(sz)){
    cerr << "warning!!!! dataset is not a power of two" << endl;
    cerr << " padding out..." << endl;

    unsigned char *newdata0 = 0;
    unsigned char *newgrad	= 0;

    pow2Celing(newdata0,
	       newgrad,
	       currentData, 
	       currentGrad,
	       sx,sy,sz);
			
    xfSize *= (float)sx/(float)xiSize;
    yfSize *= (float)sy/(float)yiSize;
    zfSize *= (float)sz/(float)ziSize;
    //hartner BUG FIXME, xfSize etc above was calculated using
    //	  the non-pow2 xiSize, yiSize, ziSize
    //	  the above statements should come AFTER we we
    //	  update the xiSize here below
    xiSize = sx;
    yiSize = sy;
    ziSize = sz;
    //hartner BUG? why is it .dbData[i], shouldn't we just
    // be iterating on volumes[i]
    delete currentData;
    delete currentGrad;
    currentData	= newdata0;
    currentGrad	=  newgrad;
    //volumes[i].xiSize = xiSize;
    //volumes[i].yiSize = yiSize;
    //volumes[i].ziSize = ziSize;
    //volumes[i].xfSize = xfSize;
    //volumes[i].yfSize = yfSize;
    //volumes[i].zfSize = zfSize;

    //gluvv.mv->wholeVol->currentData = newdata0;
  }
  //}
}

int isPow2(int sz)
{
  double tmp = sz/2.0;
  while(tmp > 1.0){
    tmp = tmp/2.0;
  }
  if(tmp != 1.0) return 0;
  else return 1;
}

int getPow2(int sz)
{
  double tmp = sz/2.0;
  int i = 2;
  while(tmp > 1.0){
    i *= 2;
    tmp = tmp/2.0;
  }
  return i;
}

void pow2Celing(unsigned char *&data0,   //there must be a better way!!!!
		unsigned char *&grad,
		const unsigned char *d0,
		const unsigned char *g,
		int &sx, int &sy, int &sz)
{
  int nx = getPow2(sx);
  int ny = getPow2(sy);
  int nz = getPow2(sz);
  cerr << "new sizes = " << nx << " " << ny << " " << nz <<endl;
  cerr << "old sizes = " << sx << " " << sy << " " << sz << endl;
	
  data0 = new unsigned char[nx*ny*nz*nelts];
  if(g)  grad  = new unsigned char[nx*ny*nz*3];

  for(int i=0; i<nz; ++i){
    for(int j=0; j<ny; ++j)
      for(int k=0; k<nx; ++k)
	for(int l=0; l<nelts; ++l)
	  data0[i*ny*nx*nelts + j*nx*nelts + k*nelts + l] = 0;
  }

  if(g){
    for(int i=0; i<nz; ++i)
      for(int j=0; j<ny; ++j)
	for(int k=0; k<nx; ++k)
	  for(int l=0; l<3; ++l)
	    grad[i*ny*nx*3 + j*nx*3 + k*3 + l] = 0;
  }

  for(int i=0; i<sz; ++i){
    for(int j=0; j<sy; ++j)
      for(int k=0; k<sx; ++k)
	for(int l=0; l<nelts; ++l)
	  data0[i*ny*nx*nelts + j*nx*nelts + k*nelts + l] = 
	    d0[i*sy*sx*nelts + j*sx*nelts + k*nelts + l];
  }

  if(g){
    for(int i=0; i<sz; ++i)
      for(int j=0; j<sy; ++j)
	for(int k=0; k<sx; ++k)
	  for(int l=0; l<3; ++l)
	    grad[i*ny*nx*3 + j*nx*3 + k*3 + l] = 
	      g[i*sy*sx*3 + j*sx*3 + k*3 + l];	
  }


  sx = nx;
  sy = ny;
  sz = nz;

}


void brick(int maxsz)
{
  if((xiSize*yiSize*ziSize) > maxsz ){
    if( numSubVols == 1 ){ //we have to rebrick
      int xd = 1;
      int yd = 1;
      int zd = 1;
      int sx = xiSize;
      int sy = yiSize;
      int sz = ziSize;
      while(((sx/(float)xd)*(sy/(float)yd)*(sz/(float)zd)) > maxsz){
	if(zd > yd){
	  if(yd > xd)
	    xd*=2;
	  else
	    yd*=2;
	} else
	  zd*=2;
      }
      cerr << "NEW brick dims " << sx/(float)xd << " " << sy/(float)yd << " " << sz/(float)zd << endl;


      int wait = 0;
      int wait2=0;

      while (wait){
	wait2 = 1;
      }
      
      
      //Volume *tempv = &volumes[0];
      //wholeVol = tempv;
      unsigned char *od = currentData; //old data
      volumes = new Volume[xd*yd*zd];	//create new volumes
      numSubVols = xd*yd*zd;
      int nsx = (int)(sx/(float)xd);	//new sizes
      int nsy = (int)(sy/(float)yd);
      int nsz = (int)(sz/(float)zd);
      cerr << "	 Creating sub-volume : ";
      for(int i=0; i<zd; ++i){
	for(int j=0; j<yd; ++j){
	  for(int k=0; k<xd; ++k){
	    int cv = i*yd*xd + j*xd + k; //current (sub)volume
	    cerr << cv << " ";
	    volumes[cv].currentData = volumes[cv].dbData[0] = new unsigned char[nsx*nsy*nsz*nelts];
	    if(currentGrad) 
	      volumes[cv].currentGrad = volumes[cv].dbGrad[0] = new unsigned char[nsx*nsy*nsz*3];
	    volumes[cv].xiSize = nsx;
	    volumes[cv].yiSize = nsy;
	    volumes[cv].ziSize = nsz;
	    volumes[cv].xfSize = xfSize*(nsx/(float)sx);
	    volumes[cv].yfSize = yfSize*(nsy/(float)sy);
	    volumes[cv].zfSize = zfSize*(nsz/(float)sz);
	    volumes[cv].xfPos  = volumes[cv].xfSize*k;
	    volumes[cv].yfPos  = volumes[cv].yfSize*j;
	    volumes[cv].zfPos  = volumes[cv].zfSize*i;
	    volumes[cv].xiPos  = volumes[cv].xiSize*k;
	    volumes[cv].yiPos  = volumes[cv].yiSize*j;
	    volumes[cv].ziPos  = volumes[cv].ziSize*i;
	    
	    for(int z=0; z<nsz; ++z){
	      for(int y=0; y<nsy; ++y){
		for(int x=0; x<nsx; ++x){
		  for(int e=0; e<nelts; ++e){
		    volumes[cv].currentData[z*nsy*nsx*nelts + y*nsx*nelts + x*nelts + e] = 
		      currentData[(i*nsz + z)*sx*sy*nelts + (j*nsy + y)*sx*nelts + (k*nsx + x)*nelts + e];
		  }//elts
		}//x
	      }//y
	    }//z----------------------
	    
	    if(currentGrad){
	      for(int z=0; z<nsz; ++z){
		for(int y=0; y<nsy; ++y){
		  for(int x=0; x<nsx; ++x){
		    for(int e=0; e<3; ++e){
		      volumes[cv].currentGrad[z*nsy*nsx*3 + y*nsx*3 + x*3 + e] = 
			currentGrad[(i*nsz + z)*sx*sy*3 + (j*nsy + y)*sx*3 + (k*nsx + x)*3 + e];
		    }//elts
		  }//x
		}//y
	      }//z----------------------
	    }
	    
	    
	  }//xd
	}//yd
      }//zd--------------------
      cerr << endl;
      //append = 1;

    } //end if we had to rebrick if( numSubVols == 1 ){
    else{
      cerr << "Warning::Re-bricking is needed!!!, auto-re-bricking is not supported...YET" << endl; 
    }
  } //if((xiSize*yiSize*ziSize) > maxsz ){
  else {
    cerr << "bricking not done because it is not needed" << endl; 
  }
  //append = 1;
  writeAll("hartner_bricks", nelts );
}

unsigned int writeAll(char *filename, int nelts)
{
  char	subfile[512];
  sprintf(subfile, "%s.trex", filename);
  FILE *f;
  if(!(f = fopen(subfile, "w"))){
    cerr << "Error: writeAll, failed to open " 
	 << subfile << " for writing" << endl;
    return 0;
  }
  fprintf(f, "#  Meta Volume  #\n");
  fprintf(f, "\n\n#    Global info\n");
  fprintf(f, "Data Set Name:	     %s\n", filename);
  fprintf(f, "#  NOTE: Data Set Files uses an implicit extension: filename.timestep.subvol\n");
  fprintf(f, "Data Set Files:	   %s\n", filename);
  //change this for time varying
  fprintf(f, "Number of Time Steps:  1, 0, 0\n");
  fprintf(f, "Volume Size int:	   %d, %d, %d\n", xiSize, yiSize, ziSize);
  fprintf(f, "Volume Size float:	   %f, %f, %f\n", xfSize, yfSize, zfSize);
  fprintf(f, "\n\n#    Subvolume info\n");
  fprintf(f, "Number of Sub Volumes: %d\n", numSubVols);
	
  unsigned int n = 0;

  for(int i =0; i<numSubVols; ++i){
    char subvolfile[512];
    //if(append)	//change this for time varying
    sprintf(subvolfile, "%s.%04d.%02d", filename, 0, i);
    //else		
    //sprintf(subvolfile, "%s", filename);
    cerr << "writing " << filename << " sub-volume " << i << "  as " << subvolfile << endl;
    n += volumes[i].writeVol(f, subvolfile,nelts);
  }

  fclose(f);

  return n;
}
