//------------------------------------------------------------------------
// gluvvui.cpp
//   Joe Kniss
//	a user interface for gluvv state
//
//		     ________	 ____	___ 
//		    |	     \	/    | /  /
//		    +---+     \/     |/	 /
//		    +--+|  |\	 /|	< 
//		    |  ||  | \	/ |  |\	 \ 
//		    |	   |  \/  |  | \  \ 
//		     \_____|	  |__|	\__\
//			 Copyright  2001 
//			Joe Michael Kniss
//		     <<< jmk@cs.utah.edu >>>
//		 "All Your Base are Belong to Us"
//-------------------------------------------------------------------------
//////////////////////////////////////////////////////////////////////

#include "gluvvui.h"
#ifdef WIN32
#include "glui/glui.h"
#else
#include <glui.h>
#endif
#include "gluvv.h"
#include "VectorMath.h"
#include <iostream>
using namespace std;

#include "glutredirect.h"

#define __VERSION_2 1
#define __VERSION_3 1

//======================================== Static ui callback
//===========================================================
void uicb(int control){
  gui.control_cb(control);
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

gluvvui::gluvvui()
{
  on = 0;
}

gluvvui::~gluvvui()
{

}

//================================================= INIT !!!!
//===========================================================
int gluvvui::init()
{
  glui_window = GLUI_Master.create_glui("Simian~ SpaceMonkey");

  /*** Add invisible panel to hold rest of controls ***/
  GLUI_Panel *panel1 = glui_window->add_panel( "", GLUI_PANEL_NONE );

  //close the menu gui
  glui_window->add_button_to_panel(panel1, ">>>>>>>> Close <<<<<<<<", _close, uicb);

  //roll out for window level --------------------------------------------------------
  GLUI_Rollout *roll0 = glui_window->add_rollout_to_panel(panel1, "Classification", true);
  {

    /// name
    TriName = 
      glui_window->add_edittext_to_panel(roll0, "Feature Name:",GLUI_EDITTEXT_TEXT, NULL, _triname, uicb);
    TriName->set_text("None Selected");
    TriName->set_w( 300 );

    /// window / level settings
    GLUI_Panel *winlev = glui_window->add_panel_to_panel(roll0,"Level & Window Settings",GLUI_PANEL_EMBOSSED);

    //level
    Levsp = 
      glui_window->add_spinner_to_panel(winlev, "Level", 
					GLUI_SPINNER_FLOAT, 
					NULL, _levsp, uicb);
    Levsp->set_float_limits(0, 1, GLUI_LIMIT_CLAMP);
    Levsp->set_float_val(1);

		
    glui_window->add_column_to_panel(winlev,false);

    // window
    Winsp = 
      glui_window->add_spinner_to_panel(winlev, "Window", 
					GLUI_SPINNER_FLOAT, 
					NULL, _winsp, uicb);
    Winsp->set_float_limits(0, 1, GLUI_LIMIT_CLAMP);
    Winsp->set_float_val(1);


    /// surface settings
    GLUI_Panel *surfpan = glui_window->add_panel_to_panel(roll0,"Surface Enhancement",GLUI_PANEL_EMBOSSED);

    // surface
    Surfsp = 
      glui_window->add_spinner_to_panel(surfpan, "Min", 
					GLUI_SPINNER_FLOAT, 
					NULL, _surfsp, uicb);
    Surfsp->set_float_limits(0, 1, GLUI_LIMIT_CLAMP);
    Surfsp->set_float_val(0);

    glui_window->add_column_to_panel(surfpan,false);

    // surface
    MaxSurfsp = 
      glui_window->add_spinner_to_panel(surfpan, "Max", 
					GLUI_SPINNER_FLOAT, 
					NULL, _maxsurfsp, uicb);
    MaxSurfsp->set_float_limits(0, 1, GLUI_LIMIT_CLAMP);
    MaxSurfsp->set_float_val(1);


    /// color - alpha settings
    GLUI_Panel *color = glui_window->add_panel_to_panel(roll0,"",GLUI_PANEL_NONE);

    Alphasp = 
      glui_window->add_spinner_to_panel(color, "Opacity", 
					GLUI_SPINNER_FLOAT, 
					NULL, _alphasp, uicb);
    Alphasp->set_float_limits(0, 1, GLUI_LIMIT_CLAMP);
    Alphasp->set_float_val(1);

    glui_window->add_column_to_panel(color,false);

    Huesp = 
      glui_window->add_spinner_to_panel(color, "Hue", 
					GLUI_SPINNER_FLOAT, 
					NULL, _huesp, uicb);
    Huesp->set_float_limits(0, 1, GLUI_LIMIT_CLAMP);
    Huesp->set_float_val(1);

    /// kinds
    GLUI_Panel *kinds = glui_window->add_panel_to_panel(roll0,"",GLUI_PANEL_NONE);

    //radio group for types
    TriType = glui_window->add_radiogroup_to_panel(kinds,NULL, _tritype, uicb);
    glui_window->add_radiobutton_to_group(TriType, "Basic Win-Lev");
    glui_window->add_radiobutton_to_group(TriType, "Iso-value");
    glui_window->add_radiobutton_to_group(TriType, "Surface Iso-value");
    TriType->set_int_val(0);

    glui_window->add_column_to_panel(kinds,true);

    TriHilight = glui_window->add_checkbox_to_panel(kinds,"Hi-Light",NULL,_trihilight,uicb);
    TriHilight->set_int_val(1);

    TriHide = glui_window->add_checkbox_to_panel(kinds,"Hide",NULL,_trihide,uicb);
    TriHide->set_int_val(0);

    /// next - previous
    GLUI_Panel *trinext = glui_window->add_panel_to_panel(roll0,"",GLUI_PANEL_NONE);
    glui_window->add_button_to_panel(trinext, "<<<< Prev", _prevtri, uicb);
    glui_window->add_column_to_panel(trinext,true);
    glui_window->add_button_to_panel(trinext, "Next >>>>", _nexttri, uicb);

    /// new - delete
    GLUI_Panel *trinewdel = glui_window->add_panel_to_panel(roll0,"",GLUI_PANEL_NONE);
    glui_window->add_button_to_panel(trinewdel, " New ", _newtri, uicb);
    glui_window->add_column_to_panel(trinewdel,true);
    glui_window->add_button_to_panel(trinewdel, " Delete ", _deltri, uicb);
  }

  //roll out for dual domain interaction --------------------------------------------------------
  GLUI_Rollout *roll1 = glui_window->add_rollout_to_panel(panel1, "Dual Domain Interaction", false);
  {
    //radio group for brushes
    ddBrushes = glui_window->add_radiogroup_to_panel(roll1,(int*)&gluvv.probe.brush, _brushgrp, uicb);
    glui_window->add_radiobutton_to_group(ddBrushes, "OFF (k)");
    //glui_window->add_radiobutton_to_group(ddBrushes, "Ellipse (b)");
    glui_window->add_radiobutton_to_group(ddBrushes, "Auto Ellipse (a)");
    glui_window->add_radiobutton_to_group(ddBrushes, "Triangle (t)");
    //glui_window->add_radiobutton_to_group(ddBrushes, "1D (O)");
    //glui_window->add_radiobutton_to_group(ddBrushes, "Auto 1D (o)");
    //column
    glui_window->add_column_to_panel(roll1, true);
    //buttons for clear and paint
    glui_window->add_button_to_panel(roll1, "Paint (spc)", _ddpaint, uicb);
    glui_window->add_button_to_panel(roll1, "Drop (d)", _dddrop, uicb);
    glui_window->add_button_to_panel(roll1, "Clear (n)", _ddclear, uicb);
    //spinner for brush size
    ddSizeSpin = 
      glui_window->add_spinner_to_panel(roll1, "Size", 
					GLUI_SPINNER_FLOAT, 
					NULL, _ddsize, uicb);
    ddSizeSpin->set_float_limits(0, 1, GLUI_LIMIT_CLAMP);
    ddSizeSpin->set_float_val(gluvv.probe.slider);
  }//end dual domain -----------------------------------------------------------------------------

  //roll out for shading ------------------------------------------------------------------------
  GLUI_Rollout *roll2 = glui_window->add_rollout_to_panel(panel1, "Shading", false); 
  {
    //radio group for shading
    Shaders = glui_window->add_radiogroup_to_panel(roll2,(int*)&gluvv.probe.brush, _shadegrp, uicb);
    glui_window->add_radiobutton_to_group(Shaders, "Emmisive (F9)");
    glui_window->add_radiobutton_to_group(Shaders, "Shaded (F10)");
    glui_window->add_radiobutton_to_group(Shaders, "Faux (F11)");
    Shaders->set_int_val(2);

#ifdef DO_EXPERIMENTAL_FEATURES
    // ambiant contribution
    ambientSpin = 
      glui_window->add_spinner_to_panel(roll2, "Ambient", 
					GLUI_SPINNER_FLOAT, 
					NULL, _ambient, uicb);
    ambientSpin->set_float_limits(0, 1, GLUI_LIMIT_CLAMP);
    ambientSpin->set_float_val(gluvv.light.amb);
#endif //DO_EXPERIMENTAL_FEATURES		
    //lighting
    GLUI_Rollout *rollLight = glui_window->add_rollout_to_panel(roll2, "Lights", false);
    {
      //light radius
      lightRadSpin = 
	glui_window->add_spinner_to_panel(rollLight, "Light Radius", 
					  GLUI_SPINNER_FLOAT, 
					  NULL, _lightRad, uicb);
      lightRadSpin->set_float_limits(.1, 1000, GLUI_LIMIT_CLAMP);
      lightRadSpin->set_float_val(normV3(gluvv.light.pos));
      //check box for light
      Light = glui_window->add_checkbox_to_panel(rollLight, "Light Widget", NULL, _lightchk, uicb);

#ifdef DO_EXPERIMENTAL_FEATURES
      LtAtt = glui_window->add_checkbox_to_panel(rollLight, "Light Attenuation", NULL, _ltatt, uicb);

      LtAttThresh = 
	glui_window->add_spinner_to_panel(rollLight, "Light Attenuation", 
					  GLUI_SPINNER_FLOAT, 
					  NULL, _ltattthresh, uicb);
      LtAttThresh->set_float_limits(0, 1, GLUI_LIMIT_CLAMP);
      LtAttThresh->set_float_val(1.0 - gluvv.light.lattThresh);
			
      LtAttNear = 
	glui_window->add_spinner_to_panel(rollLight, "Attinuation Start", 
					  GLUI_SPINNER_FLOAT, 
					  NULL, _ltattnear, uicb);
      LtAttNear->set_float_limits(0, 10, GLUI_LIMIT_CLAMP);
      LtAttNear->set_float_val(gluvv.light.lattLimits[0]*2.0);
			
      LtAttFar = 
	glui_window->add_spinner_to_panel(rollLight, "Attinuation Finish", 
					  GLUI_SPINNER_FLOAT, 
					  NULL, _ltattfar, uicb);
      LtAttFar->set_float_limits(-10, 0, GLUI_LIMIT_CLAMP);
      LtAttFar->set_float_val(gluvv.light.lattLimits[1]*2.0);
#endif //DO_EXPERIMENTAL_FEATURES

    } //end lighting
#if __VERSION_3

#ifdef DO_EXPERIMENTAL_FEATURES
    //shadow qualities
    GLUI_Rollout *rollShad = glui_window->add_rollout_to_panel(roll2, "Shadows", false);
    {	
      //check box for shadows
      Shadow = glui_window->add_checkbox_to_panel(rollShad, "Shadows", NULL, _shadowchk, uicb);
      Shadow->set_int_val(gluvv.light.shadow);
      Soft = glui_window->add_checkbox_to_panel(rollShad, "Soft Shadows", NULL, _softchk, uicb);
      Soft->set_int_val(gluvv.light.softShadow);
      showLT = glui_window->add_checkbox_to_panel(rollShad, "Light View", NULL, _showltchk, uicb);
      showLT->set_int_val(gluvv.light.showView);
			
      GShadowQualSpin = 
	glui_window->add_spinner_to_panel(rollShad, "Good Shadow Quality", 
					  GLUI_SPINNER_FLOAT, 
					  NULL, _gshadowqual, uicb);
      GShadowQualSpin->set_float_limits(.1, 1, GLUI_LIMIT_CLAMP);
      GShadowQualSpin->set_float_val(gluvv.light.gShadowQual);
      IShadowQualSpin = 
	glui_window->add_spinner_to_panel(rollShad, "Interactive Shadow Quality", 
					  GLUI_SPINNER_FLOAT, 
					  NULL, _ishadowqual, uicb);
      IShadowQualSpin->set_float_limits(.1, 1, GLUI_LIMIT_CLAMP);
      IShadowQualSpin->set_float_val(gluvv.light.iShadowQual);
    } //end shadows
#endif //DO_EXPERIMENTAL_FEATURES
#endif

#ifdef DO_EXPERIMENTAL_FEATURES		
    //fog
    GLUI_Rollout *rollFog = glui_window->add_rollout_to_panel(roll2, "Fog", false);
    {
      GLUI_Panel *panelFC = glui_window->add_panel_to_panel(rollFog, "Fog Color");
      FogColors = glui_window->add_radiogroup_to_panel(panelFC, 0, _fogcolgrp, uicb);
      glui_window->add_radiobutton_to_group(FogColors, "Fog Off");
      glui_window->add_radiobutton_to_group(FogColors, "Blue");
      glui_window->add_radiobutton_to_group(FogColors, "White");
      glui_window->add_radiobutton_to_group(FogColors, "Black");
			
      FogThresh = 
	glui_window->add_spinner_to_panel(rollFog, "Fog Intensity", 
					  GLUI_SPINNER_FLOAT, 
					  NULL, _fogthresh, uicb);
      FogThresh->set_float_limits(0, 1, GLUI_LIMIT_CLAMP);
      FogThresh->set_float_val(1.0 - gluvv.light.fogThresh);
			
      FogNear = 
	glui_window->add_spinner_to_panel(rollFog, "Fog Start", 
					  GLUI_SPINNER_FLOAT, 
					  NULL, _fognear, uicb);
      FogNear->set_float_limits(0, 10, GLUI_LIMIT_CLAMP);
      FogNear->set_float_val(gluvv.light.fogLimits[0]*2.0);
			
      FogFar = 
	glui_window->add_spinner_to_panel(rollFog, "Fog Finish", 
					  GLUI_SPINNER_FLOAT, 
					  NULL, _fogfar, uicb);
      FogFar->set_float_limits(-10, 0, GLUI_LIMIT_CLAMP);
      FogFar->set_float_val(gluvv.light.fogLimits[1]*2.0);
    } //end fog
#endif //DO_EXPERIMENTAL_FEATURES		

#ifdef DO_EXPERIMENTAL_FEATURES		
    //silhouette edges
    Sill = glui_window->add_checkbox_to_panel(roll2, "Silhouette Edges", NULL, _sillchk, uicb);
    Sill->set_int_val(gluvv.light.sill);
#endif //DO_EXPERIMENTAL_FEATURES		

  }
  //end shading ----------------------------------------------------------------------------------

#ifdef DO_EXPERIMENTAL_FEATURES		
  //perturbation ---------------------------------------------------------------------------------
  GLUI_Rollout *rollpert = glui_window->add_rollout_to_panel(panel1, "Perturbation", false);
  {
    PertW1 = 
      glui_window->add_spinner_to_panel(rollpert, "Weight 1", 
					GLUI_SPINNER_FLOAT, 
					NULL, _pertW1, uicb);
    PertW1->set_float_limits(0, 1, GLUI_LIMIT_CLAMP);
    PertW1->set_float_val(.2);
      
    PertS1 = 
      glui_window->add_spinner_to_panel(rollpert, "Scale 1", 
					GLUI_SPINNER_FLOAT, 
					NULL, _pertS1, uicb);
    PertS1->set_float_limits(0, 65, GLUI_LIMIT_CLAMP);
    PertS1->set_float_val(.2);

    PertW2 = 
      glui_window->add_spinner_to_panel(rollpert, "Weight 2", 
					GLUI_SPINNER_FLOAT, 
					NULL, _pertW2, uicb);
    PertW2->set_float_limits(0, 1, GLUI_LIMIT_CLAMP);
    PertW2->set_float_val(0);

    PertS2 = 
      glui_window->add_spinner_to_panel(rollpert, "Scale 2", 
					GLUI_SPINNER_FLOAT, 
					NULL, _pertS2, uicb);
    PertS2->set_float_limits(0, 65, GLUI_LIMIT_CLAMP);
    PertS2->set_float_val(2.1);

    PertW3 = 
      glui_window->add_spinner_to_panel(rollpert, "Weight 3", 
					GLUI_SPINNER_FLOAT, 
					NULL, _pertW3, uicb);
    PertW3->set_float_limits(0, 1, GLUI_LIMIT_CLAMP);
    PertW3->set_float_val(0);

    PertS3 = 
      glui_window->add_spinner_to_panel(rollpert, "Scale 3", 
					GLUI_SPINNER_FLOAT, 
					NULL, _pertS3, uicb);
    PertS3->set_float_limits(0, 65, GLUI_LIMIT_CLAMP);
    PertS3->set_float_val(4.5);

    PertW4 = 
      glui_window->add_spinner_to_panel(rollpert, "Weight 4", 
					GLUI_SPINNER_FLOAT, 
					NULL, _pertW4, uicb);
    PertW4->set_float_limits(0, 1, GLUI_LIMIT_CLAMP);
    PertW4->set_float_val(0);

    PertS4 = 
      glui_window->add_spinner_to_panel(rollpert, "Scale 4", 
					GLUI_SPINNER_FLOAT, 
					NULL, _pertS4, uicb);
    PertS4->set_float_limits(0, 1, GLUI_LIMIT_CLAMP);
    PertS4->set_float_val(8.7);
  }

  //end perturbation -----------------------------------------------------------------------------
#endif //DO_EXPERIMENTAL_FEATURES		

  //clipping planes ------------------------------------------------------------------------------
  GLUI_Rollout *roll3 = glui_window->add_rollout_to_panel(panel1, "Clipping Plane", true);
  Clips = glui_window->add_radiogroup_to_panel(roll3,NULL, _clipgrp, uicb);
  glui_window->add_radiobutton_to_group(Clips, "OFF");
  glui_window->add_radiobutton_to_group(Clips, "X positive");
  glui_window->add_radiobutton_to_group(Clips, "X negative");
  glui_window->add_radiobutton_to_group(Clips, "Y positive");
  glui_window->add_radiobutton_to_group(Clips, "Y negative");
  glui_window->add_radiobutton_to_group(Clips, "Z positive");
  glui_window->add_radiobutton_to_group(Clips, "Z negative");

  glui_window->add_column_to_panel(roll3,true);
	
  ClipPos = glui_window->add_spinner_to_panel(roll3, "Position", GLUI_SPINNER_FLOAT, NULL, _clippos, uicb);
  ClipPos->set_float_limits(-2,2, GLUI_LIMIT_CLAMP);
  ClipPos->set_float_val(0);

  ClipAlpha = glui_window->add_spinner_to_panel(roll3, "Opacity", GLUI_SPINNER_FLOAT, NULL, _clipalpha, uicb);
  ClipAlpha->set_float_limits(0,1, GLUI_LIMIT_CLAMP);
  ClipAlpha->set_float_val(0);

  ClipSize = glui_window->add_spinner_to_panel(roll3, "Size", GLUI_SPINNER_FLOAT, NULL, _clipsize, uicb);
  ClipSize->set_float_limits(.2,2, GLUI_LIMIT_CLAMP);
  ClipSize->set_float_val(.2);

  ClipProbe = glui_window->add_checkbox_to_panel(roll3, "Probe Clip Plane", NULL, _clipprobe, uicb);
  ClipProbe->set_int_val(1);
	
  //end clipping --------------------------------------------------------------------------------

  //sample rates --------------------------------------------------------------------------------
  GLUI_Panel *panel2 = glui_window->add_panel_to_panel(panel1, "Sample Rate");
  interactSpin = 
    glui_window->add_spinner_to_panel(panel2, "Interactive", 
				      GLUI_SPINNER_FLOAT, 
				      NULL, _interactSR, uicb);
  interactSpin->set_float_limits(0, 20, GLUI_LIMIT_CLAMP);
  interactSpin->set_float_val(gluvv.volren.interactSamp);

  glui_window->add_column_to_panel(panel2,true);
  
  goodSpin = 
    glui_window->add_spinner_to_panel(panel2, "High Quality", 
				      GLUI_SPINNER_FLOAT, 
				      NULL, _goodSR, uicb);
  goodSpin->set_float_limits(0, 20, GLUI_LIMIT_CLAMP);
  goodSpin->set_float_val(gluvv.volren.goodSamp);
  //end sample rates ---------------------------------------------------------------------------

  /// color - alpha settings
  GLUI_Panel *probe = glui_window->add_panel_to_panel(panel1,"Probe & Position");

  ProbeValue = glui_window->add_spinner_to_panel(probe, "Probe Val", GLUI_SPINNER_FLOAT);
  ProbeValue->set_float_limits(0,100000, GLUI_LIMIT_CLAMP);
  ProbeValue->set_float_val(0);

  ProbeGrad = glui_window->add_spinner_to_panel(probe,  "      Grad", GLUI_SPINNER_FLOAT);
  ProbeGrad->set_float_limits(0,100000, GLUI_LIMIT_CLAMP);
  ProbeGrad->set_float_val(0);

  glui_window->add_column_to_panel(probe,true);

  ProbeX = glui_window->add_spinner_to_panel(probe,  "X", GLUI_SPINNER_FLOAT);
  ProbeX->set_float_limits(0,100000, GLUI_LIMIT_CLAMP);
  ProbeX->set_float_val(0);

  ProbeY = glui_window->add_spinner_to_panel(probe,  "Y", GLUI_SPINNER_FLOAT);
  ProbeY->set_float_limits(0,100000, GLUI_LIMIT_CLAMP);
  ProbeY->set_float_val(0);

  ProbeZ = glui_window->add_spinner_to_panel(probe,  "Z", GLUI_SPINNER_FLOAT);
  ProbeZ->set_float_limits(0,100000, GLUI_LIMIT_CLAMP);
  ProbeZ->set_float_val(0);
	
  //histogram
  Histo = glui_window->add_checkbox_to_panel(panel1, "Histogram", NULL, _histochk, uicb);
  Histo->set_int_val(1);
  
  //back ground color --------------------------------------------------------------------------
  GLUI_Panel *panelBG = glui_window->add_panel_to_panel(panel1, "Background Color");
  bGround = glui_window->add_radiogroup_to_panel(panelBG, 0, _bggrp, uicb);
  glui_window->add_radiobutton_to_group(bGround, "White");
  glui_window->add_radiobutton_to_group(bGround, "Black");
  //end back ground color ----------------------------------------------------------------------

#ifdef DO_EXPERIMENTAL_FEATURES		
  //normal computation buttons -----------------------------------------------------------------
  glui_window->add_button_to_panel(panel1, "Post Classification Normals", _normpostBtn, uicb);
  glui_window->add_button_to_panel(panel1, "Pre Classification Normals", _normpreBtn, uicb);
#endif //DO_EXPERIMENTAL_FEATURES		
  //use normal computation buttons -------------------------------------------------------------

  on = 1;
  return 0;
}

//====================================================== Show
//===========================================================
void gluvvui::show()
{
#ifdef PARALLEL
  if (worldId != 0) return;
#endif

  init();
  on = 1;
}

//====================================================== Hide
//===========================================================
void gluvvui::hide()
{
#ifdef PARALLEL
  if (worldId != 0) return;
#endif

  //glutSetWindow(gluvv.mainWindow);
  glui_window->close();
  //GLUI_Master.close_all();
  on = 0;
  glutSetWindow(gluvv.mainWindow);
}

//=================================================== Visible
//===========================================================
int  gluvvui::visible()
{
  return on;
}

void gluvvui::update()
{
#ifdef PARALLEL
  if (worldId != 0 || !visible()) return;
#endif
  if (!visible()) return;
  if(gluvv.clip.clipWidget)
    {	

      ClipPos->set_float_val( gluvv.clip.clipWidget->getOrthoPos() );
      ClipAlpha->set_float_val( gluvv.clip.clipWidget->getAlpha() );
      ClipSize->set_float_val( gluvv.clip.clipWidget->getSize() );
      ClipProbe->set_int_val( gluvv.clip.clipWidget->getProbe() );
    }
	
  ProbeValue->set_float_val( gluvv.probe.vval[0] * 1000 );
  ProbeGrad->set_float_val( gluvv.probe.vval[1] * 1000 );

  ProbeX->set_float_val( gluvv.clip.mousepnt[0] * gluvv.mv->xiSize * gluvv.mv->xSpc );
  ProbeY->set_float_val( gluvv.clip.mousepnt[1] * gluvv.mv->yiSize * gluvv.mv->ySpc );
  ProbeZ->set_float_val( gluvv.clip.mousepnt[2] * gluvv.mv->ziSize * gluvv.mv->zSpc );

}

//=================================================== updateTF
//===========================================================
void gluvvui::updateTF()
{
  //#ifdef PARALLEL
  //if (worldId != 0) return;
  //#endif
  
  //pop all alpha
  LevWidget *trit = gluvv.tf.tfWidget->getTris();
  while( trit )
    {
      trit->popAlpha();
      trit = trit->getNext();
    }
  gluvv.tf.loadme = 1;

  if(!gluvv.tf.curTri) return;
	
  LevWidget *tri = gluvv.tf.curTri;

  if(!tri) return;

  tri->setPicked(1);
#ifdef PARALLEL
  if (worldId == 0){
#endif
  TriName->set_text(tri->getName());
  //TriName->set_text("not implemented");
  
  Levsp->set_float_val(tri->getLevel());
  Winsp->set_float_val(tri->getWindow());

  Surfsp->set_float_val( tri->getSurfMin() );
  MaxSurfsp->set_float_val( tri->getSurfMax() );

  Alphasp->set_float_val( tri->getAlpha() );
  Huesp->set_float_val( tri->getHue() );

  TriHide->set_int_val( tri->getHide() ) ;

#ifdef PARALLEL
  } //endif (worldId == 0){
#endif
  switch(tri->getType()){
  case LevWidget::LWtriangle:
#ifdef PARALLEL
    if (worldId == 0){
#endif
    TriType->set_int_val(2);
#ifdef PARALLEL
    }
#endif    
    break;
  case LevWidget::LW1d:
    if(tri->getInteraction() == LevWidget::LWStandard)
      {
#ifdef PARALLEL
    if (worldId == 0){
#endif
	TriType->set_int_val(1);
#ifdef PARALLEL
    }
#endif 
      }
    else
      {
#ifdef PARALLEL
    if (worldId == 0){
#endif 
	TriType->set_int_val(0);
 #ifdef PARALLEL
    }
#endif
      }

    break;
  }

  if( /*TriHilight->get_int_val() && */ gluvv.hilight && (tri != gluvv.tf.tfWidget->getTris()) && (!tri->getHide()))
    {
      trit = gluvv.tf.tfWidget->getTris();
      while( trit )
	{
	  if(trit != tri)
	    {
	      trit->pushAlpha( trit->getAlpha() * .1 );
	    }
	  trit = trit->getNext();
	}
			
    }

}

extern void bcast_control(int control, int ival0, float fval0);

//========================================== Control callback
//===========================================================
void gluvvui::control_cb(int control)
{
  if(!on) return;
	
  int val;
	
  int ival0 = 0;
  float fval0 = 0.0;
  char tval0[1024];

  glutSetWindow(gluvv.mainWindow);

  //LevWidget *lwtmp;
  //float b[2],l[2],r[2],t[2];

  switch(control){
  case _close:
    break;


  //////**********************
  case _newtri:
    break;
  case _deltri:
    break;
  case _trihide:
    if(gluvv.tf.curTri)
      ival0 = TriHide->get_int_val();
    break;
  case _trihilight:
    ival0 = TriHilight->get_int_val();
    break;
  case _nexttri:
    break;
  case _prevtri:
    break;
  case _triname:
#ifdef PARALLEL
    if(worldId == 0){
#endif
      if(gluvv.tf.curTri){
	gluvv.tf.curTri->setName(TriName->get_text());
      }
#ifdef PARALLEL
    }
#endif
    break;
  case _winsp:
    if(gluvv.tf.curTri)
      fval0 = Winsp->get_float_val();
    break;
  case _levsp:
    if(gluvv.tf.curTri)
      fval0 = Levsp->get_float_val();
    break;
  case _surfsp:
    if(gluvv.tf.curTri)
	fval0 = Surfsp->get_float_val();
    break;
  case _maxsurfsp:
    if(gluvv.tf.curTri)
      fval0 = MaxSurfsp->get_float_val();
    break;
  case _alphasp:
    if(gluvv.tf.curTri)
      fval0 = Alphasp->get_float_val();
    break;
  case _huesp:
    if(gluvv.tf.curTri)
      fval0 = Huesp->get_float_val();
    break;
  case _tritype:
    ival0 = TriType->get_int_val();
    break;
  //////**********************
    
    
    //------------------------------------------------------ brushes
  case _brushgrp: //brush select
    ival0 = ddBrushes->get_int_val();
    aprint( << "brush group id = " << val << endl);
    break;

    //---------------------------------------------------- painting
  case _ddpaint: //paint
    break;
  case _dddrop: //drop
    break;
  case _ddclear: //clear
    break;
  case _ddsize: //brush size
    fval0 = ddSizeSpin->get_float_val();
    break;

    //---------------------------------------------------- shading
  case _shadegrp: //shade mode
    ival0 = Shaders->get_int_val();
    break;
  case _ambient:
    fval0 = ambientSpin->get_float_val();
    break;
    //Lighting --------------------
  case _lightRad:
    fval0 = lightRadSpin->get_float_val();
    break;
  case _lightchk:
    break;
  case _ltatt:
    ival0 = LtAtt->get_int_val();
    break;
  case _ltattthresh:
    fval0 = LtAttThresh->get_float_val();
    break;
  case _ltattnear:
    fval0 = LtAttNear->get_float_val();
    break;
  case _ltattfar:
    fval0 = LtAttFar->get_float_val();
    break;
    //end Lighting ----------------
    //Shadows ---------------------
  case _shadowchk:
    ival0 = Shadow->get_int_val();
    break;
  case _softchk:
    ival0 = Soft->get_int_val();
    break;
  case _gshadowqual:
    fval0 = GShadowQualSpin->get_float_val();
    break;
  case _ishadowqual:
    fval0 = IShadowQualSpin->get_float_val();
    break;
  case _showltchk:
    ival0 = showLT->get_int_val();
    break;
    //end Shadows -----------------
    //Silhouettes -----------------
  case _sillchk:
    ival0 = Sill->get_int_val();
    break;
    //Fog -------------------------
  case _fogcolgrp:    
    ival0 = FogColors->get_int_val();
    break;
  case _fogthresh:
    fval0 = FogThresh->get_float_val();
    break;
  case _fognear:
    fval0 = FogNear->get_float_val();
    break;
  case _fogfar:
    fval0 = FogFar->get_float_val();
    break;
    //end Fog ---------------------
    //---------------------------------------------------- perturbation
  case _pertW1:
    fval0 = PertW1->get_float_val();
    break;
  case _pertS1:
    fval0 = PertS1->get_float_val();
    break;
  case _pertW2:
    fval0 = PertW2->get_float_val();
    break;
  case _pertS2:
    fval0 = PertS2->get_float_val();
    break;
  case _pertW3:
    fval0 = PertW3->get_float_val();
    break;
  case _pertS3:
    fval0 = PertS3->get_float_val();
    break;
  case _pertW4:
    fval0 = PertW4->get_float_val();
    break;
  case _pertS4:
    fval0 = PertS4->get_float_val();
    break;
    //---------------------------------------------------- sample rates
  case _interactSR: //interactive sample rate
    fval0 = interactSpin->get_float_val();
    break;
  case _goodSR:  //high quality sample rate
    fval0 = goodSpin->get_float_val();
    break;
		
    //---------------------------------------------------- clipping
  case _clipgrp: //clipping plane selection
    ival0 = Clips->get_int_val();
    break;

  case _clippos:
    if(gluvv.clip.clipWidget) 
      fval0 = ClipPos->get_float_val();
    break;
  case _clipalpha:
    if(gluvv.clip.clipWidget) 
      fval0 = ClipAlpha->get_float_val();
    break;
  case _clipsize:
    if(gluvv.clip.clipWidget)
      fval0 = ClipSize->get_float_val();
    break;
  case _clipprobe:
    if(gluvv.clip.clipWidget)
      ival0 = ClipProbe->get_int_val();
    break;
    
    //---------------------------------------------------- histo
  case _histochk:
    ival0 = Histo->get_int_val();
    break;

    //---------------------------------------------------- normals
  case _normpostBtn:
    aprint( << "Loading post-classification normals" << endl);
    break;
  case _normpreBtn:
    aprint( << "Loading pre-classification normals" << endl);
    break;

    //--------------------------------------------------- back ground color
  case _bggrp:
    ival0 = bGround->get_int_val();
    break;
  }

#ifdef PARALLEL
  bcast_control(control, ival0, fval0);
#else
  remote_control_cb(control, ival0, fval0);
#endif
}

void gluvvui::remote_control_cb(int control, int ival0, float fval0)
{
  int val;
  LevWidget *lwtmp;
  float b[2],l[2],r[2],t[2];
  
  switch(control){
  case _close:
    hide();
    break;

    //------------------------------------------------------ brushes
  //////**********************
  case _newtri:
    lwtmp = new LevWidget(&gluvv);
    lwtmp->oneDType();
    b[0] = 0; b[1] = 0;
    l[0] = 0; l[1] = 1;
    r[0] = 1; r[1] = 1;
    t[0] = 1; t[1] = 0;
    lwtmp->setPos(b,l,r,t[0],t[1]);
    lwtmp->setInteraction(LevWidget::LWWinLev);
    lwtmp->setWindow(1);
    lwtmp->setLevel(.5);
    gluvv.tf.curTri = lwtmp;
    lwtmp->setName("New Classification");
    srand(2893);
    lwtmp->setHue( (float)rand()/(float)RAND_MAX );
    gluvv.tf.tfWidget->getTris()->insert(lwtmp);
    updateTF();
    gluvv.tf.loadme = 1;
    PglutPostRedisplay();
    break;
  case _deltri:
    if( gluvv.tf.curTri && (gluvv.tf.curTri != gluvv.tf.tfWidget->getTris()) )
      {
	gluvv.tf.tfWidget->getTris()->remove( gluvv.tf.curTri->id );
	delete gluvv.tf.curTri;
	gluvv.tf.curTri = 0;
	updateTF();
      }
    break;


  case _trihide:
    if(gluvv.tf.curTri)
      {
	gluvv.tf.curTri->setHide(ival0);
	updateTF();
      }
    break;

  case _trihilight:
    gluvv.hilight = ival0;
    updateTF();
    PglutPostRedisplay();
    break;

  case _nexttri:
    if(gluvv.tf.curTri)
      {
	gluvv.tf.curTri->setPicked(0);
	LevWidget *nt = gluvv.tf.curTri->getNext();
	if(nt) gluvv.tf.curTri = nt;
	updateTF();
      }
    else
      {
	LevWidget *nt = gluvv.tf.tfWidget->getTris();
	if(nt) nt= nt->getNext();
	if(nt)
	  {
	    gluvv.tf.curTri = nt;
	    updateTF();
	  }
			
      }
    PglutPostRedisplay();
    break;

  case _prevtri:

    if( !gluvv.tf.curTri )
      {
	LevWidget *nt = gluvv.tf.tfWidget->getTris();
	if(nt) nt= nt->getNext();
	if(nt)
	  {
	    gluvv.tf.curTri = nt;
	    updateTF();
	  }
      }
    else
      {
	LevWidget *pt = gluvv.tf.tfWidget->getTris();
	while(pt && (pt->getNext() != gluvv.tf.curTri) ) pt = pt->getNext();
	if(pt)
	  {
	    gluvv.tf.curTri->setPicked(0);
	    gluvv.tf.curTri = pt; 
	    updateTF();
	  }

      }
    PglutPostRedisplay();

    break;

  case _triname:

    //if(gluvv.tf.curTri) gluvv.tf.curTri->setName( tval0 );

    break;

  case _winsp:
		
    if(gluvv.tf.curTri)
      gluvv.tf.curTri->setWindow( fval0 );

    PglutPostRedisplay();
    break;
  case _levsp:

    if(gluvv.tf.curTri)
      gluvv.tf.curTri->setLevel( fval0 );

    PglutPostRedisplay();
    break;

  case _surfsp:

    if(gluvv.tf.curTri)
      {
	gluvv.tf.curTri->setSurfMin( fval0 );
	updateTF();
	PglutPostRedisplay();
      }

    break;

  case _maxsurfsp:

    if(gluvv.tf.curTri)
      {
	gluvv.tf.curTri->setSurfMax( fval0 );
	updateTF();
	PglutPostRedisplay();
      }

    break;
  case _alphasp:

    if(gluvv.tf.curTri)
      {
	gluvv.tf.curTri->setAlpha( fval0 );
	gluvv.tf.loadme = 1;
	PglutPostRedisplay();
      }

    break;

  case _huesp:
    if(gluvv.tf.curTri)
      {
	gluvv.tf.curTri->setHue( fval0 );
	PglutPostRedisplay();
      }

    break;


  case _tritype:
    switch( ival0 ){
    case 0:
      if(gluvv.tf.curTri)
	{
	  gluvv.tf.curTri->oneDType();
	  gluvv.tf.curTri->setInteraction(LevWidget::LWWinLev);
	  updateTF();
	  PglutPostRedisplay();
	}

      break;
    case 1:
      if(gluvv.tf.curTri)
	{
	  gluvv.tf.curTri->oneDType();
	  gluvv.tf.curTri->setInteraction(LevWidget::LWStandard);
	  updateTF();
	  PglutPostRedisplay();
	}
      break;
    case 2:
      if(gluvv.tf.curTri)
	{
	  gluvv.tf.curTri->triangleType();
	  gluvv.tf.curTri->setWindow( gluvv.tf.curTri->getWindow() );
	  gluvv.tf.curTri->setLevel(gluvv.tf.curTri->getLevel() );
	  gluvv.tf.curTri->setInteraction(LevWidget::LWStandard);
	  updateTF();
	  PglutPostRedisplay();
	}
      break;
    }
		
    break;
    
    //////**********************
  case _brushgrp: //brush select
    val = ival0;
    switch(val){ //brush type-----
    case 0: //off
      gluvv.tf.brushon = 0;
      gluvv.tf.brushon = 0;
      gluvv.tf.loadme = 1;
      PglutPostRedisplay();
      break;
    case 1: //ellipse brush
      gluvv.tf.brushon = 1;
      gluvv.probe.brush = EllipseBrush;
      gluvv.tf.loadme = 1; 
      PglutPostRedisplay();
      break;
    case 2: //auto ellipse
      gluvv.tf.brushon = 1;
      //gluvv.probe.slider = .5;
      gluvv.probe.brush = AutoEllipseBrush;
      gluvv.tf.loadme = 1; 
      PglutPostRedisplay();
      break;
    case 3: //triangle brush
      gluvv.tf.brushon = 1;
      //gluvv.probe.slider = .5;
      gluvv.probe.brush = TriangleBrush;
      gluvv.tf.loadme = 1; 
      PglutPostRedisplay();
      break;
    case 4:
      gluvv.tf.brushon = 1;
      //gluvv.probe.slider = .5;
      gluvv.probe.brush = OneDBrush;
      gluvv.tf.loadme = 1; 
      PglutPostRedisplay();
      break;
    case 5: //one d
      gluvv.tf.brushon = 1;
      //gluvv.probe.slider = .5;
      gluvv.probe.brush = AutoOneDBrush;
      gluvv.tf.loadme = 1; 
      PglutPostRedisplay();
      break;
    }//----------------------------
    break;

    //---------------------------------------------------- painting
  case _ddpaint: //paint
    gluvv.tf.paintme = 1;
    PglutPostRedisplay();
    break;
  case _dddrop: //drop
    gluvv.tf.dropme = 1;
    PglutPostRedisplay();
    break;
  case _ddclear: //clear
    gluvv.tf.clearpaint = 1;
    PglutPostRedisplay();
    break;
  case _ddsize: //brush size
    gluvv.probe.slider = 1.0 - fval0;
    gluvv.tf.loadme = 1;
    PglutPostRedisplay();
    break;

    //---------------------------------------------------- shading
  case _shadegrp: //shade mode
    val = ival0;
    switch(val){  //select shade mode
    case 0: //emmisive
      gluvv.shade = gluvvShadeAmb;
      gluvv.tf.loadme = 1;
      PglutPostRedisplay();
      break;
    case 1: //shaded
      gluvv.shade = gluvvShadeDSpec;
      gluvv.tf.loadme = 1;
      PglutPostRedisplay();
      break;
    case 2: //faux
      gluvv.shade = gluvvShadeFaux;
      gluvv.tf.loadme = 1;
      PglutPostRedisplay();
      break;
    }
    break;
  case _ambient:
    gluvv.light.amb = fval0;
    PglutPostRedisplay();
    break;
    //Lighting --------------------
  case _lightRad:
    normalizeV3(gluvv.light.pos);
    scaleV3(fval0, gluvv.light.pos);
    ltwr.genXForm();
    PglutPostRedisplay();
    break;
  case _lightchk:
    ltwr.onoff();
    PglutPostRedisplay();
    break;
  case _ltatt:
    gluvv.light.latt = ival0;
    PglutPostRedisplay();
    break;
  case _ltattthresh:
    gluvv.light.lattThresh = 1.0 - fval0;
    PglutPostRedisplay();
    break;
  case _ltattnear:
    gluvv.light.lattLimits[0] = fval0/2;
    PglutPostRedisplay();
    break;
  case _ltattfar:
    gluvv.light.lattLimits[1] = fval0/2;
    PglutPostRedisplay();
    break;
    //end Lighting ----------------
    //Shadows ---------------------
  case _shadowchk:
    gluvv.light.shadow = ival0;
    PglutPostRedisplay();
    break;
  case _softchk:
    gluvv.light.softShadow = ival0;
    PglutPostRedisplay();
    break;
  case _gshadowqual:
    gluvv.light.gShadowQual = fval0;
    PglutPostRedisplay();
    break;
  case _ishadowqual:
    gluvv.light.iShadowQual = fval0;
    PglutPostRedisplay();
    break;
  case _showltchk:
    gluvv.light.showView = ival0;
    PglutPostRedisplay();
    break;
    //end Shadows -----------------
    //Silhouettes -----------------
  case _sillchk:
    gluvv.light.sill = ival0;
    PglutPostRedisplay();
    break;
    //Fog -------------------------
  case _fogcolgrp:
    switch(ival0){
    case 0:
      gluvv.light.fog = 0;
      break;
    case 1:
      gluvv.light.fog = 1;
      gluvv.light.fogColor[0] = 0;
      gluvv.light.fogColor[1] = 0;
      gluvv.light.fogColor[2] = .5;
      break;
    case 2:
      gluvv.light.fog = 1;
      gluvv.light.fogColor[0] = .9;
      gluvv.light.fogColor[1] = .9;
      gluvv.light.fogColor[2] = .9;
      break;
    case 3:
      gluvv.light.fog = 1;
      gluvv.light.fogColor[0] = .1;
      gluvv.light.fogColor[1] = .1;
      gluvv.light.fogColor[2] = .1;				
      break;
    }
    PglutPostRedisplay();
    break;
  case _fogthresh:
    gluvv.light.fogThresh = 1.0 - fval0;
    PglutPostRedisplay();
    break;
  case _fognear:
    gluvv.light.fogLimits[0] = fval0/2;
    PglutPostRedisplay();
    break;
  case _fogfar:
    gluvv.light.fogLimits[1] = fval0/2;
    PglutPostRedisplay();
    break;
    //end Fog ---------------------
    //---------------------------------------------------- perturbation
  case _pertW1:
    gluvv.pert.weights[0] = fval0;
    PglutPostRedisplay();
    break;
  case _pertS1:
    gluvv.pert.scales[0] = fval0;
    PglutPostRedisplay();
    break;
  case _pertW2:
    gluvv.pert.weights[1] = fval0;
    PglutPostRedisplay();
    break;
  case _pertS2:
    gluvv.pert.scales[1] = fval0;
    PglutPostRedisplay();
    break;
  case _pertW3:
    gluvv.pert.weights[2] = fval0;
    PglutPostRedisplay();
    break;
  case _pertS3:
    gluvv.pert.scales[2] = fval0;
    PglutPostRedisplay();
    break;
  case _pertW4:
    gluvv.pert.weights[3] = fval0;
    PglutPostRedisplay();
    break;
  case _pertS4:
    gluvv.pert.scales[3] = fval0;
    PglutPostRedisplay();
    break;
    //---------------------------------------------------- sample rates
  case _interactSR: //interactive sample rate
    gluvv.volren.interactSamp = fval0;
    PglutPostRedisplay();
    break;
  case _goodSR:  //high quality sample rate
    gluvv.volren.goodSamp = fval0;
    PglutPostRedisplay();
    break;
		
    //---------------------------------------------------- clipping
  case _clipgrp: //clipping plane selection
    val = ival0;
    switch(val){
    case 0: //off
      gluvv.clip.on = 0;
      PglutPostRedisplay();
      break;
    case 1: //x pos
      gluvv.clip.on = 1;
      gluvv.clip.oaxis = VolRenAxisXPos;
      PglutPostRedisplay();
      break;
    case 2: //x neg
      gluvv.clip.on = 1;
      gluvv.clip.oaxis = VolRenAxisXNeg;
      PglutPostRedisplay();
      break;
    case 3: //y pos
      gluvv.clip.on = 1;
      gluvv.clip.oaxis = VolRenAxisYPos;
      PglutPostRedisplay();
      break;
    case 4: //y neg
      gluvv.clip.on = 1;
      gluvv.clip.oaxis = VolRenAxisYNeg;
      PglutPostRedisplay();
      break;
    case 5: //z pos
      gluvv.clip.on = 1;
      gluvv.clip.oaxis = VolRenAxisZPos;
      PglutPostRedisplay();
      break;
    case 6: //z neg
      gluvv.clip.on = 1;
      gluvv.clip.oaxis = VolRenAxisZNeg;
      PglutPostRedisplay();
      break;
    }
    update();
    break;

  case _clippos:
    if(gluvv.clip.clipWidget) 
      gluvv.clip.clipWidget->setOrthoPos( fval0 );
    break;
  case _clipalpha:
    if(gluvv.clip.clipWidget) 
      gluvv.clip.clipWidget->setAlpha( fval0 );
    break;
  case _clipsize:
    if(gluvv.clip.clipWidget)
      gluvv.clip.clipWidget->setSize( fval0 );
    break;
  case _clipprobe:
    if(gluvv.clip.clipWidget)
      gluvv.clip.clipWidget->setProbe( ival0 );
    break;
		
    //---------------------------------------------------- histo
  case _histochk:
    gluvv.tf.histOn = ival0;
    PglutPostRedisplay();
    break;

    //---------------------------------------------------- normals
  case _normpostBtn:
    gluvv.volren.usePostCNorm = 1;
    gluvv.volren.loadNorms	= 1;
    PglutPostRedisplay();
    break;
  case _normpreBtn:
    PglutPostRedisplay();
    break;

    //--------------------------------------------------- back ground color
  case _bggrp:
    gluvv.env.bgColor = ival0;
    PglutPostRedisplay();
    break;
  }
}
