//------------------------------------------------------------------------
//
//   Joe Kniss
//    A glut implementation of the Volume Renderer
//   gluvv.cpp "GL Utilites for Virtual reality Volume rendering"
//     3-4-01
//		     ________	 ____	___ 
//		    |	     \	/    | /  /
//		    +---+     \/     |/	 /
//		    +--+|  |\	 /|	< 
//		    |  ||  | \	/ |  |\	 \ 
//		    |	   |  \/  |  | \  \ 
//		     \_____|	  |__|	\__\
//			 Copyright  2001 
//			Joe Michael Kniss
//		     <<< jmk@cs.utah.edu >>>
//		 "All Your Base are Belong to Us"
//
//	 Linux implementation --------------------------------------
//-------------------------------------------------------------------------

#include <unistd.h>
#include <iostream>
#include <fstream>
using namespace std;
#include <string.h>
#include <stdlib.h>
#ifdef WIN32
#include <windows.h>		//  must be included before any OpenGL
#endif
#include <GL/glut.h>	   //glut includes gl and glu for you!

#include "Trackball.h"	   //SGI quaternion trackball

#include "VectorMath.h"	   //usefull vector/matrix operations

#include "gluvv.h"	   //global definitions

#include "glUE.h"	   //gl utilities and extensions

#include "gluvvui.h"	   //the user interface
#include "saveppm.h"

#include "glutredirect.h"

#include <signal.h> //sigaction
#include <execinfo.h>

//----------------------------------------------------------------
//----------- Renderable includes --------------------------------
#include "NV20VolRen3D.h"
#include "TFWidgetRen1.h"
#include "CPWidgetRen.h"
#include "DPWidgetRen.h"
#include "LTWidgetRen.h"

// What: Some compiling parameters.
//
#undef DEBUG
#undef DEBUG_BLENDING

#include "MpiStuff.h"
#include "SortLast.h"

int SCREEN_RES=512;

#define MAX_MPI_PROCESSES 128

// What: MPI stuff
// _______________

MPI_Comm  worldComm;   // The MPI world communicator.
int	  worldSize;   // The number of MPI processes.
MpiIdType worldId;     // The process MPI id (in world AND slaves comm).

MPI_Comm  slavesComm;	 // The slaves communicator.
int	  slavesSize;	 // The number of slaves
int	  masterIsSlave; // Is the master (MPI #0) part of the slaves?

char hostName[256];  // The name of the local host where the process runs.

NV20VolRen3D *myVolRen; // The vol renderer of the process (to ask for Z).
TFWidgetRen *tfwr; //hartner+
SortLast *mySortLast;	// The sort last of the process.
int continuousDisplay = 1;

//----------------------------------------------------------------
//---- the global data structures --------------------------------
gluvvGlobal	 gluvv;	      //the global gluvv data structure (gluvv.h)
Trackball	 tball;	      //global trackball
Trackball	 lball;	      //for moving the light (now using light widget)
char		 loadname[1024];
gluvvPrimitive	 renderables; //all renderables attatch to this one
//gluvvPrimitive   widgets; //all renderables attatch to this one
TFWindow	 tfwin;	      //separate transfer function window (not used??)
LTWidgetRen	 ltwr;	      //need the light widget to be accesable to other things
gluvvui		 gui;	      //the user interface

int diag_show_node = 0;
bool diag_show_subvol_bbox = false;
bool diag_show_vol_bbox = false;
bool diag_show_objs_behind = true;
bool diag_show_objs_inside = true;
bool diag_show_objs_infront = true;
bool synch_dont_redraw = false;
bool synch_firstpass = true;

FILE *lfile = 0;
bool record = false;
bool playback = false;

FILE *ffile = 0;
bool logframes = false;

double ntime = 0.0; //null time, which we don't care about but need to start stopwatch counters
double atime = 0.0; //animation time, time to read and interpret .rec file
struct timeval ats;
struct timeval ate;
double rtime = 0.0; //local render time
struct timeval rts;
struct timeval rte;
double srttime = 0.0; //local render time
struct timeval sts;
struct timeval ste;
double cotime = 0.0; //composite time
struct timeval cts;
struct timeval cte;
double dtime = 0.0; //display time
struct timeval dts;
struct timeval dte;
double ttime = 0.0; //total time
struct timeval tts;
struct timeval tte;

#include <sys/time.h>

//#define DOSTATS
#ifdef DOSTATS
#define timestampe(et, ts,te) \
  {\
    if (logframes) {\
      if (gettimeofday(&te, 0) != 0) cerr << "Error getting time" << endl;\
      et = te.tv_sec-ts.tv_sec + (te.tv_usec-ts.tv_usec)*1.e-6;\
      ts.tv_sec = te.tv_sec;\
      ts.tv_usec = te.tv_usec;\
    }\
  }
#else
#define timestampe(et, ts,te) {}
#endif

//----------------------------------------------------------------
//---- GLUT Stuff -------------------------------------------------
int  initGLUTWindow(int argc, char * argv[]);
void glutSlaveIdle(void);
void glutSlaveDisp(void);
void glutSlaveReshape(int w, int h);

void glutMasterDisp(void);
void glutMasterIdle(void);
void glutMasterReshape(int w, int h);
void glutMasterMouse(int button, int state, int x, int y);
void glutMasterMotion(int x, int y);
void glutMasterKey(unsigned char key, int x, int y);
void glutMasterSpecial(int key, int x, int y);

void selfDisplay(void);
float display(int pick);
void selfReshape(int w, int h);
void selfMouse(int button, int state, int x, int y);
void selfMotion(int x, int y);
void selfKey(unsigned char key, int x, int y);
void selfSpecial(int key, int x, int y);

int oldW = 0;
int oldH = 0;

//----------------------------------------------------------------
//---- GL Stuff ---------------------------------------------------
void	initGL();  //initialize gl state

//----------------------------------------------------------------
//---- Utils ------------------------------------------------------
void	initGluvv();		      //"Constructor for gluvv data structures"
void	initRenderables();	      //this gets called after gl context extists
int	pre_parse(int argc, char **argv); //returns 0 if no errors
int	parse(int argc, char **argv); //returns 0 if no errors
void	gluvvQuit();		      //essentialy a destructor
int	pick();			      //pick objects
int	initData();		      //make sure data is in the right format
int	  save(char *name);
int	  load(char *name);
void updateUI() { gui.update(); }

//---- Misc -------------------------------------------------------
int	checkExts();  //returns 1 if everything is available, 0 else
int	blurnorms;    //flag set if normals are to be blurred

//____________________________________________________________________________
//
// Frames per second.
//____________________________________________________________________________

int fpsFrames = 0;
int fpsTime = 0;
int fpsTimeBase = 0;
char fpsString[64] = "0.0 fps";
void *fpsFont = GLUT_BITMAP_9_BY_15;

//____________________________________________________________________________
//
// Initial enter.
//____________________________________________________________________________

void initalEnter(int argc, char * argv[])
{
  MPI_Init(&argc, &argv);
  worldComm = MPI_COMM_WORLD;
  MPI_Comm_size(worldComm, &worldSize);
  MPI_Comm_rank(worldComm, &worldId);
  
  diag_show_node = worldSize;

  if ( worldSize == 3 || worldSize == 5 || worldSize == 9 || 
       worldSize == 17 || worldSize == 33 )
    {
      if ( worldId == 0 )
	{
	  // MPI #0 will be a stand alone master.

	  // XC NOTE: in this case slavesComm is the master's communicator.
	  MPI_Comm_split(worldComm, 0, 0, &slavesComm);
	}
      else
	{
	  // All other MPI #n will be slaves.
	  MPI_Comm_split(worldComm, 1, 0, &slavesComm);
	}      
      masterIsSlave = 0;
      slavesSize = worldSize - 1;
    }
  else
    {
      // All MPI #n are slaves.
      MPI_Comm_split(worldComm, 1, 0, &slavesComm);	
      masterIsSlave = 1;
      slavesSize = worldSize;
    }
  
#ifdef MPE_LOGGING
  MPE_Init_log();
  
  if ( worldId == 0 )
    {
      MPE_Describe_state(MPE_EVENT_RENDER_VOL_B, 
			 MPE_EVENT_RENDER_VOL_E,
			 "Render volume", "green");
      MPE_Describe_state(MPE_EVENT_READ_PIXELS_B, 
			 MPE_EVENT_READ_PIXELS_E,
			 "Read pixels", "blue");
      MPE_Describe_state(MPE_EVENT_SEND_RECV_PIXELS_B, 
			 MPE_EVENT_SEND_RECV_PIXELS_E,
			 "Send & Recv pixels", "red");
      MPE_Describe_state(MPE_EVENT_COMPOSE_B, 
			 MPE_EVENT_COMPOSE_E,
			 "Compose", "yellow");
      MPE_Describe_state(MPE_EVENT_COLLECT_PIXELS_B, 
			 MPE_EVENT_COLLECT_PIXELS_E,
			 "Collect pixels", "orange");
      MPE_Describe_state(MPE_EVENT_DRAW_PIXELS_B, 
			 MPE_EVENT_DRAW_PIXELS_E,
			 "Draw pixels", "purple");
      MPE_Describe_state(MPE_EVENT_BCAST_B,
			 MPE_EVENT_BCAST_E,
			 "Broadcast", "white");
    }
#endif
}

//____________________________________________________________________________
//
// Final exit.
//____________________________________________________________________________

void finalExit(void)
{

#ifdef MPE_LOGGING
  if ( worldId == 0 ) cerr << "Writing log file..." << endl;
  MPE_Finish_log("psimian");
  if ( worldId == 0 ) cerr << "Log file is psimian.clog" << endl;
#endif
  
  MPI_Comm_free(&slavesComm);
  MPI_Finalize();


  if ( worldId == 0 ) cerr << endl << "Au revoir!" << endl << endl;
}

//=================================================================
//=================================================================

//================================================= the Main func.
//===============================================================
int main(int argc, char **argv)
{

  initalEnter(argc, argv);
    
  unsigned int myStages;
  unsigned int myConfig[8];
  
  switch ( slavesSize )
    {
    case 1:
      myStages = 1;
      myConfig[0] = 1;
    case 2:
      myStages = 2;
      myConfig[0] = 2;
      myConfig[1] = 1;	    
      break;
    case 4:
      myStages = 2;
      myConfig[0] = 2;
      myConfig[1] = 2;	    
      break;
    case 8:
      myStages = 3;
      myConfig[0] = 2;
      myConfig[1] = 2;
      myConfig[2] = 2;
      break;
    case 16:
      myStages = 4;
      myConfig[0] = 2;
      myConfig[1] = 2;
      myConfig[2] = 2;
      myConfig[3] = 2;
      break;
    case 32:
      myStages = 5;
      myConfig[0] = 2;
      myConfig[1] = 2;
      myConfig[2] = 2;
      myConfig[3] = 2;
      myConfig[4] = 2;
      break;
    default:
      if ( worldId == 0 )
	{
	  aprint( << endl << argv[0] << ": FATAL ERROR" << endl);
	  aprint( << "Supported number of MPI processes are: 1, 2/3, 4/5, 8/9, 16/17 and 32/33" << endl);
	}
      finalExit();
      return 1;
    }
  
  gethostname(hostName, 256);
  
  start_logger("CF");

  // All should be set for the MPI part.
  // ___________________________________

  //===== Initialization ========================
  //------------------------------------init gluvv
  if (pre_parse(argc, argv)) return 1;

  initGluvv();
  loadname[0] = '\0';

  //------------------------------------parse 
  if(parse(argc, argv)) return 1;  
  
  if ( worldId || masterIsSlave )
    {
      if ( initData() )
	{ 
	  aprint( << "Data read failed. Make sure you have the correct path specified" << endl);
	  finalExit();
	  return 1;
	}
    }
  
  initGLUTWindow(argc, argv);
  
  if( ! checkExts() )
    {
      aprint( << "Sorry, this platform is not supported" << endl);
      aprint( << "	 Please make sure you have a GeForce 3 or Radeon 8K with the correct driver" << endl);
      finalExit();
      return 1;
    }
  
  if ( worldId || masterIsSlave )
    {
      myVolRen = new NV20VolRen3D;
      renderables.setNext((gluvvPrimitive*)myVolRen);  
    }

  tfwr = new TFWidgetRen;
  renderables.setNext((gluvvPrimitive*)tfwr);
  gluvv.tf.tfWidget = tfwr;
  CPWidgetRen *cpwr = new CPWidgetRen;
  renderables.setNext((gluvvPrimitive*)cpwr);
  gluvv.clip.clipWidget = cpwr;
  DPWidgetRen *dpwr = new DPWidgetRen;
  renderables.setNext((gluvvPrimitive*)dpwr);  
  renderables.setNext((gluvvPrimitive*)&ltwr);

  initRenderables();
  if(loadname[0] != '\0')
    load(loadname);

  gui.show();
  
  mySortLast = new SortLast(myStages, myConfig, 
			    gluvv.win.width, gluvv.win.height,
			    worldComm, slavesComm, 
			    slavesSize, masterIsSlave);
  
  // Just to synchronize before going.
  MPI_Barrier(worldComm);

  alog( << "MPI #" << worldId << " / " << worldSize << " is running on " << hostName << endl);
  
  cerr << "MPI #" << worldId << " entering Main Loop" << endl;
 
  glutMainLoop();
  
  // Should never get there.
  finalExit();
  return 0;  
}
//=================================================================
//=================================================================

//=============================================== init Gluvv
//==========================================================
void	initGluvv(){		      //"Constructor for gluvv data structures"
  //-------------------- initialize global state ---------------

  gluvv.mv	    = 0;
  gluvv.mv1	    = 0;
  gluvv.mv2	    = 0;
  gluvv.mv3	    = 0;
#ifdef PARALLEL
  gluvv.normalsRestored = 0; //normals restored from a trex brick file?
#endif
  gluvv.front2backSlicing = 1;
  gluvv.hilight = 1; //emphasize selected levoy widget in the rendered volume
  gluvv.numHues = 5;
  gluvv.currentHue = 0;
  gluvv.hues = new float[gluvv.numHues];
  gluvv.hues[0] = 0.3;
  gluvv.hues[1] = 0.7;
  gluvv.hues[2] = 0.1;
  gluvv.hues[3] = 0.5;
  gluvv.hues[4] = 0.9;
  
  
  gluvv.debug	    = 0;
  gluvv.dmode	    = GDM_V1;
  
  gluvv.picking = 0;
  gluvv.mprobe	= clipPlane;
  gluvv.shade	= gluvvShadeFaux;
  
  gluvv.reblend = GB_NONE;
  
  //gluvv.spec	  = 0;	//not in spectral mode
  
  //------------------------------------window
  gluvv.win.width  = SCREEN_RES;     //size
  gluvv.win.height = SCREEN_RES;
  gluvv.win.xPos   = 100;     //position
  gluvv.win.yPos   = 100;
  //------------------------------------environment
  gluvv.env.eye[0]     = 0;   //eye
  gluvv.env.eye[1]     = 0;
  gluvv.env.eye[2]     = -7;
  gluvv.env.at[0]      = 0;   //at
  gluvv.env.at[1]      = 0;
  gluvv.env.at[2]      = 0;
  gluvv.env.up[0]      = 0;   //up
  gluvv.env.up[1]      = 1;
  gluvv.env.up[2]      = 0;
  gluvv.env.frustum[0] = -.2;  //left
  gluvv.env.frustum[1] = .2;   //right
  gluvv.env.frustum[2] = -.2;  //bottom
  gluvv.env.frustum[3] = .2;   //top
  gluvv.env.clip[0]    = 1;    //front
  gluvv.env.clip[1]    = 20;   //back
  gluvv.env.bgColor	  = 0;	  //white background
  //------------------------------------mouse
  gluvv.mouse.alt     = 0;
  gluvv.mouse.ctrl    = 0;
  gluvv.mouse.shift   = 0;
  //------------------------------------lights & lighting
  gluvv.light.pos[0] = gluvv.light.startpos[0] = 0;  
  gluvv.light.pos[1] = gluvv.light.startpos[1] = 0;
  gluvv.light.pos[2] = gluvv.light.startpos[2] = -5;
  gluvv.light.buffsz[0] = 1024; //shadow buffer size
  gluvv.light.buffsz[1] = 1024;
  gluvv.light.shadow	= 0;   //shadows off
  gluvv.light.softShadow= 1;   //soft shadows
  gluvv.light.showView	= 0;   //show the view from the light
  gluvv.light.sill	= 0;   //show silhouette edges
  gluvv.light.amb	= .05;	//shadow strenght
  gluvv.light.intens	= .75;
  gluvv.light.latt	= 0;
  gluvv.light.lattLimits[0] = .5;  //attinuation limits (start)
  gluvv.light.lattLimits[1] = -.5; //attinuation limits (finish)
  gluvv.light.lattThresh    = 1;   //amount of attinuation 
  gluvv.light.gShadowQual  = .5;    //good shadow quality
  gluvv.light.iShadowQual  = .20;  //interactive shadow quality
  gluvv.light.csz	   = 32;   //size of the cube map
  gluvv.light.cubeKey[0]   = 0;	  //x+ cube key not created yet
  gluvv.light.cubeKey[1]   = 0;	  //x- cube key not created yet
  gluvv.light.cubeKey[2]   = 0;	  //y+ cube key not created yet
  gluvv.light.cubeKey[3]   = 0;	  //y- cube key not created yet
  gluvv.light.cubeKey[4]   = 0;	  //z+ cube key not created yet
  gluvv.light.cubeKey[5]   = 0;	  //z- cube key not created yet
  gluvv.light.load	   = 1;	  //load lighting info
  gluvv.light.gload	   = 1;	  //load the good lighting
  gluvv.light.fog	   = 0;	  //fog is off
  gluvv.light.fogThresh	   = .5;  //fog blend threshold
  gluvv.light.fogLimits[0] = .5;  //near fog ramp (start)
  gluvv.light.fogLimits[1] = -.5; //far fog ramp (finish)
  //------------------------------------render info
  identityMatrix(gluvv.rinfo.xform);  //init the rotation
  gluvv.rinfo.scale	  = SCREEN_RES/512;  //scale dep on res
  gluvv.rinfo.trans[0]	  = 0;	      //no translation
  gluvv.rinfo.trans[1]	  = 0;
  gluvv.rinfo.trans[2]	  = 0;
  //------------------------------------volume rendering stuff
  gluvv.volren.interactSamp= .6;      //Interactive sample rate
  gluvv.volren.goodSamp	   = 2.5;     //High quality sample ..
  gluvv.volren.sampleRate  = gluvv.volren.goodSamp; //samples per voxel
  gluvv.volren.shade	   = 0;	      //shade off (this flag no longer used)
  gluvv.volren.deptex	   = 0;	      //1st & 2nd axies transfer function
  gluvv.volren.deptex2	   = 0;	      //3rd axis transfer function
  gluvv.volren.deptex3	   = 0;	      //scattering (auxilary 2D) transfer
  gluvv.volren.loadTLUT	   = 0;	      //no need to load the tf
  gluvv.volren.scaleAlphas = 1;	      //scale alphas to the sample rate
  gluvv.volren.gamma	   = 1;	      //gamma identitiy
  gluvv.volren.timestep	   = 0;	      //current timestep
  gluvv.volren.usePostCNorm= 0;	      //use pre-computed Normals to start with
  gluvv.volren.loadNorms   = 0;	      //no need to reload normals
  //------------------------------------transfer function
  gluvv.tf.tfWindow	   = 0;	      //not in separate window
  gluvv.tf.loadme	   = 0;	      //no need to load the tf
  gluvv.tf.paintme	   = 0;	      //nothing to paint into it
  gluvv.tf.dropme	   = 0;	      //not widgets to drop
  gluvv.tf.clearpaint	   = 0;	      //already cleared
  gluvv.tf.brushon	   = 0;	      //brush is off
  gluvv.tf.slider1	   = 1;	      //3rd axis slider high
  gluvv.tf.slider1hi	   = 1;	      //these are sliders for special cases
  gluvv.tf.slider1lo	   = 0;
  gluvv.tf.slider2	   = 1;
  //gluvv.tf.loadHist	     = 0;
  gluvv.tf.histOn	   = 1;	      //histogram on
  gluvv.tf.ptexsz[0]	   = 256;     //size of the transfer function axies
  gluvv.tf.ptexsz[1]	   = 256;
  gluvv.tf.ptexsz[2]	   = 1;	      //3rd axis handled sparately
  gluvv.tf.numelts	   = 4;	      // RGBA colors
  gluvv.tf.curTri          = 0;
  gluvv.tf.movedSinceDropped = 1;     //has the TF brused moved since it was last droppped
  //gluvv.tf.use3D	     = 0;	//not using a 3D tf
  //------------------------------------clip plane
  gluvv.clip.mousepnt[0] = 0;
  gluvv.clip.mousepnt[1] = 0;
  gluvv.clip.mousepnt[2] = 0;
  gluvv.clip.ortho	 = 1;
  gluvv.clip.oaxis	 = VolRenAxisXPos;
  gluvv.clip.clipWidget  = 0;
  //------------------------------------perturbation
  gluvv.pert.on		= 0;  
  gluvv.pert.numHarm	= 0;  //this is not used right now
  gluvv.pert.weights[0] = 0;
  gluvv.pert.scales[0]	= 1;
  gluvv.pert.weights[1] = 0;
  gluvv.pert.scales[1]	= 2;
  gluvv.pert.weights[2] = 0;
  gluvv.pert.scales[2]	= 4;
  gluvv.pert.weights[3] = 0;
  gluvv.pert.scales[3]	= 8;
  //------------------------------------data probe
  gluvv.probe.brush = NoBrush;
  gluvv.probe.vval[0] = 0;
  gluvv.probe.vval[1] = 0;
  gluvv.probe.vval[2] = 0;
  //------------------------------------trackball
  tball.clear();
  lball.clear();
  
  //------------------------------------misc
  blurnorms = 0;

  //------------------------------------scene graph
  renderables.setName("Dummy Node");
}

//================================================ init GLUT
//==========================================================
int initGLUTWindow(int argc, char **argv)
{
  char displaymode[256];
  strcpy(displaymode, "double rgba depth");
  
  glutInit(&argc, argv);

  glutInitDisplayString(displaymode);
  //glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);

  glutInitWindowSize(gluvv.win.width, gluvv.win.height);
  glutInitWindowPosition(gluvv.win.xPos, gluvv.win.yPos);
  gluvv.mainWindow =
    glutCreateWindow("GL Utilites for Virtual Reality Volume Rendering");
  
#ifdef DEBUG
  //print out all of the extensions for the card!
  cerr << endl
       << ">>>>>>>>>>>>> GL Extensions <<<<<<<<<<<<<<" << endl
       << "------------------------------------------" << endl;
  QueryExtension("Print");
  cerr << "------------------------------------------" << endl << endl;
#endif

#ifdef WIN32
  LoadAllExtensions();
#endif
  
  // print out some window attributes for our (ribbed) pleasure. 
  {
    int r, g, b, a, ar, ag, ab, aa, d, s;
    GLboolean db,st, aux;
    
    glGetIntegerv(GL_RED_BITS,		 &r);
    glGetIntegerv(GL_GREEN_BITS,	 &g);
    glGetIntegerv(GL_BLUE_BITS,		 &b);
    glGetIntegerv(GL_ALPHA_BITS,	 &a);
    glGetIntegerv(GL_ACCUM_RED_BITS,	 &ar);
    glGetIntegerv(GL_ACCUM_GREEN_BITS,	 &ag);
    glGetIntegerv(GL_ACCUM_BLUE_BITS,	 &ab);
    glGetIntegerv(GL_ACCUM_ALPHA_BITS,	 &aa);
    glGetIntegerv(GL_DEPTH_BITS,	 &d);
    glGetIntegerv(GL_STENCIL_BITS,	 &s);
    glGetBooleanv(GL_DOUBLEBUFFER,	 &db);
    glGetBooleanv(GL_STEREO,		 &st);
    glGetBooleanv(GL_AUX_BUFFERS,	 &aux);
    
    cerr << "Window Attributes" << endl;
    cerr << "-----------------" << endl;
    cerr << "color:    " << r + g + b + a << " bits";
    cerr << " (" << r << "," << g << "," << b << "," << a << ")" << endl;
    cerr << "accum:    " << ar+ag+ab+aa	  << " bits";
    cerr << " (" << ar << "," << ag << "," << ab << "," << aa << ")" << endl;
    cerr << "depth:    " << d << " bits" << endl;
    cerr << "stencil:  " << s << " bits" << endl;
    cerr << "double:   "; 
    if(db) cerr << "YES" << endl;
    else   cerr << "NO" << endl;
    cerr << "aux:      ";
    if(aux) cerr << "YES" << endl;
    else    cerr << "NO" << endl;
    cerr << "stereo :  ";
    if(st) cerr << "YES" << endl;
    else   cerr << "NO" << endl;
    cerr << "-----------------" << endl;
  }
   
  initGL();
   
  if ( worldId == 0 )
    {
      glutDisplayFunc(glutMasterDisp);
      glutReshapeFunc(glutMasterReshape);
      glutMouseFunc(glutMasterMouse);
      glutMotionFunc(glutMasterMotion);
      glutKeyboardFunc(glutMasterKey);	 
      glutSpecialFunc(glutMasterSpecial);
      glutIdleFunc(glutMasterIdle);    
    }
  else
    {
      glutDisplayFunc(glutSlaveDisp);
      glutIdleFunc(glutSlaveIdle);
      glutReshapeFunc(glutSlaveReshape);
    }
  
  return 0;
}
//=================================================================
//=================================================================

//=================================================== init GL
//===========================================================
void initGL( void )
{
  /* Black background and interpolated shading. */
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glShadeModel(GL_FLAT);

  /* Set up lighting. mostly for widgets */
  static float ambient[]	     = {0.4, 0.4, 0.4, 4.0};
  static float diffuse[]	     = {0.5, 1.0, 1.0, 1.0};
  static float front_mat_shininess[] = {60.0};
  static float front_mat_specular[]  = {0.2, 0.2,  0.2,	 1.0};
  static float front_mat_diffuse[]   = {0.5, 0.28, 0.38, 1.0};
  static float lmodel_ambient[]	     = {0.2, 0.2,  0.2,	 1.0};
  static float lmodel_twoside[]	     = {GL_FALSE};
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_AMBIENT,  ambient);
  glLightfv(GL_LIGHT0, GL_DIFFUSE,  diffuse);
  glLightfv(GL_LIGHT0, GL_POSITION, gluvv.light.startpos);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT,  lmodel_ambient);
  glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE, lmodel_twoside);
  
  glEnable(GL_LIGHTING);
  glEnable(GL_COLOR_MATERIAL);
  
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, front_mat_shininess);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR,	front_mat_specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,	front_mat_diffuse);
  
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);
  glDisable(GL_CULL_FACE);
  
  //glEnable(GL_LINE_SMOOTH);
  glLineWidth(3);
  
  GlErr("gluvv::","initGL()");
  
  gluvv.env.diff[0] = front_mat_diffuse[0];
  gluvv.env.diff[1] = front_mat_diffuse[1];
  gluvv.env.diff[2] = front_mat_diffuse[2];
  gluvv.env.diff[3] = front_mat_diffuse[3];
  gluvv.env.spec[0] = front_mat_specular[0];
  gluvv.env.spec[1] = front_mat_specular[1];
  gluvv.env.spec[2] = front_mat_specular[2];
  gluvv.env.spec[3] = front_mat_specular[3];
  
}
//=================================================================
//=================================================================

//========================================== init Renderables
//===========================================================
void initRenderables()
{
  gluvvPrimitive *tmp = renderables.getNext();
  while(tmp){
    tmp->init();
    tmp = tmp->getNext();
  }
}
//=================================================================
//=================================================================

void bcast_control(int control, int ival0, float fval0) {

  int buffer[CMD_BUFFER_MAX_SIZE];
  buffer[0] = CMD_GUI;
  buffer[1] = control;
  buffer[2] = ival0;
  *((float*)(&buffer[3])) = fval0;
  alog( << "s bcast GUI control " << control << endl);

  MPI_Bcast(buffer, CMD_BUFFER_MAX_SIZE, MPI_INT, 0, worldComm);
  alog( << "e bcast GUI control " << control << endl);

  gui.remote_control_cb(control, ival0, fval0);
}

//================================================= glut Disp
//===========================================================
void glutMasterDisp()
{
  // What: The master must display the current view.
  
  // Tell the slaves to display and compose.
  //
  int buffer[CMD_BUFFER_MAX_SIZE];  
  buffer[0] = CMD_DRAW;

  alog( << "s bcast cmd draw " << endl);
  MPI_Bcast(buffer, CMD_BUFFER_MAX_SIZE, MPI_INT, 0, worldComm);
  alog( << "e bcast cmd draw " << endl);
  
  // The master displays itself (if masterIsSlave) and compose.

  timestampe(ntime, tts, tte)
  selfDisplay();
  timestampe(ttime, tts, tte)

  // Frames per second.
  ++fpsFrames;
  fpsTime = glutGet(GLUT_ELAPSED_TIME);
  if ( fpsTime - fpsTimeBase > 1000 ) 
    {
      sprintf(fpsString, "%4.2f fps", 
	      fpsFrames * 1000.0 / (fpsTime - fpsTimeBase));
      fpsTimeBase = fpsTime;		
      fpsFrames = 0;	
    }

  if (logframes) {
    double fps = 1.0/ttime;
    if (fps < 100.0) fprintf(ffile, "D a= %g r= %g s= %g c= %g d= %g t= %g\n", atime, rtime, srttime, cotime, dtime, ttime);
  }

}
//=================================================================
//=================================================================

//================================================== Display
//==========================================================
float display(int pick)
{
  glPushMatrix(); {
    
    if (!pick) {

      //just adjust the diagnostic images to account for the fact that if !mis then we have an extra
      if (
	  (masterIsSlave && ((diag_show_node == worldSize) || (diag_show_node == worldId)))
	 ||
	  (!masterIsSlave && ((diag_show_node == worldSize) || (diag_show_node == (worldId-1))))
	  )
	{

	gluvvPrimitive *tmp;

	if (diag_show_objs_inside) {

	  //draw inside each subvolume
	  
	  if (diag_show_subvol_bbox) myVolRen->draw_bbox(true);
	  
	  if (myVolRen) myVolRen->clip_bbox(true);	
	  
	  tmp = renderables.getNext();	  
	  while(tmp) { 
	    tmp->draw();
	    tmp = tmp->getNext();

	  }
	  
	  if(gluvv.reblend == GB_UNDER){ //have to do this for front to back volumes
	    tmp = renderables.getNext();
	    while(tmp){
	      tmp->draw();
	      tmp = tmp->getNext();
	    }

	    glDisable(GL_BLEND);
	  }
	  
	  gluvv.reblend = GB_NONE;
	  
	  glDisable(GL_CLIP_PLANE0);
	  glDisable(GL_CLIP_PLANE1);
	  glDisable(GL_CLIP_PLANE2);
	  glDisable(GL_CLIP_PLANE3);
	  glDisable(GL_CLIP_PLANE4);
	  glDisable(GL_CLIP_PLANE5);
	}

	if (diag_show_objs_infront) {
	  
	  //draw in front of everything
	  
	  glDepthFunc(GL_LEQUAL);
	  
	  glClear( GL_DEPTH_BUFFER_BIT );
	  
	  if (!diag_show_vol_bbox) 
	    glColorMask(false,false,false,false); 
	  
	  if (myVolRen) myVolRen->draw_bbox(false); //this needs to be entire volume bbox
	  
	  glColorMask(true,true,true,true);
	  
	  gluvv.reblend = GB_ZERO;
	  
	  tmp = renderables.getNext();	  
	  while(tmp){
	    if (tmp != myVolRen) tmp->draw();
	    tmp = tmp->getNext();
	  }
	}
      }

    } else {
      
      glDepthFunc(GL_LEQUAL);
      
      glClear( GL_DEPTH_BUFFER_BIT );
      
      glColorMask(false,false,false,false); 
	        
      gluvvPrimitive *tmp = renderables.getNext();	  
      while(tmp){
	if (tmp != myVolRen) tmp->draw();
	tmp = tmp->getNext();
      }

      glColorMask(true, true, true, true); 

    }
  } glPopMatrix();

  if (!masterIsSlave && worldId==0) return 0;
  return myVolRen->getZVal();
}
//=================================================================
//=================================================================

//================================================== Mouse
//========================================================
void selfMouse(int button, int state, int x, int y)
{
  gluvv.mouse.button = button;
  gluvv.mouse.state  = state;

  if(state == GLUT_DOWN){  //try picking something
    gluvv.pick.pos[0] = x;
    gluvv.pick.pos[1] = gluvv.win.height - y;
    if (pick()){
      //aprint( << worldId << " Pick succeeded." << endl);
      if(gluvv.pick.prim){
	if ( gluvv.pick.prim->pick() &&
	     gluvv.pick.prim->mouse(button, state, x, y) )
	  return;
      }
    } else {
      //aprint( << worldId << " Pick failed." << endl);
    }
  }
  
  if(state == GLUT_UP){ //send pick events & reset title
    //cerr << worldId << " UP succeeded." << endl;
    if(gluvv.pick.prim){
      //cerr << worldId << " have a prim." << endl;
      gluvv.pick.prim->mouse(button, state, x, y);
      //cerr << worldId << " done with up." << endl;
      gluvv.mouse.button = -1;
      return;
    }
  }


  //standard button actions...
  switch(button){//////////////////////////////
  case GLUT_LEFT_BUTTON: //+++++++++++++++++++++++++++++++++
    switch(state){//)))))))))))))))))))
    case GLUT_UP:
      gluvv.volren.sampleRate = gluvv.volren.goodSamp;
      PglutPostRedisplay();
      break;
    case GLUT_DOWN:
      if(gluvv.mouse.ctrl && gluvv.mouse.shift){
	glutSetWindowTitle("Rotate **LIGHT**");
	lball.start((2.0 * x - gluvv.win.width) / gluvv.win.width,
		    (2.0 * y - gluvv.win.height) / gluvv.win.height);
	gluvv.mouse.pos[0] = gluvv.mouse.last[0] = x;
	gluvv.mouse.pos[1] = gluvv.mouse.last[1] = y;
	gluvv.mouse.ctrl = 1;
	gluvv.mouse.shift= 1;
      } else {
	glutSetWindowTitle("Rotate");
	tball.start((2.0 * x - gluvv.win.width) / gluvv.win.width,
		    (2.0 * y - gluvv.win.height) / gluvv.win.height);
	gluvv.mouse.pos[0] = gluvv.mouse.last[0] = x;
	gluvv.mouse.pos[1] = gluvv.mouse.last[1] = y;
	gluvv.volren.sampleRate = gluvv.volren.interactSamp;
      }
      PglutPostRedisplay();
      break;
    }//)))))))))))))))))))))))))))))))))
    break; //++++++++++++++++++++++++++++++++++++++++++++++
  case GLUT_RIGHT_BUTTON: //+++++++++++++++++++++++++++++
    if(gluvv.mouse.ctrl){ //simulate a middle button
      gluvv.mouse.button = GLUT_MIDDLE_BUTTON;
      switch(state){//))))))))))))))))))))))
      case GLUT_UP:
	gluvv.volren.sampleRate = gluvv.volren.goodSamp;
	PglutPostRedisplay();
	break;
      case GLUT_DOWN:
	glutSetWindowTitle("Translate");
	gluvv.mouse.pos[0] = gluvv.mouse.last[0] = x;
	gluvv.mouse.pos[1] = gluvv.mouse.last[1] = y;
	gluvv.volren.sampleRate = gluvv.volren.interactSamp;
	PglutPostRedisplay();
	break;
      }//))))))))))))))))))))))))))))))))))
    }
    else {
      switch(state){//))))))))))))))))))))
      case GLUT_UP:
	gluvv.volren.sampleRate = gluvv.volren.goodSamp;
	PglutPostRedisplay();
	break;
      case GLUT_DOWN:
	glutSetWindowTitle("Zoom");
	gluvv.mouse.pos[0] = gluvv.mouse.last[0] = x;
	gluvv.mouse.pos[1] = gluvv.mouse.last[1] = y;
	gluvv.volren.sampleRate = gluvv.volren.interactSamp;
	PglutPostRedisplay();
	break;
      }//))))))))))))))))))))))))))))))))
    }
    break;//++++++++++++++++++++++++++++++++++++++++++++
  case GLUT_MIDDLE_BUTTON://++++++++++++++++++++++++++
    switch(state){//)))))))))))))))))))))))
    case GLUT_UP:
      gluvv.volren.sampleRate = gluvv.volren.goodSamp;
      PglutPostRedisplay();
      break;
    case GLUT_DOWN:
      glutSetWindowTitle("Translate");
      gluvv.mouse.pos[0] = gluvv.mouse.last[0] = x;
      gluvv.mouse.pos[1] = gluvv.mouse.last[1] = y;
      gluvv.volren.sampleRate = gluvv.volren.interactSamp;
      PglutPostRedisplay();
      break;
    }//))))))))))))))))))))))))))))))))))))
    break;//+++++++++++++++++++++++++++++++++++++++++
  default:
    break;
  }////////////////////////////////////////////
}
//=================================================================
//=================================================================

//================================================== Motion
//=========================================================
void selfMotion(int x, int y)
{
  float dx, dy;
	
  if(gluvv.pick.prim){ //send events to a picked object
    //cerr << worldId << " Motion pick succeeded." << endl;
    if(gluvv.pick.prim->move(x, y)){
      //sync gluvv.volren.gamma
      //     gluvv.volren.loadTLUT = 1;
      return;
    }
  } else {	
    //cerr << worldId << " Motion pick failed." << endl;
  }
  
  switch(gluvv.mouse.button){
  case GLUT_LEFT_BUTTON: 
    switch(gluvv.mouse.state){
    case GLUT_DOWN:
      if(gluvv.mouse.ctrl && gluvv.mouse.shift){
	lball.update((2.0 * x - gluvv.win.width) /gluvv.win.width,
		     (2.0 * y - gluvv.win.height) /gluvv.win.height);
	float lxf[16];
	lball.buildRotMatrix((float (*)[4])lxf);
	translateV3(gluvv.light.pos, lxf, gluvv.light.startpos);
	glLightfv(GL_LIGHT0, GL_POSITION, gluvv.light.pos);
      } else {
	tball.update((2.0 * x - gluvv.win.width) /gluvv.win.width,
		     (2.0 * y - gluvv.win.height) /gluvv.win.height);
	tball.buildRotMatrix((float (*)[4])gluvv.rinfo.xform);
      }
      gluvv.mouse.last[0] = x;
      gluvv.mouse.last[1] = y;
      PglutPostRedisplay();
      break;
    case GLUT_UP: 
      break;
    }/////////////////////////////////////
    break; //+++++++++++++++++++++++++++++++
  case GLUT_RIGHT_BUTTON://+++++++++++++++
    switch(gluvv.mouse.state){////////////
    case GLUT_DOWN:
      dx = (x - gluvv.mouse.last[0])/(float)gluvv.win.width;
      dy = (y - gluvv.mouse.last[1])/(float)gluvv.win.height;
	  
      // XC FIXME: (int) cast to check (was no cast)
      if ( abs((int)dy) > abs((int)dx) )
	{ 
	  //scale object
	  gluvv.rinfo.scale += 5.0 * dy;
	  if ( gluvv.rinfo.scale < .05 ) gluvv.rinfo.scale = .05;
	  if ( gluvv.rinfo.scale > 10.0 ) gluvv.rinfo.scale = 10;
	} 
      else 
	{ 
	  //zoom in
	  gluvv.env.frustum[0] *= 1+dx;
	  gluvv.env.frustum[1] *= 1+dx;
	  gluvv.env.frustum[2] *= 1+dx;
	  gluvv.env.frustum[3] *= 1+dx;
	}
	  
      gluvv.mouse.last[0] = x;
      gluvv.mouse.last[1] = y;
      PglutPostRedisplay();
      break;
    case GLUT_UP: 
      break;
    }/////////////////////////////////////
    break; //++++++++++++++++++++++++++++++
  case GLUT_MIDDLE_BUTTON://+++++++++++++++
    switch ( gluvv.mouse.state )
      {////////////
      case GLUT_DOWN:
	dx = (x - gluvv.mouse.last[0])/(float)gluvv.win.width;
	dy = (y - gluvv.mouse.last[1])/(float)gluvv.win.height;
	gluvv.rinfo.trans[0] += dx *
	  (gluvv.env.frustum[1]-gluvv.env.frustum[0]) * 
	  gluvv.env.eye[2]/gluvv.env.clip[0];
	gluvv.rinfo.trans[1] += dy *
	  (gluvv.env.frustum[3]-gluvv.env.frustum[2]) * 
	  gluvv.env.eye[2]/gluvv.env.clip[0];
	gluvv.mouse.last[0] = x;
	gluvv.mouse.last[1] = y;
	PglutPostRedisplay();
	break;
      case GLUT_UP: 
	break;
      }/////////////////////////////////////
    break;//+++++++++++++++++++++++++++++++
      
  default:
    break;
  }//////////////////////////////////////////////
}
//=================================================================
//=================================================================

//================================================== GLUT key
//===========================================================
void glutMasterKey(unsigned char key, int x, int y)
{

  int buffer[CMD_BUFFER_MAX_SIZE];
  buffer[0] = CMD_KEY;
  buffer[1] = key;
  buffer[3] = x;
  buffer[4] = y;
  alog( << "s bcast key " << endl);
  MPI_Bcast(buffer, CMD_BUFFER_MAX_SIZE, MPI_INT, 0, worldComm);
  alog( << "e bcast key " << endl);

  selfKey(key, x, y);
}


void selfKey(unsigned char key, int x, int y) 
{
  if ( key == 27 )
    { 
      // excape key	
      int buffer[CMD_BUFFER_MAX_SIZE];
      buffer[0] = CMD_QUIT;
      MPI_Bcast(buffer, CMD_BUFFER_MAX_SIZE, MPI_INT, 0, worldComm);

      if (worldId == 0) {
	if (record) {
	  fprintf(lfile, "Q -1.0 -1.0 -1 -1\n");
	}
	if (record || playback) fclose(lfile);
	if (logframes) fclose(ffile);
      }

      finalExit();
      gluvvQuit();
    }
  
  if(gluvv.pick.prim)
    if(gluvv.pick.prim->key(key, x, y)) 
      return;

  switch(key){ ///////////////////////////////////
  case '2':
    diag_show_node++;
    if (masterIsSlave) {
      if (diag_show_node > worldSize) diag_show_node = 0;
      if (diag_show_node == worldSize) {
	aprint( << "drawing the whole volume" << endl);
      } else {
	aprint( << "drawing node " << diag_show_node << "'s data " << endl);
      }
    } else {
      if (diag_show_node == (worldSize-1)) diag_show_node++; //ick, 1 dead space in there
      if (diag_show_node > worldSize) diag_show_node = 0;
      if (diag_show_node == worldSize) {
	aprint( << "drawing the whole volume" << endl);
      } else {
	aprint( << "drawing node " << diag_show_node+1 << "'s data " << endl);
      }
    }
    
    PglutPostRedisplay();
    break;
  case '3':
    diag_show_subvol_bbox = !diag_show_subvol_bbox;
    PglutPostRedisplay();
    aprint( << "draw sub volume bbox now " << diag_show_subvol_bbox << endl);
    break;
  case '4':
    diag_show_vol_bbox = !diag_show_vol_bbox;
    PglutPostRedisplay();
    aprint( << "draw total volume bbox now " << diag_show_vol_bbox << endl);
    break;
  case '5':
    diag_show_objs_behind = !diag_show_objs_behind;
    PglutPostRedisplay();
    aprint( << "draw behind volume now " << diag_show_objs_behind << endl);
    break;
  case '6':
    diag_show_objs_inside = !diag_show_objs_inside;
    PglutPostRedisplay();
    aprint( << "draw inside volume now " << diag_show_objs_inside << endl);
    break;
  case '7':
    diag_show_objs_infront = !diag_show_objs_infront;
    PglutPostRedisplay();
    aprint( << "draw in front of volume now " << diag_show_objs_infront << endl);
    break;
  case 'm':
    if (worldId == 0) {
      if (gui.visible()) 
	gui.hide();
      else 
	gui.show();
    }
    break;
  case '.': //increase interactive sample rate
    gluvv.volren.interactSamp *= 1/.95;
    aprint( << "*Interactive Sample Rate = " << gluvv.volren.interactSamp << endl);
    break;
  case '<': //decrease good sample rate
    gluvv.volren.goodSamp *= .95;
    aprint( << "*High Quality Sample Rate = " << gluvv.volren.goodSamp << endl);
    break;
  case '>': //increase good sample rate
    gluvv.volren.goodSamp *= 1/.95;
    aprint( << "*High Quality Sample Rate = " << gluvv.volren.goodSamp << endl);
    break;
  case 'p': //enable/disable data probe
    {
    int probeInsideVolume = 0;
    if(gluvv.tf.brushon){
      float*   bot = gluvv.tf.tfWidget->getBrush()->getBottom();
      float*  left = gluvv.tf.tfWidget->getBrush()->getLeft();
      float* right = gluvv.tf.tfWidget->getBrush()->getRight();
      if (bot[0] == 0 && bot[1] == 0 &&
	    left[0] == 0 && left[1] == 0 &&
	    right[0] == 0 && right[1] == 0){
	probeInsideVolume = 0;
	//cout<<"The TFWidget is not being rendered"<<endl;
      }
      else{
	probeInsideVolume = 1;
      }    
    }
    
    if(gluvv.probe.on){
      // take probe out of TF
      gluvv.probe.on = 0;
      if (probeInsideVolume) gluvv.tf.loadme = 1;
    }
    else {
      // put probe back into TF
      gluvv.probe.on = 1;
      if (probeInsideVolume) gluvv.tf.loadme = 1;
    }
    PglutPostRedisplay();
    }
    break;
     case 'P':
    if (worldId == 0){
      //unsigned int FrameBuffer[gluvv.win.width*gluvv.win.height];
      unsigned int* FrameBuffer;
      //glReadBuffer(GL_BACK); //glReadBuffer(GL_BACK);
      cout<<"starting screen capture..."<<endl;
      glPixelStorei(GL_UNPACK_ALIGNMENT, 8);

      FrameBuffer = new unsigned int[gluvv.win.width*gluvv.win.height];
    
      glReadPixels(0, 0, gluvv.win.width, gluvv.win.height,
		   GL_RGBA, GL_UNSIGNED_INT_8_8_8_8, FrameBuffer);

      char tmp[256];
      sprintf(tmp, "FrameBuffer%d.ppm", 1);
      save_ppm(tmp, gluvv.win.width, gluvv.win.height, 
	       4, (const unsigned char *)(FrameBuffer));
      delete FrameBuffer;
      //sprintf(tmp, "FrameBuffer-alpha%d.ppm", 1);
      //save_ppm_alpha(tmp, gluvv.win.width, gluvv.win.height,
      //		   4, (const unsigned char *)(FrameBuffer));
    }
    break;
  case 'c': //enable/disable clipping plane
    {
    int probeInsideVolume = 0;
    if(gluvv.tf.brushon){
      float*   bot = gluvv.tf.tfWidget->getBrush()->getBottom();
      float*  left = gluvv.tf.tfWidget->getBrush()->getLeft();
      float* right = gluvv.tf.tfWidget->getBrush()->getRight();
      if (bot[0] == 0 && bot[1] == 0 &&
	    left[0] == 0 && left[1] == 0 &&
	    right[0] == 0 && right[1] == 0){
	probeInsideVolume = 0;
	//cout<<"The TFWidget is not being rendered"<<endl;
      }
      else{
	probeInsideVolume = 1;
      }
    }
    
    if(gluvv.clip.on){
      // take probe out of TF
      gluvv.clip.on = 0;
      if (probeInsideVolume) gluvv.tf.loadme = 1;
    }
    else{
      // put probe back into TF
      gluvv.clip.on = 1;
      if (probeInsideVolume) gluvv.tf.loadme = 1;
    }
    PglutPostRedisplay();
    }
    break;
  case 'k':
    gluvv.tf.brushon = 0;
    gluvv.tf.loadme = 1; 
    PglutPostRedisplay();
    break;
 case 'B':
    gluvv.env.bgColor = !gluvv.env.bgColor;
    PglutPostRedisplay();
    break;    
  case 'b': //enable/disable transfer function brush		
    if(gluvv.tf.brushon && 
       (gluvv.probe.brush == EllipseBrush)) 
      gluvv.tf.brushon = 0;
    else gluvv.tf.brushon = 1;
    gluvv.probe.brush = EllipseBrush;
    gluvv.tf.loadme = 1; 
    PglutPostRedisplay();
    break;
  case 't': //automatically sized triangle brush
    if(gluvv.tf.brushon && 
       (gluvv.probe.brush == TriangleBrush)) 
      gluvv.tf.brushon = 0;
    else gluvv.tf.brushon = 1;
    gluvv.probe.slider = .5;
    gluvv.probe.brush = TriangleBrush;
    gluvv.tf.loadme = 1; 
    PglutPostRedisplay();
    break;
  case 'a': //automatically sized ellipse brush
    if(gluvv.tf.brushon && 
       (gluvv.probe.brush == AutoEllipseBrush)) 
      gluvv.tf.brushon = 0;
    else gluvv.tf.brushon = 1;
    gluvv.probe.slider = .5;
    gluvv.probe.brush = AutoEllipseBrush;
    gluvv.tf.loadme = 1; 
    PglutPostRedisplay();
    break;
  case 'o': //auto one d brush
    if(gluvv.tf.brushon && 
       (gluvv.probe.brush == AutoOneDBrush)) 
      gluvv.tf.brushon = 0;
    else gluvv.tf.brushon = 1;
    gluvv.probe.slider = .5;
    gluvv.probe.brush = AutoOneDBrush;
    gluvv.tf.loadme = 1; 
    PglutPostRedisplay();
    break;
  case 'O': //one d brush
    if(gluvv.tf.brushon && 
       ((gluvv.probe.brush == AutoOneDBrush)||
	(gluvv.probe.brush = OneDBrush))) 
      gluvv.tf.brushon = 0;
    else gluvv.tf.brushon = 1;
    gluvv.probe.slider = .5;
    gluvv.probe.brush = OneDBrush;
    gluvv.tf.loadme = 1; 
    PglutPostRedisplay();
    break;
  case 'h': //toggle histogram
    if(gluvv.tf.histOn) gluvv.tf.histOn = 0;
    else gluvv.tf.histOn = 1;
    PglutPostRedisplay();
    break;
  case 'w':
    if (worldId == 0){
      cerr << "Please input a file name: ";
      char name[256];
      cin>>name;
      save(name);
    }
    break;
  case 'W':
    if (worldId == 0){
      cerr<<"Please input a file name: ";
      char name[256];
      cin>>name;
      gluvv.mv->writeAll("hartner_bricks", gluvv.mv->nelts );
    }
    break;
  case 'f':
    gluvv.front2backSlicing = 1;
    PglutPostRedisplay();
    break;
  case 'F':
    gluvv.front2backSlicing = 0;
    PglutPostRedisplay();
    break;
  case 's': //toggle shade modes
  case 'S':
    if(gluvv.shade == gluvvShadeAmb) gluvv.shade = gluvvShadeDiff;
    else if(gluvv.shade == gluvvShadeDiff) gluvv.shade = gluvvShadeDSpec;
    else if(gluvv.shade == gluvvShadeDSpec) gluvv.shade = gluvvShadeFaux;
    else if(gluvv.shade == gluvvShadeFaux) gluvv.shade = gluvvShadeAmb;
    gluvv.tf.loadme = 1;
    PglutPostRedisplay();
    break;
  case ' ': //paint action
    gluvv.tf.paintme = 1;
    PglutPostRedisplay();
    break;
  case 'd': //drop action
    gluvv.tf.dropme = 1;
    PglutPostRedisplay();
    break;
  case 'n': //clear painted tf
    gluvv.tf.clearpaint = 1;
    PglutPostRedisplay();
    break;
  case 'l': //toggle light widget on and off
    //#ifdef WIN32
    ltwr.onoff();
    PglutPostRedisplay();
    //#endif
    break;
  case '-':
  case '_': //change the current timestep - down
    aprint( << "-_ TODO\n");
    /*
      gluvv.volren.timestep -= 1;
      if(gluvv.volren.timestep < gluvv.mv->tstart){
      gluvv.volren.timestep = gluvv.mv->tstart;
      cerr << "We are at the first time step :" << gluvv.volren.timestep << endl;
      } else {
      cerr << "Current time step: " << gluvv.volren.timestep << endl;
      if(!gluvv.mv->swapTStep(gluvv.volren.timestep)){
      gluvv.mv->readAll(gluvv.volren.timestep);
      if(gluvv.mv1)
      gluvv.mv1->readAll(gluvv.volren.timestep);
      if(gluvv.mv2)
      gluvv.mv2->readAll(gluvv.volren.timestep);
      if(gluvv.mv3)
      gluvv.mv3->readAll(gluvv.volren.timestep);
      initData();
      gluvv.mv->cacheTStep();
      }
      }
      PglutPostRedisplay();
    */
    break;
  case '=':
  case '+': //change the current timesetp - up
    cerr << "=+ TODO\n";
    
    /*
      gluvv.volren.timestep += 1;
      if(gluvv.volren.timestep > gluvv.mv->tstop){
      gluvv.volren.timestep = gluvv.mv->tstop;				
      cerr << "We are at the last time step :" << gluvv.volren.timestep << endl;
      } else {
      cerr << "Current time step: " << gluvv.volren.timestep << endl;
      if(!gluvv.mv->swapTStep(gluvv.volren.timestep)){
      gluvv.mv->readAll(gluvv.volren.timestep);
      if(gluvv.mv1)
      gluvv.mv1->readAll(gluvv.volren.timestep);
      if(gluvv.mv2)
      gluvv.mv2->readAll(gluvv.volren.timestep);
      if(gluvv.mv3)
      gluvv.mv3->readAll(gluvv.volren.timestep);
      initData();
      gluvv.mv->cacheTStep();
      }
      }
      PglutPostRedisplay();
    */
    break;
  case 'x':
    if ( continuousDisplay ) 
      {
	continuousDisplay = 0;
	aprint( << "Continuous display is OFF" << endl);
      }
    else 
      {
	continuousDisplay = 1;
	aprint( << "Continuous display is ON" << endl);
      }
    break;
  default:
    aprint( << "Unused Key press: '" << (char)key << "' = " << (int) key << " at " << x << ", " << y << endl);
    break;
  }
}

void glutMasterSpecial(int key, int x, int y)
{

  int buffer[CMD_BUFFER_MAX_SIZE];
  buffer[0] = CMD_SPECIAL;
  buffer[1] = key;
  buffer[3] = x;
  buffer[4] = y;
  alog( << "s bcast special " << endl);
  MPI_Bcast(buffer, CMD_BUFFER_MAX_SIZE, MPI_INT, 0, worldComm);
  alog( << "e bcast special " << endl);

  selfSpecial(key, x, y);
}

void selfSpecial(int key, int x, int y)
{
  if(gluvv.pick.prim)
    if(gluvv.pick.prim->special(key, x, y)) 
      return;
  
  switch(key){
  case GLUT_KEY_F1:  //open tf window
    break;

  case GLUT_KEY_F9:  //ambiant shade mode
    gluvv.shade = gluvvShadeAmb;
    gluvv.tf.loadme = 1;
    PglutPostRedisplay();
    break;

  case GLUT_KEY_F10: //shaded
    gluvv.shade = gluvvShadeDSpec;
    //gluvv.tf.loadme = 1;
    PglutPostRedisplay();
    break;

  case GLUT_KEY_F11: //faux shaded
    gluvv.shade = gluvvShadeFaux;
    gluvv.tf.loadme = 1;
    PglutPostRedisplay();
    break;

  case GLUT_KEY_F12: //????
    gluvv.shade = gluvvShadeDiff;
    //gluvv.tf.loadme = 1;
    PglutPostRedisplay();
    break;

  default:
    break;
  }
}

//=================================================================
//=================================================================

//================================================== GLUT idle
//============================================================
void glutMasterIdle(void)
{  
  // What: The master gets bored with nothing to do.
  
  if ( glutGetWindow() != gluvv.mainWindow ) 
    {
      glutSetWindow(gluvv.mainWindow);	
    }
  
  if ( continuousDisplay ) 
    {
      glutPostRedisplay();
    }

  if ( record ) {
    fprintf(lfile, "I -1.0 -1.0 -1 -1\n");
  }
  if ( playback ) {
    timestampe(ntime, ats, ate);    
    char cchar;
    float f1, f2;
    int i1, i2;
    fscanf(lfile, "%c %f %f %d %d\n", &cchar, &f1, &f2, &i1, &i2);
    timestampe(atime, ats, ate);

    switch (cchar) {
    case 'I':
      break;
    case 'B':
      glutMasterMouse(i1, i2, (int)(f1*SCREEN_RES), (int)(f2*SCREEN_RES));
      break;
    case 'M':
      glutMasterMotion((int)(f1*SCREEN_RES), (int)(f2*SCREEN_RES));
      break;
    case 'Q':
      fclose(lfile);
      if (logframes) fclose(ffile);
      cerr << "Read Quit." << endl;
      glutMasterKey((unsigned char)27, 0, 0);
      break;
    default:
      cerr << "PLAYBACK ERROR\n";
      exit(-1);
    }
  }
}

void glutSlaveIdle(void)
{  
#ifdef MPE_LOGGING
  MPE_Log_event(MPE_EVENT_BCAST_B, 0, "broadcast");
#endif

  int buffer[CMD_BUFFER_MAX_SIZE];  
  alog( << "s bcast slave idle " << endl);
  MPI_Bcast(buffer, CMD_BUFFER_MAX_SIZE, MPI_INT, 0, worldComm);
  alog( << "e bcast slave idle " << endl);
  
#ifdef MPE_LOGGING
  MPE_Log_event(MPE_EVENT_BCAST_E, 0, "broadcasted");
#endif

  switch ( buffer[0] )
    {
    case CMD_DRAW:
      synch_dont_redraw = false;
      glutPostRedisplay();
      break;
    case CMD_RESHAPE:
      //cerr << worldId << " GOT RESHAPE " << endl;
      selfReshape(buffer[1], buffer[2]);
      break;
    case CMD_MOUSE:
      gluvv.mouse.ctrl = buffer[5];
      gluvv.mouse.shift = buffer[6];
      gluvv.mouse.alt = buffer[7];
      selfMouse(buffer[1], buffer[2], buffer[3], buffer[4]);
      break;
    case CMD_MOTION:
      selfMotion(buffer[1], buffer[2]);
      break;
    case CMD_KEY:
      selfKey((unsigned char)buffer[1], buffer[2], buffer[3]);
      break;
    case CMD_SPECIAL:
      selfSpecial(buffer[1], buffer[2], buffer[3]);
      break;
    case CMD_MASTERONLY:
      break;
    case CMD_GUI: 
      {
	int control = buffer[1];
	int ival0 = buffer[2];
	float fval0 = *((float*)(&buffer[3]));
	gui.remote_control_cb(control, ival0, fval0);
      }
      break;
    default:
      // Shouldn't go there.
      cerr << worldId << " Unknown command #" << buffer[0] << endl;
      break;
    }
}

//================================================== Reshape
//==========================================================

void innerReshape(int w, int h)
{
  glViewport(0,0, (GLint) w, (GLint) h);
  //alog( << worldId << " innerReshape to " << w << "x" << h << endl);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluvv.env.frustum[0] *= (float)w/(float)gluvv.win.width;
  gluvv.env.frustum[1] *= (float)w/(float)gluvv.win.width;
  gluvv.env.frustum[2] *= (float)h/(float)gluvv.win.height;
  gluvv.env.frustum[3] *= (float)h/(float)gluvv.win.height;
  
  glFrustum(gluvv.env.frustum[0], 
	    gluvv.env.frustum[1], 
	    gluvv.env.frustum[2], 
	    gluvv.env.frustum[3], 
	    gluvv.env.clip[0], 
	    gluvv.env.clip[1]);
  
  glMatrixMode(GL_MODELVIEW);
  
  gluvv.win.width  = w;
  gluvv.win.height = h;
  
  mySortLast->Resize(w,h);

  alog( << "s barrier inner reshape " << endl);
  MPI_Barrier(worldComm);
  alog( << "e barrier inner reshape " << endl);
}

void selfReshape(int w, int h)
{
  if ((w != oldW) || (h != oldH)) {
    oldW = w;
    oldH = h;
    
    innerReshape(w,h);

    if (!masterIsSlave || worldId) {
      //alog(<<"GROWING WINDOW" << endl);
      glutReshapeWindow(w,h);
    }
  }
  alog( << "s barrier self reshape " << endl);
  MPI_Barrier(worldComm);
  alog( << "e barrier self reshape " << endl);
}

//=================================================================
//=================================================================


//=================================================== Pick
//========================================================
//================================================== begin
//--- lesson in picking ---------------------------------------
// rules for gluvv picking say that first you push your 
//  'this' pointer (1 int for 32 bit, 2 for 64) 
//	 -(must be a subclass of gluvvPrimitive!)
//  finaly push three (3) ints on the stack 
//	 -(sub object names - anything you want)
// to recap: you push 4 ints (total) in 32 bit, and 5 (total)in 64
//--------------------------------------------------------------

int pick()
{
  gluvv.pick.data1  = 0;
  gluvv.pick.data2  = 0;
  gluvv.pick.name32 = 0;
  gluvv.pick.name64 = 0;
  gluvv.pick.z	    = 0xffffffff;
  if(gluvv.pick.prim){
    gluvv.pick.prim->release();
    gluvv.pick.prim   = 0;
  }
	
  GLint viewport[4];
  glGetIntegerv(GL_VIEWPORT, viewport);
	
  GLfloat projection[16];
  glGetFloatv(GL_PROJECTION_MATRIX, projection);
  
  int pickBufferSize = 128;
  GLuint *pickBuffer = new GLuint[pickBufferSize];
	
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  {
    glLoadIdentity();
    GLdouble pickbox = 4;
    gluPickMatrix((GLdouble)gluvv.pick.pos[0], (GLdouble)gluvv.pick.pos[1], 
		  pickbox, pickbox, viewport);
    
    glMultMatrixf(projection);
    glMatrixMode(GL_MODELVIEW);
    glSelectBuffer(pickBufferSize, pickBuffer);
    glRenderMode(GL_SELECT);
		
    glInitNames();
    glPushName(0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    
    // ---- RENDER -------------------------------
    //cycle through scene graph ------------------
    gluvv.picking = 1;
    display(1);
    gluvv.picking = 0;
  }
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
	
  glFlush();
	
  // now see what we got!
  GLint hits = glRenderMode(GL_RENDER);
  //cerr << "hits = " << hits;
  GLuint nnames, z, minz = 0xffffffff;
  int gotone = 0;
	
  if(hits > 0){
    
    //i tracks the number of hits, j tracks postion in pick buffer
    int i, j; 
    for(i=0, j=0; (j<pickBufferSize && i<hits); i++){
      //first item on stack is the number of items
      nnames = pickBuffer[j++]; 
      z = pickBuffer[j++];	//second is the z value for this hit
      if(nnames > 1 && z <= minz){
	gotone = 1;
	minz= z;
	gluvv.pick.z = minz;
	
#if (_MIPS_SZPTR == 64) //64 bit
	j += nnames - 5;
	//reconstruct your 'this' pointer
	unsigned long o1 = pickBuffer[j++];
	unsigned long o2 = pickBuffer[j++];
	gluvv.pick.name64     = (o1<<32)|o2;
	gluvv.pick.data1      = pickBuffer[j++];
	gluvv.pick.data2      = pickBuffer[j++];
	gluvv.pick.data3      = pickBuffer[j++];
	gluvv.pick.prim	      = (gluvvPrimitive*)gluvv.pick.name64;
#else
	j += nnames - 3;
	gluvv.pick.name32 = pickBuffer[j++];
	gluvv.pick.data1  = pickBuffer[j++];
	gluvv.pick.data2  = pickBuffer[j++];
	gluvv.pick.data3  = pickBuffer[j++];
	gluvv.pick.prim	  = (gluvvPrimitive*)gluvv.pick.name32;
#endif
	//cerr << "here: " << gluvv.pick.name32	 << " " << gluvv.pick.data1 
	//  << " " << gluvv.pick.data2
	// << " " << gluvv.pick.data3 << endl;
      }
      else{
	j += nnames + 1;
      }
    }
  }
	
  
  delete[] pickBuffer;
	
  if (gotone) return 1;
	
  return 0;
  
}
//=================================================== Pick
//========================================================
//==================================================== end

//======================================= check extensions
//========================================================
int checkExts(void)
{
  //use this to check if we can support the things we want to do
  // in the future this should check to see what render modes
  //  that we can use.
  // Many thanks to Nate Robins for testing this on his GF2!!!!!
  
  gluvv.plat = (gluvvPlatform) CardProfile();
  
  if( glTexImage3D == 0 )
    return 0; //error
  if( glActiveTextureARB == 0)
    return 0; //error
  // if ( ! glCombinerParameterfvNV )
  //return 0;
  
  return 1;
}
//=================================== end check extensions
//========================================================

//============================================== init Data
//========================================================
int initData(void)
{
  int ret = 0;
  
  cerr << "Initializing Data " << worldId << endl;
  switch (gluvv.dmode){
  case GDM_V1:
  case GDM_V1G:
  case GDM_V2:
  case GDM_V2G:
  case GDM_V3:
  case GDM_V3G:
  case GDM_V4:
    ret += gluvv.mv->mergeMV(blurnorms,1,0,gluvv.mv1, gluvv.mv2, gluvv.mv3);
    break;
  case GDM_V1GH:
  case GDM_V2GH:
    ret += gluvv.mv->mergeMV(blurnorms,1,1,gluvv.mv1, gluvv.mv2, gluvv.mv3);
    break;
  case GDM_VGH:
  case GDM_VGH_VG:
  case GDM_VGH_V:
#ifdef PARALLEL
    if (gluvv.normalsRestored ){
      //do nothing, we read the normals from the TREX file
    }
    else{
#endif //PARALLEL
      gluvv.mv->normalsVGH(blurnorms,1);
#ifdef PARALLEL
    }
#endif
    break;
  }
  
  int bricksz = 128 * 128 * 256;
  
  switch (gluvv.dmode){
  case GDM_V1:
  case GDM_VGH_V:
    bricksz = 128 * 128 * 256;
    break;
  case GDM_V1G:
  case GDM_V2:
  case GDM_VGH_VG:
    bricksz = 128 * 128 * 256;
    break;
  default:
    bricksz = 128 * 128 * 256;
  }
  
  gluvv.mv->padOut();

#ifdef PARALLEL
  int dataSize = gluvv.mv->xiSize * gluvv.mv->yiSize * gluvv.mv->ziSize;
  bricksz = dataSize/worldSize;
#endif
  gluvv.mv->brick(bricksz);
  gluvv.mv->printInfo();
  
  aprint( << " -done with data initialization" << endl);
  return ret;
}
//=================================================================
//=================================================================

//================================================== Parse
//========================================================
int pre_parse(int argc, char *argv[])
{
  for ( int i = 0; i < argc; ++i )
    {
      if (!strcmp(argv[i], "-res"))
	{
	  ++i;
	  if(i>=argc){
	    cerr << "parse Error: -l specified without a file name \n";
	    return 1;
	  }
	  SCREEN_RES = atoi(argv[i]);
	}
    }
  
  return 0;
}

int parse(int argc, char *argv[])
{
  //Yeah yeah, I should use lex and yack for this, so shoot me...
  if ( argc < 3 ){
    cerr << argv[0] << "  Usage:\n";
    cerr << " File flags:\n";
    cerr << "	   <-addG>		   Add a gradient measure to N channel data\n";
    cerr << "	   <-addH>		   Add a hessian measure to N channel data\n";
    cerr << "	    -n	 [file name.nrrd]  Nrrd file\n";
    cerr << "	    -vgh [file name.nrrd]  VGH Nrrd file\n";
    cerr << "	   <-useVG>		   Only use V&G from VGH Nrrd file (2B)\n";
    cerr << "	   <-useV>		   Only use V from VGH Nrrd file (1B)\n";
    cerr << " Other options:\n";
    cerr << "	   <-use3D>		   Use a full 3D transfer function\n";
    cerr << "	   <-blurNormals>	   Blurr normals\n";	       
      
    MPI_Finalize();
    gluvvQuit();		
  }
  
  for ( int i = 0; i < argc; ++i )
    {
      //-------------- TF widget state load ------------------------------
      if(!strcmp(argv[i], "-l")){
	++i;
	if(i>=argc){
	  cerr << "parse Error: -l specified without a file name \n";
	  return 1;
	}
	strcpy(loadname, argv[i]);
	cerr << "got a simian file: " << loadname << endl;
	gluvv.tf.loadme    = 1;
      }
      
      //-------------- trex flags ------------------------------
      else if(!strcmp(argv[i], "-t")){ //parse a trex file**********************************
	++i;
	if(i>=argc){
	  cerr << "parse Error: -t specified without a file name \n";
	  return 1;
	}
	cerr << "  Loading TREX file " << argv[i] << endl;
	gluvv.dmode = GDM_VGH;
	gluvv.mv = new MetaVolume(argv[i]);
	gluvv.normalsRestored = 1;
	if (worldId || masterIsSlave)
	  if (!gluvv.mv->readAll(gluvv.mv->tstart)) return 1;
	gluvv.volren.timestep = gluvv.mv->tstart;
      }

      //----------------- Nrrd flags --------------------------
      // parse a nrrd file******************************
      else if ( ! strcmp(argv[i], "-n") )
	{ 
	  ++i;
	  if( i >= argc )
	    {
	      cerr << "parse Error: -n specified without a file name" << endl;
	      return 1;
	    }
	  
#ifdef DEBUG
	  cerr << "  Loading NRRD file " << argv[i] << endl;
#endif
	  
	  if ( worldId || masterIsSlave )
	    {
	      gluvv.mv = new MetaVolume(argv[i], 1);
	      if ( ! gluvv.mv->readNrrd(gluvv.debug) ) return 1;
	      
	      if      ( gluvv.mv->nelts == 1 ) gluvv.dmode = GDM_V1;
	      else if ( gluvv.mv->nelts == 2 ) gluvv.dmode = GDM_V2;
	      else if ( gluvv.mv->nelts == 3 ) gluvv.dmode = GDM_V3;
	      else if ( gluvv.mv->nelts == 4 ) gluvv.dmode = GDM_V4;
	    }
	}
      // parse an old school vgh file****************
      else if ( ! strcmp(argv[i], "-vgh") )
	{ 
	  ++i;
	  if ( i >= argc )
	    {
	      cerr << "parse Error: -vgh specified without a file name \n";
	      return 1;
	    }

#ifdef DEBUG
	  cerr << "  Loading VGH NRRD file " << argv[i] << endl;
#endif

	  int fileId = masterIsSlave ? worldId : worldId - 1;
	  
	  if ( worldId || masterIsSlave )
	    {
	      gluvv.mv = new MetaVolume(argv[i], 1);

	      if (worldId == 0){
		if ( ! gluvv.mv->readNrrd(gluvv.debug) ) return 1;
	      }
	      else{
		alog( << "s barrier load vgh nrrd " << endl);
		MPI_Barrier(slavesComm); 
		alog( << "e barrier load vgh nrrd " << endl);
		if ( ! gluvv.mv->readNrrd(gluvv.debug) ) return 1;
		
	      }
	      
	      if ( (gluvv.dmode != GDM_VGH_VG) && (gluvv.dmode != GDM_VGH_V) )
		{
		  gluvv.dmode = GDM_VGH;
		}

	      if ( gluvv.mv->nelts == 1 )
		{  
		  if ( gluvv.dmode == GDM_VGH_V )
		    {
		      gluvv.dmode = GDM_V1;
		    }
		  if ( gluvv.dmode == GDM_VGH_VG )
		    {
		      gluvv.dmode = GDM_V1G;
		    }
		  else
		    {
		      gluvv.dmode = GDM_V1GH;
		    }
		} 
	      else if ( gluvv.mv->nelts != 3 )
		{
		  //I cannot recover from this one, explode!!
		  cerr << "Error: -vgh flag is for nrrd volume with pre-computed VGH"
		       << endl;
		  return 1;
		}
	    }
	} 
      else if ( !strcmp(argv[i], "-useVG") )
	{ 
	  //only use VG from a VGH file*****
	  gluvv.dmode = GDM_VGH_VG;
	}
      else if ( !strcmp(argv[i], "-useV") )
	{ 
	  //only use VG from a VGH file******
	  gluvv.dmode = GDM_VGH_V;
	}
      // add a gradient comp************************
      else if ( ! strcmp(argv[i], "-addG") )
	{ 
	  switch (gluvv.dmode)
	    {
	    case GDM_V1:
	      gluvv.dmode = GDM_V1G;
	      break;
	    case GDM_V2:
	      gluvv.dmode = GDM_V2G;
	      break;
	    case GDM_V3:
	      gluvv.dmode = GDM_V3G;
	      break;
	    default:
	      cerr << "parse Error: you cannot add gradient measure to this data type" << endl;
	      return 1;
	    }
	}
      // add a hessian comp***************************
      else if ( ! strcmp(argv[i], "-addH") )
	{
	  switch (gluvv.dmode)
	    {
	    case GDM_V1G:
	      gluvv.dmode = GDM_V1GH;
	      break;
	    case GDM_V2G:
	      gluvv.dmode = GDM_V2GH;
	      break;
	    default:
	      cerr << "parse Error: you cannot add hessian measure to this data type"
		   << endl;
	      return 1;
	    }
	}
      //------------------ misc flags --------------------
      else if ( (!strcmp(argv[i], "-blurNormals")) ||
		(!strcmp(argv[i], "-blurnormals")) )
	{
	  //blur normals************
	  blurnorms = 1;
	}
      else if ( (!strcmp(argv[i], "-use3D")) ||
		(!strcmp(argv[i], "-use3D")) )
	{
	  //3D transfer function************
	  //gluvv.tf.use3D = 1;
	}
      else if ( !strcmp(argv[i],"-shadow") )
	{
	  gluvv.light.shadow = 1;
	}
      else if ( !strcmp(argv[i], "-d") )
	{ 
	  //Print some debug info***************
	  gluvv.debug = 1;
	}
      else if (!strcmp(argv[i], "-record")) 
	{
	  if (worldId==0) {
	    record = true; playback = false;
	    if(!(lfile = fopen("cfs.rec", "w"))){
	      printf("Problem opening log output file.\n");
	    }
	  }
        }
      else if (!strcmp(argv[i], "-playback")) 
	{
	  if (worldId==0) {
	    record = false; playback = true;
	    if(!(lfile = fopen("cfs.rec", "r"))){
	      printf("Problem opening log input file.\n");
	    }
	  }
        }
      else if (!strcmp(argv[i], "-logframes")) 
	{
	  if (worldId==0) {
	    logframes = true;
	    if(!(ffile = fopen("fps.log", "w"))){
	      printf("Problem opening framerate output file.\n");
	    }
	  }
        }
      else if (!strcmp(argv[i], "-res"))
	{
	  i++;
	}
    }
  
  return 0;
}

//=================================================================
//=================================================================

int	save(char *name)
{
  FILE *f;
  if(!(f = fopen(name, "w"))){
    cerr << "Error: Simian save, failed to open " 
	 << name << " for writing" << endl;
    return 0;
  }
  else {
    cerr << "Saving ...." << endl;
    gluvvPrimitive *tmp = renderables.getNext();
    while(tmp){
      tmp->save(f);
      tmp = tmp->getNext();
    }
    fclose(f);
    cerr << "Done Saving" << endl;
    return 1;
  }
  return 0;
}


int	load(char *name)
{
  FILE *f;
  if(!(f = fopen(name, "r"))){
    cerr << "Error: Simian load, failed to open " 
	 << name << " for reading" << endl;
    return 0;
  }
  else {
    cerr << "Loading ...." << endl;

    char line[1024];
    
    while(fgets(line, 1024, f)){
      gluvvPrimitive *tmp = renderables.getNext();
      cerr << "	 load: " << line << endl;
      while(tmp){
	tmp->load(line, f);
	tmp = tmp->getNext();
      }
    }

    fclose(f);
    cerr << "Done Loading" << endl;
    return 1;
  }
}


//================================================== Gluvv Quit
//==== delete crap here =======================================
void gluvvQuit(void)
{
  cerr << endl;
  exit(0); //clean exit
}
//=================================================================
//=================================================================
//=================================================================
//=================================================================
//=================================================================
//=================================================================
//=================================================================
//=================================================================
//=================================================================
//=================================================================

void glutSlaveDisp(void)
{
  if (!synch_firstpass && synch_dont_redraw) {
    //alog(<< "slave not redrawing" << endl);
  } else {
    synch_firstpass = false;
    //alog(<< "slave is redrawing" << endl);
    selfDisplay();
  }
  synch_dont_redraw = false;
}


//____________________________________________________________________________
//
// The master's glut stuff
//____________________________________________________________________________

void glutSlaveReshape(int w, int h)
{
  //alog(<< "RESHAPE " << w << " " << h << endl);
  synch_dont_redraw = true;
}

void glutMasterReshape(int w, int h)
{

  // What: The window is being reshaped.

  alog( << "M: starting handling reshape interrupt: " << w << ", " << h << endl);
  //aprint( << "M: starting handling reshape interrupt: " << w << ", " << h << endl);
  
  // Tell the slaves to reshape to wxh window.
  int buffer[CMD_BUFFER_MAX_SIZE];
  buffer[0] = CMD_RESHAPE;
  buffer[1] = w;
  buffer[2] = h;
  alog( << "s bcast master reshape " << endl);
  MPI_Bcast(buffer, CMD_BUFFER_MAX_SIZE, MPI_INT, 0, worldComm);
  alog( << "e bcast master reshape " << endl);
  
  selfReshape(w, h);

  alog( << "M: finished handling reshape interrupt: " << w << ", " << h << endl);
}

//____________________________________________________________________________

void glutMasterMouse(int button, int state, int x, int y)
{
  // What: The mouse has been used.
  
  if (record) {
    fprintf(lfile, "B %f %f %d %d\n", (float)x/(float)SCREEN_RES, (float)y/(float)SCREEN_RES, button, state);
  }

#ifdef DEBUG
  cerr << "Masters's mouse: " 
       << button << " " << state << " " << x << " " << y << endl;
#endif
  
  if (!playback) {
    if ( glutGetModifiers() & GLUT_ACTIVE_CTRL ) gluvv.mouse.ctrl = 1;
    else gluvv.mouse.ctrl = 0;
    if ( glutGetModifiers() & GLUT_ACTIVE_SHIFT ) gluvv.mouse.shift = 1;
    else gluvv.mouse.shift = 0;
    if ( glutGetModifiers() & GLUT_ACTIVE_ALT ) gluvv.mouse.alt = 1;
    else gluvv.mouse.alt = 0;
  }

  // Tell the slaves that the mouse has been used.
  int buffer[CMD_BUFFER_MAX_SIZE];
  buffer[0] = CMD_MOUSE;
  buffer[1] = button;
  buffer[2] = state;
  buffer[3] = x;
  buffer[4] = y;
  buffer[5] = gluvv.mouse.ctrl;
  buffer[6] = gluvv.mouse.shift;
  buffer[7] = gluvv.mouse.alt;

  alog( << "s bcast master mouse " << endl);
  MPI_Bcast(buffer, CMD_BUFFER_MAX_SIZE, MPI_INT, 0, worldComm);
  alog( << "e bcast master mouse " << endl);

  // Mouse the master.
  selfMouse(button, state, x, y);
}

//____________________________________________________________________________

void glutMasterMotion(int x, int y)
{
  if ( record ) {
    fprintf(lfile, "M %f %f %d %d\n", (float)x/(float)SCREEN_RES, (float)y/(float)SCREEN_RES, -1, -1 );
  }

  // What: The mouse has moved.

#ifdef DEBUG
  cerr << "Masters's mouse move to : " << x << " " << y << endl;
#endif

  // Tell the slaves that the mouse has moved.
  int buffer[CMD_BUFFER_MAX_SIZE];
  buffer[0] = CMD_MOTION;
  buffer[1] = x;
  buffer[2] = y;

  
  alog( << "s bcast master motion " << endl);
  MPI_Bcast(buffer, CMD_BUFFER_MAX_SIZE, MPI_INT, 0, worldComm);
  alog( << "e bcast master motion " << endl);
  
  
  // Move the master's mouse.
  selfMotion(x, y);
  
}

//____________________________________________________________________________
//
// The self stuff
//____________________________________________________________________________


void selfDisplay(void) //eye was not used anywhere in this function
{

  static int framenum = 0;

  //alog(<< "framenum is " << framenum++ << endl);

#ifdef MPE_LOGGING
  if ( worldId || masterIsSlave )
    MPE_Log_event(MPE_EVENT_RENDER_VOL_B, 0, "render");
#endif
  updateUI();
  timestampe(ntime, rts, rte);   
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(gluvv.env.eye[0],
	    gluvv.env.eye[1], 
	    gluvv.env.eye[2],  
	    gluvv.env.at[0],
	    gluvv.env.at[1], 
	    gluvv.env.at[2],  
	    gluvv.env.up[0], 
	    gluvv.env.up[1], 
	    gluvv.env.up[2]);
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(gluvv.env.frustum[0], //left
	    gluvv.env.frustum[1], //right
	    gluvv.env.frustum[2], //top
	    gluvv.env.frustum[3], //bottom
	    gluvv.env.clip[0],	  //front
	    gluvv.env.clip[1]);	  //back
  
  glMatrixMode(GL_MODELVIEW);
  
  glDrawBuffer(GL_BACK);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
  glDepthFunc(GL_LESS);
  glDisable(GL_BLEND);
 
  float myZval = -1;
  //if ( worldId || masterIsSlave )
    {
      myZval = display(0); // our local display function
    }  
  timestampe(rtime, rts, rte);

  // Compute the correct draw order.
  //________________________________
  //

  timestampe(ntime, sts, ste);
  
  MpiIdType *algRankToProc = new MpiIdType[slavesSize];	 
  
  int *algProcToRank = new int[worldSize];
  
  if ( worldId == 0 )
    {
      // Get the Z values for each subvolume.
      //
      
      // zval : MpiIdType -> float
      float *zval = new float[worldSize];
      
      int received = 0;
      
      if ( masterIsSlave )
	{
	  zval[0] = myZval;
	  ++received;
#ifdef DEBUG_BLENDING
	  cerr << "MPI #0 " << " is at Z = " << myZval << endl;
#endif
	}
      
      {
	float tmp;	
	MPI_Status status;
	
	while ( received < slavesSize )
	  { 
	    alog( << "s recv zval " << endl);
	    MPI_Recv(&tmp, 1, MPI_FLOAT, 
		     MPI_ANY_SOURCE, TAG_ZVAL, worldComm, &status);  
	    alog( << "e recv zval " << endl);
	    
	    zval[status.MPI_SOURCE] = tmp;
	    ++received;
	    
#ifdef DEBUG_BLENDING
	    cerr << "MPI #" << status.MPI_SOURCE 
		 << " is at Z = " << zval[status.MPI_SOURCE] << endl;
#endif
	  }
      }
      
      // Reorder the subvolumes from back to front.
      //
      {
	int ii, jj;
	
	for ( ii = 0; ii < slavesSize; ++ii ) 
	  {
	    if ( masterIsSlave )
	      algRankToProc[ii] = (MpiIdType)ii;
	    else
	      algRankToProc[ii] = (MpiIdType)ii + 1;
	  }
	
	for ( ii = 0; ii < slavesSize; ++ii )
	  { 
	    //now sort
	    for ( jj = ii + 1; jj < slavesSize; ++jj )
	      {
		if( zval[algRankToProc[ii]] < zval[algRankToProc[jj]])
		  {
		    //test for swap
		    MpiIdType tmp = algRankToProc[ii];
		    algRankToProc[ii] = algRankToProc[jj];
		    algRankToProc[jj] = tmp;
		  }
	      }
	  }
	
	for ( ii = 0; ii < slavesSize; ++ii )
	  {
	    algProcToRank[algRankToProc[ii]] = ii;
	  }

#ifdef DEBUG_BLENDING
	cerr << "Here is the back to front order of processes:" << flush;
	for ( ii = 0; ii < slavesSize; ++ii )
	  {
	    cerr << " #" << algRankToProc[ii] << flush;
	  }
	cerr << endl;
#endif
      }
      
      // Broadcast the order.
      alog( << "s bcast order " << endl);
      MPI_Bcast(algRankToProc, slavesSize, MPI_INT, 0, worldComm);
      MPI_Bcast(algProcToRank, worldSize, MPI_INT, 0, worldComm);
      alog( << "e bcast order " << endl);

      delete [] zval;
    }
  else
    {
      alog( << "s send zval " << endl);
      // Send the Z value to the master process.
      MPI_Send(&myZval, 1, MPI_FLOAT, 0, TAG_ZVAL, worldComm);	
      alog( << "e send zval " << endl);
      
      alog( << "s bcast order " << endl);
      // Broadcast the order.
      MPI_Bcast(algRankToProc, slavesSize, MPI_INT, 0, worldComm);
      MPI_Bcast(algProcToRank, worldSize, MPI_INT, 0, worldComm);
      alog( << "e bcast order " << endl);
    }  
  glFinish();
  timestampe(srttime, sts, ste);
  
#ifdef MPE_LOGGING
  if ( worldId || masterIsSlave )
    MPE_Log_event(MPE_EVENT_RENDER_VOL_E, 0, "rendered");
#endif
  
  timestampe(ntime, cts, cte);
  mySortLast->Compose(algRankToProc, algProcToRank);
  timestampe(cotime, cts, cte);
  
  delete [] algRankToProc;
  delete [] algProcToRank;

#ifdef MPE_LOGGING
  MPE_Log_event(MPE_EVENT_DRAW_PIXELS_B, 0, "draw");
#endif
  

  timestampe(ntime, dts, dte);
  //if ( worldId == 0 )
    {	   
      mySortLast->Display(fpsFont, fpsString); 
    }
  
    //MPI_Barrier(worldComm);

  glutSwapBuffers();
  timestampe(dtime, dts, dte);
  
#ifdef MPE_LOGGING
  MPE_Log_event(MPE_EVENT_DRAW_PIXELS_E, 0, "drawn");
#endif

  GlErr("gluvv:","glutdisp()");
}


//this draws only the parts of the widgets that lie behind the volume's 
//bounding box
void drawBackWidgets()
{
  //Mark wants to redo this to draw the front sides using bbox plane normals
  //but Dave says I don't care.

  glPushMatrix();{
    
    if (diag_show_objs_behind) {

      //draw the back of the volume bbox into the depth buffer
      gluvvPrimitive *tmp;
         
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

      glColorMask(false,false,false,false);	  
      glEnable(GL_CULL_FACE);
      glCullFace(GL_FRONT);
      if (myVolRen) myVolRen->draw_bbox(false);	  
      glDisable(GL_CULL_FACE);

      //draw the widgets that lie behind it, with correct coloration
      glColorMask(true,true,true,true);	  
      
      glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
      glEnable(GL_CULL_FACE);
      glCullFace(GL_BACK);
      gluvv.reblend = GB_ZERO;	  
      tmp = renderables.getNext();	  
      while(tmp){
	glDepthFunc(GL_GREATER);
	if (tmp != myVolRen) tmp->draw();
	tmp = tmp->getNext();
      }
      glDisable(GL_CULL_FACE);
      glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);

      if (gluvv.env.bgColor == 0) {
	glDisable(GL_LIGHTING);
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE_MINUS_DST_ALPHA, GL_ONE);
	gluvv.reblend = GB_NONE;
	glDisable(GL_DEPTH_TEST);
	
	float nodebgcolor = 1;//(float)(worldId+1)/(float)worldSize;

	glColor4f(nodebgcolor, nodebgcolor, nodebgcolor, 1);
	glBegin(GL_QUADS);{
	  glVertex3f(-10,-10,-2);
	  glVertex3f(-10,10,-2);
	  glVertex3f(10,10,-2);
	  glVertex3f(10,-10,-2);
	} glEnd();
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
      }

    }
  } glPopMatrix();
}
