#ifndef __SAVEPPM_DOT_H
#define __SAVEPPM_DOT_H

#include "saveppm.h"

#include <iostream>
#include <fstream>
using namespace std;
static int once = 0;

void save_ppm(const char * const filename,
	      int sx, int sy, 
	      int bpp, const unsigned char * buf)
  //____________________________________________________________________________
{
  int R = 3;
  int G = 2;
  int B = 1;
  int A = 0;
  
  if ( ! once )
    {
      ofstream output(filename, ios::out);
      
      if ( ! output ) 
	{
	  cerr << "ERROR: can't open file " << filename << endl;
	  return;
	}
      
      if ( bpp == 1 || bpp == 2 )
	output << "P2\n# CREATOR: " << endl;
      else if ( bpp == 3 || bpp == 4 )
	output << "P3\n# CREATOR: " << endl;
      else {
	cerr << "ERROR: unknown number of bytes per pixel " << bpp << endl;
	return;
      }
      
      output << sx << " " << sy << endl;
      output << 255 << endl;
      
      for ( int row = sy - 1; row >= 0; --row ) 
	{
	  for ( int col = 0; col < sx; ++col ) 
	    {	   
	      int p = bpp * ( row * sx + col );
	      switch (bpp) {
	      case 2:
	      case 1:
		output << (int) buf[p] << endl;
		break;
	      default:
		output << (int) buf[p + R] << " " << (int) buf[p + G]
		       << " " << (int) buf[p + B] << endl;
		break;
	      }
	    }
	}
      
      cout << "INFO: image saved in " << filename << endl;
    }
  
  return;
}

void save_ppm_alpha(const char * const filename,
		    int sx, int sy, 
		    int bpp, const unsigned char * buf)
  //____________________________________________________________________________
{
  int R = 3;
  int G = 2;
  int B = 1;
  int A = 0;
  
  if ( ! once )
    {
      ofstream output(filename, ios::out);
      
      if ( ! output ) 
	{
	  cerr << "ERROR: can't open file " << filename << endl;
	  return;
	}
      
      if ( bpp == 1 || bpp == 2 )
	output << "P2\n# CREATOR: " << endl;
      else if ( bpp == 3 || bpp == 4 )
	output << "P3\n# CREATOR: " << endl;
      else {
	cerr << "ERROR: unknown number of bytes per pixel " << bpp << endl;
	return;
      }
      
      output << sx << " " << sy << endl;
      output << 255 << endl;
      
      for ( int row = sy - 1; row >= 0; --row ) 
	{
	  for ( int col = 0; col < sx; ++col ) 
	    {	   
	      int p = bpp * ( row * sx + col );
	      switch (bpp) {
	      case 2:
	      case 1:
		output << (int) buf[p] << endl;
		break;
	      default:
		//output << (int) buf[p + R] << " " << (int) buf[p + G]
		//	 << " " << (int) buf[p + B] << endl;
		if (buf[p+A] == 0){
		  output << 255 << " " << 0 << " " << 0 <<endl;
		}
		else if (buf[p+A] == 255){
		  output << 0 << " " << 255 << " " << 0 <<endl; 
		}
		else{
		  output << 0 << " " << 0 << " " << buf[p+A] <<endl; 
		}
		
		break;
	      }
	    }
	}
      
      cout << "INFO: image saved in " << filename << endl;
    }
  
  return;
}

#endif
