//cluster simian volume preprocessor
//DeMarle 06/17/2004
//based on bricker.cpp by Mark Hartner and genVGH.cpp by Joe Kniss

//This program reads in a set of volumes prepared by vol_prep_part1, merges
//them into a format suitable for simian, and then breaks the merged volume
//into N sub blocks suitable for N rendering processors in cluster simian.
// This program operates in a slice by slice fashion from the input files, 
//so that it doesn't need to keep the whole volume in memory, which makes it 
//suitable for large files.

#include <stdio.h> //printf, scanf
#include <unistd.h> //read
#include <fcntl.h>  //O_LARGEFILE
#include <math.h> //sqrt
#include <stdlib.h> //exit
#include <string.h> //memset, memcpy

//padded onto power of two
int pzd, pyd, pxd; 
int pydXpxd = 0;
int pydXpzd = 0;
//dimensions of a subblock within the padded volume
int bzd, byd, bxd;

//floating point dimensions of volume data in view space
float fzd, fyd, fxd;


#define MAXNAMELEN 256
char filename[MAXNAMELEN]; //name of raw data file
char ifilename[MAXNAMELEN]; //name for transient an input file
char ofilename[MAXNAMELEN]; //name for transient an output file

int pq_v_fd; //padded quantized value file descriptor
int pq_gm_fd;  //padded quantized gradient magnitude file descriptor
int pq_h_fd;  //padded quantized hessian file descriptor
int pq_g_fd;  //padded quantized gradient file descriptor

int rc; //system call return code


int main(int argc, char **argv) {

  //STEP 1////////////////////////////////////////////////////////////////
  ///ASKING FOR INPUT VOLUME STATS

  printf("What is the input file name? ");
  rc = scanf("%s", filename);
  printf("filename : %s\n", filename);

  printf("What are the padded integral dimensions (in x y z format where x changes fastest)? ");
  scanf("%d %d %d", &pxd, &pyd, &pzd);
  printf("dimensions : x=%d, y=%d, z=%d\n", pxd, pyd, pzd);

  printf("What are the floating point sizes (in x. y. z. format)? ");
  scanf("%f %f %f", &fxd, &fyd, &fzd);
  printf("float dimensions : x=%f, y=%f, z=%f\n", fxd, fyd, fzd);

  int nnodes = 0;
  printf("How many rendering nodes? ");
  scanf("%d", &nnodes);
  printf("num nodes : %d\n", nnodes);
  
  //STEP 2/////////////////////////////////////////////////////////////////
  ///SPLITTING INTO SUBVOLUMES
  printf("Splitting into sub volumes.\n");

  int worldSize = nnodes;
  int dataSize = pzd * pyd * pxd;
  int maxsz = dataSize/worldSize;
 
  pydXpxd = pyd*pxd;

  {    
    //INPUT FILES
    sprintf(ifilename, "%s.pqv_tmp", filename);
    pq_v_fd = open(ifilename, O_RDONLY|O_LARGEFILE);
    if (pq_v_fd <= 0) {
      printf("Problem opening input file.\n");
      return -1;
    }
    sprintf(ifilename, "%s.pqgm_tmp", filename);
    pq_gm_fd = open(ifilename, O_RDONLY|O_LARGEFILE);
    if (pq_gm_fd <= 0) {
      printf("Problem opening input file.\n");
      return -1;
    }
    sprintf(ifilename, "%s.pqh_tmp", filename);
    pq_h_fd = open(ifilename, O_RDONLY|O_LARGEFILE);
    if (pq_h_fd <= 0) {
      printf("Problem opening input file.\n");
      return -1;
    }
    sprintf(ifilename, "%s.pqg_tmp", filename);
    pq_g_fd = open(ifilename, O_RDONLY|O_LARGEFILE);
    if (pq_g_fd <= 0) {
      printf("Problem opening input file.\n");
      return -1;
    }

    //number of sub blocks across each axis
    int nzb = 1; //number of z blocks
    int nyb = 1;
    int nxb = 1;
    while(((pzd/(float)nzb)*(pyd/(float)nyb)*(pxd/(float)nxb)) > maxsz){
      if(nzb > nyb){
	if(nyb > nxb)
	  nxb*=2;
	else
	  nyb*=2;
      } else
	nzb*=2;
    }

    bzd = (int)(pzd/(float)nzb);  //sub block dimensions
    byd = (int)(pyd/(float)nyb);
    bxd = (int)(pxd/(float)nxb);

    printf("num sub blocks %d\n", nzb*nyb*nxb);
    printf("sub block dims x=%d y=%d z=%d\n", bxd, byd, bzd);
        
    //create space for a slab of sub blocks
    unsigned char *vghdata[nyb][nxb];
    unsigned char *graddata[nyb][nxb];
    for (int cvy = 0; cvy < nyb; cvy++) {
      for (int cvx = 0; cvx < nxb; cvx++) {
	vghdata[cvy][cvx] =  new unsigned char[byd*bxd*3];
	graddata[cvy][cvx] =  new unsigned char[byd*bxd*3];
      }
    }

    //space for a slice from each of the padded quantized data files
    unsigned char dataI[pydXpxd];
    unsigned char gmagI[pydXpxd];
    unsigned char hessI[pydXpxd];
    unsigned char gradI[pydXpxd*3];

    //write the global volume information to the trex file
    sprintf(ofilename, "%s.trex", filename);
    FILE *trex;
    if(!(trex = fopen(ofilename, "w"))){
      printf("Problem opening output file.\n");
      return -1;
    }
    fprintf(trex, "#  Meta Volume  #\n");
    fprintf(trex, "\n\n#    Global info\n");
    fprintf(trex, "Data Set Name:	     %s\n", filename);
    fprintf(trex, "#  NOTE: Data Set Files uses an implicit extension: filename.timestep.subvol\n");
    fprintf(trex, "Data Set Files:	   %s\n", filename);
    fprintf(trex, "Number of Time Steps:  1, 0, 0\n");
    fprintf(trex, "Volume Size int:	   %d, %d, %d\n", pxd, pyd, pzd);
    fprintf(trex,  "Volume Size float:	   %f, %f, %f\n", fxd, fyd, fzd);
    fprintf(trex, "\n\n#    Subvolume info\n");
    fprintf(trex, "Number of Sub Volumes: %d\n", nzb*nyb*nxb);


    FILE *subvolfiles[2][nyb][nxb];

    int absz=0;
    int svol = 0, cvol = 0;
    for(int cvz=0; cvz < nzb; cvz++) { //iterate up through slabs of Z sub blocks

      //we've finished a z slab, write out each of the subvolumes we've just made
      for (int cvy = 0; cvy < nyb; cvy++) {
	for (int cvx = 0; cvx < nxb; cvx++) {
	  //append next subvolume info to the trex header file	  

	  //sub block size if same for all, but it doesn't hurt to recompute it
	  int ziSize = bzd;
	  int yiSize = byd;
	  int xiSize = bxd;
	  
	  float zfSize = fzd*(ziSize/(float)pzd);
	  float yfSize = fyd*(yiSize/(float)pyd);
	  float xfSize = fxd*(xiSize/(float)pxd);
	  	  
	  //compute location of this particular sub block in the whole
	  int ziPos  = ziSize*cvz;
	  int yiPos  = yiSize*cvy;
	  int xiPos  = xiSize*cvx;	 
	  float zfPos  = zfSize*cvz;
	  float yfPos  = yfSize*cvy;
	  float xfPos  = xfSize*cvx;
	  
	  fprintf(trex, "SubVolume {\n");
	  fprintf(trex, "\tSize int:	  %d, %d, %d\n", xiSize, yiSize, ziSize);
	  fprintf(trex, "\tSize float:	  %f, %f, %f\n", xfSize, yfSize, zfSize);
	  fprintf(trex, "\tPos int:	  %d, %d, %d\n", xiPos, yiPos, ziPos);
	  fprintf(trex, "\tPos float:	  %f, %f, %f\n", xfPos, yfPos, zfPos);
	  fprintf(trex, "}\n");
	
	  //create a temporary subvolume file for the VGH data
	  sprintf(ofilename, "%s.%04d.%02d_VGH", filename, 0, svol);	  
	  if(!(subvolfiles[0][cvy][cvx] = fopen(ofilename, "wb"))){
	    printf("Problem opening output file.\n");
	    return -1;
	  }

	  //create a temporary subvolume file for the NNN data
	  sprintf(ofilename, "%s.%04d.%02d_NNN", filename, 0, svol);	  
	  if(!(subvolfiles[1][cvy][cvx] = fopen(ofilename, "wb"))){
	    printf("Problem opening output file.\n");
	    return -1;
	  }

	  svol ++;	 
	}
      }

      for(int i = 0; i < bzd; i++) { //inside each slab, iterate up through the Z slices

	absz++;
	if (!(absz % 10)) printf("blocked slice %d\n", absz);

	//read in a slice of the processed volume from the four quantized volume files
	rc = read(pq_v_fd, dataI, pydXpxd*sizeof(unsigned char));
	if (rc != pydXpxd*sizeof(unsigned char)) {
	  printf("Problem reading input file.\n");
	  return -1;
	}
	rc = read(pq_gm_fd, gmagI, pydXpxd*sizeof(unsigned char));
	if (rc != pydXpxd*sizeof(unsigned char)) {
	  printf("Problem reading input file.\n");
	  return -1;
	}
	rc = read(pq_h_fd, hessI, pydXpxd*sizeof(unsigned char));
	if (rc != pydXpxd*sizeof(unsigned char)) {
	  printf("Problem reading input file.\n");
	  return -1;
	}
	rc = read(pq_g_fd, gradI, pydXpxd*3*sizeof(unsigned char));
	if (rc != pydXpxd*3*sizeof(unsigned char)) {
	  printf("Problem reading input file.\n");
	  return -1;
	}
		
	//iterate over each word in the slice, and merge the four input values into the proper places in the subvolume data
	for(int j=0; j<pyd; j++) {
	  for(int k=0; k<pxd; k++) {

	    int cvy = j/byd; //determine current y sub block slab
	    int cvx = k/bxd; //...x...
	    
	    int instartbyte = j*pxd + k;
	    int outstartbyte = ((j%byd)*bxd + (k%bxd))*3; //times 3 because we store V, G and H

	    vghdata[cvy][cvx][outstartbyte + 0] = dataI[instartbyte];
	    vghdata[cvy][cvx][outstartbyte + 1] = gmagI[instartbyte];
	    vghdata[cvy][cvx][outstartbyte + 2] = hessI[instartbyte];

	    int ginstartbyte = instartbyte*3; //*3 for X,Y,Z components of the gradient vector
	    graddata[cvy][cvx][outstartbyte + 0] = gradI[ginstartbyte + 0];
	    graddata[cvy][cvx][outstartbyte + 1] = gradI[ginstartbyte + 1];
	    graddata[cvy][cvx][outstartbyte + 2] = gradI[ginstartbyte + 2];
	  }
	}

	//we've finished a z slice, append to each of the subvolumes
	for (int cvy = 0; cvy < nyb; cvy++) {
	  for (int cvx = 0; cvx < nxb; cvx++) {
	    
	    //write out vgh data
	    unsigned int n;
	    n= fwrite((void *) vghdata[cvy][cvx], sizeof(unsigned char), byd*bxd*3, subvolfiles[0][cvy][cvx]);
	    if ( n != (unsigned int)byd*bxd*3) {
	      printf("Problem writing output file.\n");
	      return -1;
	    }
	    
	    //write out the normals (grads)
	    n= fwrite((void *) graddata[cvy][cvx], sizeof(unsigned char), byd*bxd*3, subvolfiles[1][cvy][cvx]);
	    if( n != (unsigned int)byd*bxd*3) {
	      printf("Problem writing output file.\n");
	      return -1;
	    }
	  }
	}
	
      }

      //we've finished a z slab, finish the subvolumes we've just made
      for (int cvy = 0; cvy < nyb; cvy++) {
	for (int cvx = 0; cvx < nxb; cvx++) {
	  //append subvolume info to the trex header file	  

	  fclose(subvolfiles[0][cvy][cvx]);
	  fclose(subvolfiles[1][cvy][cvx]);

	  char fname[MAXNAMELEN];
	  sprintf(fname, "%s.%04d.%02d", filename, 0, cvol++);

	  char cmd[3*MAXNAMELEN + 8];
	  sprintf(cmd, "cat %s_VGH %s_NNN > %s\n", fname, fname, fname);
	  system(cmd);

	  sprintf(cmd, "rm -f %s_VGH %s_NNN\n", fname, fname);
	  system(cmd);
	}
      }

    }

    fclose(trex);
    close(pq_v_fd);
    close(pq_g_fd);
    close(pq_gm_fd);
    close(pq_h_fd);
  }


  printf("Finished, use \"rm -f *.*_tmp\" to clear away the intermediate files\n");
}
