//large volume subvolume extractor
//DeMarle 06/29/2004

//This program reads in a scalar volume and outputs a new subvolume extracted
//from some user specified region inside the original.

// This program operates in a slice by slice fashion from the input 
//files, so that it doesn't need to keep the whole volume in memory, which 
//makes it suitable for large files.

#include <stdio.h> //printf, scanf
#include <unistd.h> //read
#include <fcntl.h>  //O_LARGEFILE
#include <math.h> //sqrt
#include <stdlib.h> //exit
#include <string.h> //memset, memcpy

//integral dimensions of input volume data
int izd, iyd, ixd;
int iydXixd = 0;


//dimensions of the ppm image
int dim1, dim2;

#define MAXNAMELEN 256
char filename[MAXNAMELEN]; //name of raw data file
char ofilename[MAXNAMELEN]; //name for subsampled output file

int c_v_in_fd; //unsigned character value (aka data aka scalar) input file descriptor
int c_v_out_fd; //unsigned character value (aka data aka scalar) output file descriptor


int rc; //system call return code

int main(int argc, char **argv) {
  
  //STEP 1////////////////////////////////////////////////////////////////
  ///ASKING FOR INPUT VOLUME STATS
  
  printf("What is the input file name? ");
  rc = scanf("%s", filename);
  printf("filename : %s\n", filename);
  
  printf("What are the integral dimensions (in x y z format where x changes fastest)? ");
  scanf("%d %d %d", &ixd, &iyd, &izd);
  printf("dimensions : x=%d, y=%d, z=%d\n", ixd, iyd, izd);

  if ((ixd != 1) && (iyd != 1) && (izd != 1)) {
    printf("To make a ppm at least one of the volume dimensions should be 1\n");    return -1;
  }

  if (ixd == 1) {
    dim1 = iyd;
    dim2 = izd;
  } else if (iyd == 1) {
    dim1 = ixd;
    dim2 = izd;
  } else {
    dim1 = ixd;
    dim2 = iyd;      
  }

  //STEP 2//////////////////////////////////////////////////////////////////
  
  //INPUT FILES
  c_v_in_fd = open(filename, O_RDONLY|O_LARGEFILE);
  if (c_v_in_fd <= 0) {
    printf("Problem opening input file.\n");
    return -1;
  }
  
  //OUTPUT FILES
  sprintf(ofilename, "%s.ppm", filename);
  c_v_out_fd = open(ofilename, O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR);
  if (c_v_out_fd <= 0) {
    printf("Problem opening output file.\n");
    return -1;
  }
  

  char ppmheader[256];
  sprintf(ppmheader, "P6\n%d\n%d\n255\n", dim1, dim2);
  rc = write(c_v_out_fd, (void*)ppmheader, strlen(ppmheader));
  if (rc != strlen(ppmheader)) {
    printf("Problem writing output file.\n");
    return -1;
  }
  
  iydXixd = iyd*ixd;
  unsigned char *c_v_in_slice, *c_v_out_slice;
  c_v_in_slice = new unsigned char[iydXixd];
  c_v_out_slice = new unsigned char[iydXixd*3];
  
  int readslice = 0;

  while (readslice < izd) {
    
    rc = read(c_v_in_fd, (unsigned char*)c_v_in_slice, iydXixd*sizeof(unsigned char));
    if (rc != iydXixd*sizeof(unsigned char)) {
      printf("Problem reading input file.\n");
      return -1;
    }

    for (int i = 0; i < iyd; i++) {
      for (int j = 0; j < ixd; j++) {
	unsigned char ival = c_v_in_slice[i*ixd+j];
	c_v_out_slice[(i*ixd+j)*3 + 0] = ival;
	c_v_out_slice[(i*ixd+j)*3 + 1] = ival;
	c_v_out_slice[(i*ixd+j)*3 + 2] = ival;
      }
    }
      
    rc = write(c_v_out_fd, (void*)c_v_out_slice, iydXixd*3*sizeof(unsigned char));
    if (rc != iydXixd*3*sizeof(unsigned char)) {
      printf("Problem writing output file.\n");
      return -1;
    }
    
    readslice ++;
  }

  close(c_v_in_fd);
  close(c_v_out_fd);

}
