//large volume subsampler
//DeMarle 06/29/2004

//This program reads in a scalar volume and subsamples it to 1/8 the original
//size. It does simple averaging on blocks of 8 neighboring scalars to do so.

// This program operates in a slice by slice fashion from the input 
//files, so that it doesn't need to keep the whole volume in memory, which 
//makes it suitable for large files.

#include <stdio.h> //printf, scanf
#include <unistd.h> //read
#include <fcntl.h>  //O_LARGEFILE
#include <math.h> //sqrt
#include <stdlib.h> //exit
#include <string.h> //memset, memcpy

//integral dimensions of input volume data
int izd, iyd, ixd;
int iydXixd = 0;

//integral dimensions of ouput volume data
int ozd, oyd, oxd;
int oydXoxd = 0;


#define MAXNAMELEN 256
char filename[MAXNAMELEN]; //name of raw data file
char ofilename[MAXNAMELEN]; //name for subsampled output file

int c_v_in_fd; //unsigned character value (aka data aka scalar) input file descriptor
int c_v_out_fd; //unsigned character value (aka data aka scalar) output file descriptor


int rc; //system call return code

int main(int argc, char **argv) {

  //STEP 1////////////////////////////////////////////////////////////////
  ///ASKING FOR INPUT VOLUME STATS

  printf("What is the input file name? ");
  rc = scanf("%s", filename);
  printf("filename : %s\n", filename);

  printf("What are the integral dimensions (in x y z format where x changes fastest)? ");
  scanf("%d %d %d", &ixd, &iyd, &izd);
  printf("dimensions : x=%d, y=%d, z=%d\n", ixd, iyd, izd);
  
  //STEP 2//////////////////////////////////////////////////////////////////

  //INPUT FILES
  c_v_in_fd = open(filename, O_RDONLY|O_LARGEFILE);
  if (c_v_in_fd <= 0) {
    printf("Problem opening input file.\n");
    return -1;
  }

  //OUTPUT FILES
  sprintf(ofilename, "%s.smaller", filename);
  c_v_out_fd = open(ofilename, O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR);
  if (c_v_out_fd <= 0) {
    printf("Problem opening output file.\n");
    return -1;
  }

  iydXixd = iyd*ixd;
  unsigned char *c_v_in_slice[2];
  c_v_in_slice[0] = new unsigned char[iydXixd];
  c_v_in_slice[1] = new unsigned char[iydXixd];

  ozd = izd>>1;
  oyd = iyd>>1;
  oxd = ixd>>1;
  oydXoxd = oyd*oxd;
  unsigned char *c_v_out_slice;
  c_v_out_slice = new unsigned char[oydXoxd];

  int readslice = 0;

  while (readslice < (izd-1)) {

    if (!(readslice % 10)) {
      printf("Subsampling slices %d and %d\n", readslice, readslice+1);
    }

    rc = read(c_v_in_fd, (unsigned char*)c_v_in_slice[0], iydXixd*sizeof(unsigned char));
    if (rc != iydXixd*sizeof(unsigned char)) {
      printf("Problem reading input file.\n");
      return -1;
    }

    rc = read(c_v_in_fd, (unsigned char*)c_v_in_slice[1], iydXixd*sizeof(unsigned char));
    if (rc != iydXixd*sizeof(unsigned char)) {
      printf("Problem reading input file.\n");
      return -1;
    }

    for (int i = 0; i < (iyd-1); i+=2) {
      for (int j = 0; j < (ixd-1); j+=2) {
	c_v_out_slice[(i>>1)*oxd + (j>>1)] =
	  (unsigned char) ( (
			     (float)(c_v_in_slice[0][(i+0)*ixd+(j+0)]) +
			     (float)(c_v_in_slice[0][(i+0)*ixd+(j+1)]) +
			     (float)(c_v_in_slice[0][(i+1)*ixd+(j+0)]) +
			     (float)(c_v_in_slice[0][(i+1)*ixd+(j+1)]) +
			     (float)(c_v_in_slice[1][(i+0)*ixd+(j+0)]) +
			     (float)(c_v_in_slice[1][(i+0)*ixd+(j+1)]) +
			     (float)(c_v_in_slice[1][(i+1)*ixd+(j+0)]) +
			     (float)(c_v_in_slice[1][(i+1)*ixd+(j+1)]) 
			     ) / 8.0 );		  
      }
    }

    rc = write(c_v_out_fd, (void*)c_v_out_slice, oydXoxd*sizeof(unsigned char));
    if (rc != oydXoxd*sizeof(unsigned char)) {
      printf("Problem writing output file.\n");
      return -1;
    }

    readslice += 2;
  }

  close(c_v_in_fd);
  close(c_v_out_fd);

  printf("New dimensions are now x=%d y=%d z=%d\n", oxd, oyd, ozd);
}
