//large volume subvolume extractor
//DeMarle 06/29/2004

//This program reads in a scalar volume and outputs a new subvolume extracted
//from some user specified region inside the original.

// This program operates in a slice by slice fashion from the input 
//files, so that it doesn't need to keep the whole volume in memory, which 
//makes it suitable for large files.

#include <stdio.h> //printf, scanf
#include <unistd.h> //read
#include <fcntl.h>  //O_LARGEFILE
#include <math.h> //sqrt
#include <stdlib.h> //exit
#include <string.h> //memset, memcpy

//integral dimensions of input volume data
int izd, iyd, ixd;
int iydXixd = 0;

//integral start and end points of output volume data
int szc, syc, sxc;
int ezc, eyc, exc;
//integral dimensions of ouput volume data
int ozd, oyd, oxd;
int oydXoxd = 0;

#define MAXNAMELEN 256
char filename[MAXNAMELEN]; //name of raw data file
char ofilename[MAXNAMELEN]; //name for subsampled output file

int c_v_in_fd; //unsigned character value (aka data aka scalar) input file descriptor
int c_v_out_fd; //unsigned character value (aka data aka scalar) output file descriptor


int rc; //system call return code

int main(int argc, char **argv) {

  //STEP 1////////////////////////////////////////////////////////////////
  ///ASKING FOR INPUT VOLUME STATS

  printf("What is the input file name? ");
  rc = scanf("%s", filename);
  printf("filename : %s\n", filename);

  printf("What are the integral dimensions (in x y z format where x changes fastest)? ");
  scanf("%d %d %d", &ixd, &iyd, &izd);
  printf("dimensions : x=%d, y=%d, z=%d\n", ixd, iyd, izd);

  printf("What are the start x y and z coordinates of the subvolume? (inclusive) ");
  scanf("%d %d %d", &sxc, &syc, &szc);
  if (sxc < 0) sxc = 0;
  if (sxc > ixd-1) sxc = ixd-1;
  if (syc < 0) syc = 0;
  if (syc > iyd-1) syc = iyd-1;
  if (szc < 0) szc = 0;
  if (szc > izd-1) szc = izd-1;
  printf("start coordinates : x=%d, y=%d, z=%d\n", sxc, syc, szc);

  printf("What are the end x y and z coordinates of the subvolume? (inclusive) ");
  scanf("%d %d %d", &exc, &eyc, &ezc);
  if (exc < sxc) exc = sxc;
  if (exc > ixd-1) exc = ixd-1;
  if (eyc < syc) eyc = syc;
  if (eyc > iyd-1) eyc = iyd-1;
  if (ezc < szc) ezc = szc;
  if (ezc > izd-1) ezc = izd-1;
  printf("end coordinates : x=%d, y=%d, z=%d\n", exc, eyc, ezc);
  
  ozd = ezc-szc+1;
  oyd = eyc-syc+1;
  oxd = exc-sxc+1;

  //STEP 2//////////////////////////////////////////////////////////////////

  //INPUT FILES
  c_v_in_fd = open(filename, O_RDONLY|O_LARGEFILE);
  if (c_v_in_fd <= 0) {
    printf("Problem opening input file.\n");
    return -1;
  }

  //OUTPUT FILES
  sprintf(ofilename, "%s.subvol", filename);
  c_v_out_fd = open(ofilename, O_CREAT|O_WRONLY|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR);
  if (c_v_out_fd <= 0) {
    printf("Problem opening output file.\n");
    return -1;
  }

  iydXixd = iyd*ixd;
  unsigned char *c_v_in_slice;
  c_v_in_slice = new unsigned char[iydXixd];

  oydXoxd = oyd*oxd;
  unsigned char *c_v_out_slice;
  c_v_out_slice = new unsigned char[oydXoxd];

  int readslice = 0;

  while (readslice <= ezc) {

    if (!(readslice % 10)) {
      printf("Reading slice %d\n", readslice);
    }

    rc = read(c_v_in_fd, (unsigned char*)c_v_in_slice, iydXixd*sizeof(unsigned char));
    if (rc != iydXixd*sizeof(unsigned char)) {
      printf("Problem reading input file.\n");
      return -1;
    }

    if (readslice >= szc) {
      for (int i = syc; i <= eyc; i++) {
	for (int j = sxc; j <= exc; j++) {
	  c_v_out_slice[(i-syc)*oxd+(j-sxc)] = c_v_in_slice[i*ixd+j];
	}
      }
      
      rc = write(c_v_out_fd, (void*)c_v_out_slice, oydXoxd*sizeof(unsigned char));
      if (rc != oydXoxd*sizeof(unsigned char)) {
	printf("Problem writing output file.\n");
	return -1;
      }
    }

    readslice ++;
  }

  close(c_v_in_fd);
  close(c_v_out_fd);

  printf("New dimensions are now x=%d y=%d z=%d\n", oxd, oyd, ozd);

}
